/*
C
C  _______________________________________________________________
C
C*   Licence
C    =======
C
C    You may use or modify this code for your own non commercial
C    purposes for an unlimited time. 
C    In any case you should not deliver this code without a special 
C    permission of ZIB.
C    In case you intend to use the code commercially, we oblige you
C    to sign an according licence agreement with ZIB.
C
C
C  _______________________________________________________________
C
*/

#include "kask.h"
#include "kasktri.h"
#include "kaskass.h"
#include "kasksol.h"
#include <math.h>

#define EDG_MAX 10

static INT AX,AY,BX;
static REAL *vecX;

static INT SetB(p)
  PT *p;
  {
    REAL dirVals[NO_OF_DIR_VALS];
    (actProblem->DirichF)(p->x,p->y,p->z,p->class,dirVals);
    RA(p,BX) = dirVals[0];
/*    RA(p,R_SOL) = dirVals[0];   */

    return true;
  }
static INT SetBE(ed)
  EDG *ed;
  {
	PT *P1 = ed->p1, *P2 = ed->p2;
	REAL x, y, z;
	REAL dirVals[NO_OF_DIR_VALS];

	x = (P1->x)+((P2->x)-(P1->x))*HALF;
	y = (P1->y)+((P2->y)-(P1->y))*HALF;
	z = (P1->z)+((P2->z)-(P1->z))*HALF;
	(actProblem->DirichF)(x,y,z,ed->class,dirVals);
 	RA(ed,BX) = dirVals[0];	  
    return true;
  }
void SetBound(x)
  INT x;
  {
    BX = x;
    ApplyP(SetB,dirichlet);
    if (iTo==10) ApplyE(SetBE,dirichlet);
    return;
  }

static INT SetZero(p)
  PT *p;
  {
    RA(p,BX) = ZERO;
    return true;
  }

static INT SetZeroE(ed)
  EDG *ed;
  {
    RA(ed,BX) = ZERO;
    return true;
  }

void SetBoundZero(x)
  INT x;
  {
    BX = x;
    ApplyP(SetZero,dirichlet);
    if (iTo==10) ApplyE(SetZeroE,dirichlet);
    return;
  }

static INT AddElem(t)
  TR3 *t;
  {
    INT    i, k;
    PT     **pts = &(t->p1);
    REAL   hilfe;

    if ((actProblem->NumAss)(t)==false) return false;

    for (i = iFrom; i < iTo; i++)
       {
        if ( ((pts[i]->boundP)!=DIRICHLET) || (actSolve->compNormP))   
          {
            for (k = kFrom; k < kTo; k++)
               {
                 if (i>k)
                   {
                     hilfe = assA[i][k]*RA(pts[k],AX);
                     RA(pts[i],AY) += hilfe;
                   }
                 else
                   {
                     hilfe = assA[k][i]*RA(pts[k],AX);
                     RA(pts[i],AY) += hilfe;
                   }
               }
      
           }
         else 
           {
            RA(pts[i],AY) = RA(pts[i],AX);
           }
        RA(pts[i],R_DIAG) += assA[i][i];
      }
    return true;
  }
 
INT axMul(x, y)
  INT x, y;
  {
    partP = P_STIFF;
    SetZeroField(y);
    SetZeroField(R_DIAG);
    AX = x; AY=y;
    if (ApplyT3(AddElem,all)==false) return false;
    return true;
  }

static INT divD(p)
  PT *p;
  { 
  	RA(p,AX) = RA(p,AX)/RA(p,AY);     
	return true;
  }

static INT storeElem(p)
  PT *p;
  {
	RA(p,BX) = vecX[p->index];
	return true;
  }

static INT fetchElem(p)
  PT *p;
  {
	vecX[p->index] = RA(p,BX);
	return true;
  }


INT MulD(x,y)
  INT x, y;
  {
 	vecX = (REAL*)(actSolve->rhs);
	BX = x;
	ApplyP(fetchElem,direct);

	actSolve->DirectSol( (actSolve->env),
                         (actSolve->decomp),
		            	 (actSolve->diagonal),
		            	 (actSolve->rhs),
			             (actSolve->solution),
                          actSolve->dirDim);
	vecX = (REAL*)(actSolve->solution);
	ApplyP(storeElem,direct);
 
	AX = x; AY = y;
	ApplyP(divD,nonDirect);
	return true;
  }

static INT sxE(ed)
  EDG *ed;
  {
    PT *pm = ed->pm;
	REAL hilfe;
  
    if (pm==nil)
	  if (iTo==10)
	    {
    	  hilfe = (RA(ed->p1,BX)+RA(ed->p2,BX))*0.5;
		  RA(ed,BX) += hilfe;
		  return true;
		}
	  else return true;
    hilfe = (RA(ed->p1,BX)+RA(ed->p2,BX))*0.5;
	RA(pm,BX) += hilfe;
    return true;
  }

static INT stxE(ed)
  EDG *ed;
  {
    REAL xim;
    PT *pm = ed->pm;
  
	if (pm==nil)
	  if (iTo==10)
	    {
    	  xim = RA(ed,BX)*0.5;
    	  RA(ed->p1,BX) += xim; RA(ed->p2,BX) += xim;
		  return true;
		}
	  else return true;
    xim = RA(pm,BX)*0.5;
    RA(ed->p1,BX) += xim; RA(ed->p2,BX) += xim;
    return true;
  }

void pcxMul(x,y)
  INT x, y;
  {
    assign(y,x);
	SetBoundZero(y);
	MulD(y,R_DIAG);
	assign(y,x);

	SetBoundZero(y);

    return;
  }

static INT AssB(t)
  TR3 *t;
  {
    INT i;
	PT **pts = &(t->p1);
 
    if ((actProblem->NumAss)(t)==false) return false;
    for(i = iFrom; i<iTo; i++) RA(pts[i],BX) += assB[i];
    return true;
  }
  
INT AssRSide()
  {
    partP = P_RHS;
    BX = R_RHS;
    SetZeroField(R_RHS);
    if (ApplyT3(AssB,all)==false) return false;
    SetBound(R_RHS);
    return true;
  }

static INT AssBDiag(t)
  TR3 *t;
  {
    INT i;
    PT **pts = &(t->p1);
 
    if ((actProblem->NumAss)(t)==false) return false;
    for (i = iFrom; i<iTo; i++)
       {
         RA(pts[i], R_RHS) += assB[i];

/*     necessary for the refinement strategy extraVol     */

/*        if (iFrom==4) (t->diag)[i-iFrom] = assA[i][i];   */

/*--------------------------------------------------------*/


   if ((pts[i]->boundP)==DIRICHLET)
      RA(pts[i], R_DIAG) = ONE;  
   else
      RA(pts[i], R_DIAG) += assA[i][i];
   }

   return true;
}
  
INT AssRSideDiag()
  {
    EDG *ed=actTriang->firstEdge;
    PT  *p =actTriang->firstPoint;
    InitNumAss(N_QQ);
    partP = P_ALL;
    SetZeroField(R_RHS);
    SetZeroField(R_DIAG);
    if (ApplyT3(AssBDiag,all)==false) return false;
    SetBound(R_RHS);

    return true;
  }
 
/* ************ "utilities" for simple vector operations *********** */

INT assign(a,b)
  INT a,b;
  {
    PT *p = actTriang->firstPoint;
	EDG *ed = actTriang->firstEdge;
	INT count = 0;

    if (iFrom==0)
      while (p!=nil) 
       { RA(p,a) = RA(p,b);
   /*      printf(" RA(p,a),p->boundP %e %d\n",RA(p,a),p->boundP);  */
         count++; 
         p = p->next; 
       }
    if (iTo==10)
      while (ed!=nil)
       {
        if ((ed->firstSon)==nil) { RA(ed,a) = RA(ed,b); count++; }
        ed = ed->next;
       }
    return count;
  }

void assneg(a,b)
  INT a, b;
  {
    PT  *p = actTriang->firstPoint;
    EDG *ed = actTriang->firstEdge;

    if (iFrom==0)
      while (p!=nil) 
        { RA(p,a) = -RA(p,b);
          p = p->next; 
        }
    if (iTo==10)
      while (ed!=nil)
	    {
	      if ((ed->firstSon)==nil) RA(ed,a) = -RA(ed,b);
		  ed = ed->next;
	    }
    return;
  }
 
void lin(a,b,c,x)
  INT a, b, c;
  REAL x;
  {
    PT *p = actTriang->firstPoint;
    EDG *ed = actTriang->firstEdge;

    if (iFrom==0)
      while (p!=nil) { RA(p,a) = RA(p,b)+x*RA(p,c); p = p->next; }
	if (iTo==10)
      while (ed!=nil)
	    {
	      if ((ed->firstSon)==nil) RA(ed,a) = RA(ed,b)+x*RA(ed,c);
		  ed = ed->next;
	    }
    return;
  }
 
void linneg(a,b,c,x)
  INT a, b, c;
  REAL x;
  {
    PT *p = actTriang->firstPoint;
	EDG *ed = actTriang->firstEdge;

    if (iFrom==0)
      while (p!=nil) { RA(p,a) = -RA(p,b)+x*RA(p,c); p = p->next; }
	if (iTo==10)
      while (ed!=nil)
	    {
	      if ((ed->firstSon)==nil) RA(ed,a) = -RA(ed,b)+x*RA(ed,c);
		  ed = ed->next;
	    }
    return;
  }

void add(a,b,c)
  INT a, b, c;
  {
    PT  *p = actTriang->firstPoint;
    EDG *ed = actTriang->firstEdge;

    if (iFrom==0)
      while (p!=nil) { RA(p,a) = RA(p,b)+RA(p,c); p = p->next; }
    if (iTo==10)
      while (ed!=nil)
        {
          if ((ed->firstSon)==nil) RA(ed,a) = RA(ed,b)+RA(ed,c);
          ed = ed->next;
        }
    return;
  }
 
void sub(a,b,c)
  INT a, b, c;
  {
    PT  *p = actTriang->firstPoint;
    EDG *ed = actTriang->firstEdge;

    if (iFrom==0)
      while (p!=nil)
        { RA(p,a) = RA(p,b)-RA(p,c);
          p = p->next;
        }

    if (iTo==10)
       while (ed!=nil)
		{
		  if ((ed->firstSon)==nil) RA(ed,a) = RA(ed,b)-RA(ed,c);
		  ed = ed->next;
		}
    return;
  }
 
REAL scalprod(a,b)
  INT a, b;
  {
    double sum = ZERO;
    PT *p = actTriang->firstPoint;
	EDG *ed = actTriang->firstEdge;

    if (iFrom==0)
      while (p!=nil) { sum += RA(p,a)*RA(p,b); p = p->next; }
    if (iTo==10)
      while (ed!=nil)
        {
          if ((ed->firstSon)==nil) sum += RA(ed,a)*RA(ed,b);
          ed = ed->next;
        }
   return sum;
  }

void StoreVecToField(x,index)
  REAL *x;
  INT index;
  {
    PT *p = actTriang->firstPoint;

    while (p!=nil) { RA(p,index) = x[p->index]; p = p->next; }
    return;
  }

void StoreFieldToVec(index,x)
  REAL *x;
  INT index;
  {
   PT *p = actTriang->firstPoint;

   while (p!=nil) { x[p->index] = RA(p,index); p = p->next; }
   return;
  }

void SetZeroField(index)
  INT index;
  {
    PT  *p  = actTriang->firstPoint;
    EDG *ed = actTriang->firstEdge;


    if (iFrom==0)
       while (p!=nil) 
         {  
           RA(p,index) = ZERO; 
           p = p->next;
         }

    if (iTo==10)
       while (ed!=nil)
         {
           if ((ed->firstSon)==nil) RA(ed,index) = ZERO;
           ed = ed->next;
         }

    return;
   }

REAL CompRes(ires, irq, idqq, irhs, tau)
  INT ires, irq, idqq, irhs;
  REAL *tau;
  {
    EDG *ed = actTriang->firstEdge;
    PT pm;
    REAL rvec[EDG_MAX];
    INT count = 0;
    double sum = ZERO, res, u1, u2, um;

    pm.boundP = DIRICHLET;
    pm.vec = &(rvec[0]);

    while (ed!=nil)
      {
       if ((ed->firstSon)!=nil) { ed = ed->next; continue; }

       RA(ed, irq) = RA(ed, irhs)-RA(ed, irq);

       if ((ed->boundP)==DIRICHLET)
         {
           u1 = RA(ed->p1,R_SOL);
           u2 = RA(ed->p2,R_SOL);
           StdMidPoint(ed,&pm);
           um = RD(pm,R_SOL);
           res = um-(u1+u2)*HALF;
           res = RA(ed,idqq)*res*res;
         }
       else
         {
           um = RA(ed,irq);
           res = um*um/RA(ed,idqq);

         }

       RA(ed,ires) = res;
       sum = sum+res; count++;
       ed = ed->next;
      }
   *tau = sum;
   return sum/count;
  }




REAL
CompResExtrapol(ires, irq, idqq, irhs, tau)
   INT             ires, irq, idqq, irhs;
   REAL           *tau;
{
   EDG            *ed = actTriang->firstEdge;
   PT              pm;
   REAL            rvec[EDG_MAX];
   REAL            sum = ZERO, res, fatherRes, u1, u2, um;
   REAL            newRes, compEta, resMax = ZERO;
   REAL            cutEta = ZERO;

   pm.boundP = DIRICHLET;
   pm.vec = &(rvec[0]);


   while (ed != nil)
     {
      if ((ed->firstSon) != nil) 
        {
          ed = ed->next;
          continue;
        }


      RA(ed, irq) = RA(ed, irhs) - RA(ed, irq);
      if ((ed->boundP) == DIRICHLET) 
        {
           res = ZERO;
        }
      else 
        {
           um = RA(ed, irq);
           res = um * um / RA(ed, idqq);
        }
      RA(ed, ires) = res;
      sum = sum + res;

      if ((ed->father) != nil)
        {
          fatherRes = RA((ed->father), ires);
          if (fatherRes == ZERO)
             newRes = ZERO;
          else
            {
              newRes = res * res / fatherRes;
            }
        } 
      else
        newRes = ZERO;

      if (newRes > res)    newRes = res;
      if (newRes > cutEta) cutEta = newRes;  

      ed = ed->next;
    }

    

    if ((actTriang->maxDepth) == 0)
       cutEta = sum / (actTriang->noOfEdges);


   *tau = sum;
   return cutEta;
}



void 
Lev0DirectSol(y)
	INT             y;
{
	vecX = (REAL *) (actSolve->rhs);
	BX = y;
	ApplyP(fetchElem, direct);
	actSolve->DirectSol((actSolve->env),
			    (actSolve->decomp),
			    (actSolve->diagonal),
			    (actSolve->rhs),
			    (actSolve->solution),
			    actSolve->dirDim);
	vecX = (REAL *) (actSolve->solution);
	ApplyP(storeElem, direct);

	return;
}


static INT 
ResCompare(i,j)
   EDG **i,**j;
{
   INT test;
   REAL diff = RA(*i,R_RES) < RA(*j,R_RES);
   if ( diff < ZERO ) 
      test = -1;
    else if ( diff > ZERO) test = 1;
    else test = 0 ;
   return test;
}


REAL
SortResExtrapol(ires, irq, idqq, irhs, tau)
   INT             ires, irq, idqq, irhs;
   REAL           *tau;
{
   EDG            *ed = actTriang->firstEdge, **eds;
   PT              pm;
   INT             noOfEdges = actTriang->noOfEdges, counter=0, i;
   INT             noOfPoints = actTriang->noOfPoints, minIndex;
   REAL            rvec[EDG_MAX];
   REAL            sum = ZERO, res, fatherRes, u1, u2, um, etaSum;
   REAL            iMin, iErr;
   REAL            newRes, compEta, resMax = ZERO;
   REAL            cutEta = ZERO,cutEtaII = ZERO;

   pm.boundP = DIRICHLET;
   pm.vec = &(rvec[0]);

   eds = (EDG**)ZIBAlloc( (long)noOfEdges*sizeof(EDG*) );

   while (ed != nil)
     {
      if ((ed->firstSon) != nil) 
        {
          ed = ed->next;
          continue;
        }

      eds[counter] =ed; counter++;

      RA(ed, irq) = RA(ed, irhs) - RA(ed, irq);
      if ((ed->boundP) == DIRICHLET) 
        {
           res = ZERO;
        }
      else 
        {
           um = RA(ed, irq);
           res = um * um / RA(ed, idqq);
        }
      RA(ed, ires) = res;
      sum = sum + res;
      ed = ed->next;
   }


    printf(" \n\n sort \n");
    qsort(eds,noOfEdges,sizeof(EDG*),ResCompare);

/*
    for (i=0; i< noOfEdges; i++)
       printf(" i %d res %e x,y,z %e %e %e \n", i, RA(eds[i],R_RES),
                eds[i]->p1->x, eds[i]->p1->y, eds[i]->p1->z );
*/

   etaSum = sum;


   for (i=0; i< counter; i++)
     {
      ed = eds[i];
      res = RA(ed, ires);
      if ((ed->father) != nil)
        {
          fatherRes = RA((ed->father), ires);
          if (fatherRes == ZERO)
             newRes = ZERO;
          else
            {
              newRes = res * res / fatherRes;
            }
        } 
      else
        newRes = ZERO;

     etaSum = etaSum - res + 6.0*newRes; 

      
      if (i==0)
        {
		iErr = noOfPoints / fabs(log10(etaSum));
          iMin = iErr;
          cutEta = RA(ed,ires);
          minIndex = 0;
        }
      else
        {
          if (etaSum > ZERO)
            {
              iErr = (noOfPoints+ i) / fabs(log10(etaSum));
              if (iErr < iMin) 
                {
                  iMin = iErr; 
                  cutEta = RA(ed,ires);
                  minIndex = i;
                }
/*
      if (res > ZERO)
      printf("depth, res, newRes, iErr, etaSum %d, %e %e %e %e\n",
              ed->depth,res,newRes,iErr,etaSum); 
*/ 
            }
         }
    }

    printf(" counter %d \n",counter);
    printf(" min Index %d \n",minIndex);

    

    if ((actTriang->maxDepth) == 0)
       cutEta = sum / (actTriang->noOfEdges);

   ZIBFree((PTR)eds);

   *tau = sum;
   return cutEta;
}
