/*
C
C  _______________________________________________________________
C
C*   Licence
C    =======
C
C    You may use or modify this code for your own non commercial
C    purposes for an unlimited time. 
C    In any case you should not deliver this code without a special 
C    permission of ZIB.
C    In case you intend to use the code commercially, we oblige you
C    to sign an according licence agreement with ZIB.
C
C
C  _______________________________________________________________
C
*/

#include <stdio.h>
#include <math.h>

#include "kask.h"
#include "kasktri.h"
#include "kaskass.h"
#include "kasksol.h"

static INT precond, allIt, verboseP;
static REAL eps, delta0, delta1, tau, beta;
static PROC      AXMul;
static VOIDPROC  PCXMul;
 
INT CGFirst(iaxMul, ieps, res, iprecond, ipcxMul, iverboseP)
   REAL ieps, *res;
   PROC     iaxMul;
   VOIDPROC ipcxMul;
   INT iprecond, iverboseP;
{
   INT countU = -1;
   PT *p;
 
   eps = ieps;
   precond = iprecond;
   verboseP = iverboseP;
   AXMul = iaxMul;
   PCXMul = ipcxMul;

   volumeKnown = false;

   countU = assign(R_CGG,R_RHS);

   if (verboseP)
     {
       sprintf(globBuf, "Solve: start of CG, n=%d, eps=%9.3e(%9.3e)%s\n",
                        countU, ieps, sqrt(ieps),(actSolve->preCondName) );
       ZIBStdOut(globBuf);
     }


   if ((*AXMul)(R_SOL, R_CGD)==false) return false;

   sub(R_CGG, R_CGD, R_CGG);


   if (precond)
     (*PCXMul)(R_CGG, R_CGH);
   else 
     { assign(R_CGH,R_CGG); SetBoundZero(R_CGH); }

   assneg(R_CGD, R_CGH);

   delta0 = scalprod(R_CGG, R_CGH);
   *res = delta0;
   allIt = 1;
   return true;
  }
 
INT CGStep(maxit, res, eps)
  INT *maxit;
  REAL *res, eps;
  {
    INT it = -1;
 
    while ((it++)<*maxit)
      {
	volumeKnown = true;

        (*AXMul)(R_CGD, R_CGH);

        tau = delta0/scalprod(R_CGD, R_CGH);
        lin(R_SOL, R_SOL, R_CGD, tau);
        lin(R_CGG, R_CGG, R_CGH, tau);

        if (precond) (*PCXMul)(R_CGG, R_CGH);
		else { assign(R_CGH,R_CGG); SetBoundZero(R_CGH); }

        delta1 = scalprod(R_CGG, R_CGH);
        beta = delta1/delta0;
        delta0 = delta1;
        linneg(R_CGD, R_CGH, R_CGD, beta);

        allIt++;
        if (verboseP)
          {
            sprintf(globBuf, "Solve: CG step %d: delta=%9.3e(%9.3e)\n",
                             allIt, delta0, sqrt(delta0));
            ZIBStdOut(globBuf);
          }

        if (delta1<=eps) 	break;   
      }

    *maxit = allIt;
    *res = delta1;
    return true;
  }



INT assignI(a,column)
  INT a, column;
  {
    PT *p = actTriang->firstPoint;
	EDG *ed = actTriang->firstEdge;
	INT count = 0;
printf("ifrom,ito,column,count %d %d %d %d \n",iFrom,iTo,column,count);
    if (iFrom==0)
	  while (p!=nil) 
		{ 
		if(count == column)	
			RA(p,a) = ONE; else RA(p,a) = ZERO;
		count++; p = p->next;
		}
	if (iTo==6)
      while (ed!=nil)
	    {
	      if ((ed->firstSon)==nil) 
			{
			if(count == column)	
			RA(ed,a) = ONE; else RA(ed,a) = ZERO;
			count++;
			}
		  ed = ed->next;
	    }
    return count;
  }


