/*
C
C  _______________________________________________________________
C
C*   Licence
C    =======
C
C    You may use or modify this code for your own non commercial
C    purposes for an unlimited time. 
C    In any case you should not deliver this code without a special 
C    permission of ZIB.
C    In case you intend to use the code commercially, we oblige you
C    to sign an according licence agreement with ZIB.
C
C
C  _______________________________________________________________
C
*/

#include "stdio.h"
#include "math.h"
 
#include "kask.h"
#include "kasktri.h"
#include "kaskass.h"
#include "kasksol.h"



INT CholDecEnvelope( d, env, l, n)
  REAL  *d, **l;
  INT *env;
  INT   n;
  {
    INT       i, j, k, len, nextN, envMax;
    REAL      x, y, z;
    PT        *p = actTriang->firstPoint;
    NEIGHBOR  *ngb, nb;

    Set0MatSparse(l, n, true);
    Set0VecFull(d, n);
 
    while (p!= nil)
     {
      i = p->index;
      len = p->noOfNeighbors;
      ngb = p->neighbors;
      nextN = 0;
      nb = ngb[nextN];
      for (j = env[i]; j<i+1; j++)
        { 
          if (j == (nb.p)->index)
             {
              x = nb.stiff;
              nextN++;
              if (nextN < len)
                  nb = ngb[nextN];
             }
          else
              x = ZERO;

          if (i==j)
            {
              for (k = j-1; k>=env[i]; k--)
                {
                  y = l[i][k-env[i]];
                  z = (l[i][k-env[i]] = y*d[k]);
                  x -= y*z;
                }
              if (x==ZERO)
                {
                  return false;
                }
              d[i] = ONE/x;
            }
          else
            {
              envMax = (env[i]>env[j])?env[i]:env[j];
              for (k = j-1; k>=envMax; k--) 
                  {
                   x -= l[i][k-env[i]]*l[j][k-env[j]];
                  }
              l[i][j-env[i]]  = x;
            }
        }
      p = p->next;
     }

    for (k = 0; k<n; k++) l[k][k-env[k]] = ONE;

    return true;
  }
 

 
INT CholSolEnvelope(env, l, p, b, x, n)
  REAL **l, *p, *b, *x;
  INT *env;
  INT n;
  {
    INT i, k;
    REAL y;

 
    Set0VecFull(x, n);

    for (i = 0; i<n; i++)      /* Forward substitution */
      { 
        y = b[i];
        for (k = i-1; k>=env[i]; k--) 
           {
            y -= l[i][k-env[i]]*x[k];
           }
        x[i] = y;
        b[i] = y;
      }

       

    for (i = n-1; i>=0; i--)   /* Backward substitution */
       {
        b[i] = b[i]*p[i];
       }

    for (i = n-1; i>=0; i--)   
      {
        x[i] = b[i];
        for (k = env[i]; k<=i-1; k++) b[k] -= l[i][k-env[i]]*x[i];
      }

    return true;
  }
 
static INT SetCountZero(p)
    PT *p;
   {
     p->counter = 0;
     return true; 
   }
    
  

INT CholDecSparse(env, l, n)   
    INT       *env, n;
 /*   NEIGHBOR  **l;   */
    PTR  *l;   

  {
    INT       i, j, k, kCol, kStart;
    INT       noOfPoints = actTriang->noOfPoints;
    PT        *p, *pCol, *first = actTriang->firstPoint;
    NEIGHBOR  *np, *npCol, *nb;
    PTSELEM    *columns=nil, *colList;


    columns = (PTSELEM*)ZIBAlloc((long)(noOfPoints)*sizeof(PTSELEM));
    if (columns == nil)
       { printf(" no memory for columns \n"); return false;}

    columns[0].last = nil;
    columns[0].next = &columns[1];
    for (i=1;i<noOfPoints-1;i++)
       {
         columns[i].last = &columns[i-1];
         columns[i].next = &columns[i+1];
       }
    columns[noOfPoints-1].last = &columns[noOfPoints-2];
    columns[noOfPoints-1].next = nil;

    p = first;
    i=0;
    while (p!=nil)
        {
          columns[i].p = p;
          p =  p->next;
          i++;
        }


    ApplyP(SetCountZero,all);
    p = first;
    while (p!=nil)
     {
      np = p->neighbors;
      j=0;
      while ( np[j].p != p )
        j++;
      np[j].ilu = np[j].stiff;
      for (k = j+1; k < (p->noOfNeighbors); k++)
         {
           np[k].ilu = np[k].stiff;
         }

      if (j < (p->noOfNeighbors-1))
         p->counter = j+1; 

 
      colList = columns;
/*      pCol = first;   */
      pCol = colList->p;
      while (pCol != p)
        {
         npCol = pCol->neighbors;
         i   = pCol->counter;
         if ( (npCol[i].p) == (p) )
           {
             np[j].ilu -= (npCol[i].ilu) * (npCol[i].ilu);
             kStart = j+1;
             for (kCol=i+1; kCol < pCol->noOfNeighbors; kCol++)
                {
                 for (k = kStart; k < (p->noOfNeighbors); k++)
                    { 
                     if (npCol[kCol].p == np[k].p)
                       {
                        np[k].ilu -= npCol[i].ilu * npCol[kCol].ilu;
                        kStart = k + 1;
                        break;
                       }
                    }
                 }
              if (pCol->counter < (pCol->noOfNeighbors-1))   
               (pCol->counter)++;
               if (pCol->counter == (pCol->noOfNeighbors))   
                 { if (colList->last != nil)
                     (colList->last)->next = colList->next;
                   (colList->next)->last = colList->last;
                 }

             }
          colList = colList->next;
/*          pCol = pCol->next;   */
          pCol = colList->p;   
        }


      np[j].ilu = sqrt(np[j].ilu);     
      for (k=j+1; k < (p->noOfNeighbors); k++)
           np[k].ilu /= np[j].ilu;

       p = p->next;
     }


    p = first;
    while (p != nil)
      {
        k = p->index;
        l[k] = (PTR)(p->neighbors);
        env[k] = p->noOfNeighbors;
        p->neighbors = nil;
        p->noOfNeighbors = 0;
        p = p->next;
      }
     
    if (columns != nil)
       ZIBFree((PTR)columns);

    return true;
  }
 

 
INT CholSolSparse(env, l, pp, b, x, n)
/*   NEIGHBOR    **l;  */
   PTR    *l;
   INT         *env;
   REAL        *pp, *b, *x;
   INT         n;
  {
    INT i, j, k, index;
    REAL y;
    PT        *p;
    PT *first = actTriang->firstPoint, *last = actTriang->lastDirectPoint;
    NEIGHBOR  *np;


    Set0VecFull(x, n);

    p = first;                /* Forward substitution */
    for (j=0; j < n; j++)
       {
         index = p->index;
         np = (NEIGHBOR*)l[index];
         i = 0;
         while (np[i].p != p)
            i++;

         x[index] = b[index]/np[i].ilu;
         for (k = i+1; k < env[index]; k++)   
           {
             b[np[k].p->index] -= np[k].ilu*x[index];
           }
          p = p->next;
       }

    p = last;                /* Backward substitution */
    while (p != nil)
      { 
        index = p->index;
        np = (NEIGHBOR*)l[index];
        i = 0;
        while (np[i].p != p)
            i++;

        y = x[index];
        for (k = i+1; k < env[index]; k++) 
           {
            y -= np[k].ilu*x[np[k].p->index];
           }
        x[index] = y/np[i].ilu;

        p = p->last;
      }

     
    return true;
  }



 
INT CholSolSparseOld(l, pp, b, x, n)
   REAL **l, *pp, *b, *x;
   INT n;
  {
    INT       i, j, k;
    REAL      y;
    PT        *p;
    PT        *first = actTriang->firstPoint, *last = actTriang->lastPoint;
    NEIGHBOR  *np;


    Set0VecFull(x, n);

    p = first;                /* Forward substitution */
    while (p != nil)
       {
         np = p->neighbors;
         i = 0;
         while (np[i].p != p)
            i++;

         x[p->index] = b[p->index]/np[i].ilu;
         for (k = i+1; k < p->noOfNeighbors; k++)   
           {
             b[np[k].p->index] -= np[k].ilu*x[p->index];
           }
          p = p->next;
       }

    p = last;                /* Backward substitution */
    while (p != nil)
      { 
        np = p->neighbors;
        i = 0;
        while (np[i].p != p)
            i++;

        y = x[p->index];
        for (k = i+1; k < p->noOfNeighbors; k++) 
           {
            y -= np[k].ilu*x[np[k].p->index];
           }
        x[p->index] = y/np[i].ilu;

        p = p->last;
      }

     
    return true;
  }
