/*
C
C  _______________________________________________________________
C
C*   Licence
C    =======
C
C    You may use or modify this code for your own non commercial
C    purposes for an unlimited time. 
C    In any case you should not deliver this code without a special 
C    permission of ZIB.
C    In case you intend to use the code commercially, we oblige you
C    to sign an according licence agreement with ZIB.
C
C
C  _______________________________________________________________
C
*/

#include <strings.h>
#include <stdio.h>
#include <math.h>
 
#include "kask.h"
#include "kasktri.h"
#include "kaskass.h"
#include "kasksol.h"
 
#define	PI		(3.14159265358979323846)
 
TRIANGULATION *actTriang = nil, *firstTriang = nil;
TR2 *tNextApply;

static TR2 greenT1, greenT2;
static EDG greenE;
static INT level;
static INT countP, countE, countT, countTd;

static void DeleteSons();

static DummyP() { return; }
 
INT InitTri(ptD,edgD,trD,tr3D)
  INT ptD, edgD, trD, tr3D;
  {
    if(!InitFixedLists(MAXLISTS)) return false;
	InitList(PTVEC, ptD*sizeof(REAL), 50);
	InitList(EDGVEC, edgD*sizeof(REAL), 50);
/*	InitList(TRVEC, trD*sizeof(REAL), 50);   */
/*	InitList(TR3VEC, tr3D*sizeof(REAL), 50); */
	InitList(PTELEM, sizeof(PT), 50);
	InitList(EDGELEM, sizeof(EDG), 50);
	InitList(TRELEM, sizeof(TR2), 50);
	InitList(TR3ELEM, 8*sizeof(TR3), 50);
/*	InitList(LOCALSTIFF, 9*sizeof(REAL), 50);   */
	InitList(BOUNDELEM, sizeof(BOUNDARY), 10);
	
	return true;
  }

 /* ************************************************************
 StdMidPoint   computes midpoint of a line edge
 -----------
 EDG *ed       adress of edge
 PT *p         if not nil, adress of point given by user
 return *PT    adress of new point, nil if failed
************************************************************ */
PT *StdMidPoint(ed,p)
  EDG *ed;
  PT  *p;
  {
   REAL dirVals[NO_OF_DIR_VALS];

    PT *P1 = ed->p1, *P2 = ed->p2;
 
    if (p==nil)
      {
        p = NewP();
        if (p==nil) return nil;
                                 /*  Set boundary point */
/*        if ((ed->boundP)!=false) p->boundP=ed->boundP;    */
        p->boundP=ed->boundP;

                                 /* index of the new point */
        p->index = (actTriang->noOfPoints)++;
        p->mark = false;
      }
    p->x = (P1->x)+((P2->x)-(P1->x))*HALF;
    p->y = (P1->y)+((P2->y)-(P1->y))*HALF;
    p->z = (P1->z)+((P2->z)-(P1->z))*HALF;
    if ((p->boundP)==DIRICHLET)
      {
        (actProblem->DirichF)(p->x,p->y,p->z,ed->class,dirVals);
        RA(p,R_SOL) = dirVals[0];
        if (actSolve->keepP) 
          RA(p,R_SOL_LAST) = RA(p,R_SOL);
      }
    else
      { 
        RA(p,R_SOL) = (RA(P1,R_SOL)+RA(P2,R_SOL))*0.5;
        if (actSolve->keepP) 
          RA(p,R_SOL_LAST) = (RA(P1,R_SOL_LAST)+RA(P2,R_SOL_LAST))*0.5;

      }


    return p;
  }


REAL Arcus(x,y)   /* utility for ArcMidPoint */
  REAL x,y;
  {
    INT quadrant;
    REAL phi,xx,yy;
 
    if (x>=ZERO)
      quadrant = (y>=0)?0:3;
    else
      quadrant = (y>=0)?1:2;
    xx = fabs((quadrant==0)||(quadrant==2)?x:y);
    yy = fabs((quadrant==0)||(quadrant==2)?y:x);
    phi = atan2(yy,xx)+quadrant*REALPI2;
    return phi;
  }
 
 /* ************************************************************
 ArcMidPoint   computes midpoint of a line edge
 -----------
 EDG *ed       adress of edge
 PT *p         if not nil, adress of point given by user
 return *PT    adress of new point, nil if failed
************************************************************ */
PT *ArcMidPoint(ed,p)
   EDG *ed;
   PT  *p;
{
   REAL dirVals[NO_OF_DIR_VALS];
   REAL xm, ym, zm, rm, rmx, rmy, rmz;
   REAL rx, ry, rz, r;
   PT *P1 = ed->p1, *P2 = ed->p2;
 
   if (p==nil)
     {
       p = NewP();
       if (p==nil) return nil;
       p->boundP=ed->boundP;

       p->index = (actTriang->noOfPoints)++;
       p->mark = false;
     }

   p->x = (P1->x)+((P2->x)-(P1->x))*0.5;
   p->y = (P1->y)+((P2->y)-(P1->y))*0.5;
   p->z = (P1->z)+((P2->z)-(P1->z))*0.5;
   if ((p->boundP)==DIRICHLET)
     {
       (actProblem->DirichF)(p->x,p->y,p->z,ed->class,dirVals);
       RA(p,R_SOL) = dirVals[0];
     }
   else
     { 
       RA(p,R_SOL) = (RA(P1,R_SOL)+RA(P2,R_SOL))*0.5;
     }


   xm  = (ed->bound)->x;
   ym  = (ed->bound)->y;
   zm  = (ed->bound)->z;
   rmx = (p->x)-xm;
   rmy = (p->y)-ym;
   rmz = (p->z)-zm;
   rm  = sqrt(rmx*rmx + rmy*rmy + rmz*rmz);

   rx = (P1->x)-xm;
   ry = (P1->y)-ym;
   rz = (P1->z)-zm;
   r  = sqrt(rx*rx + ry*ry + rz*rz);
   r  = r/rm;

   p->x = xm + r*rmx;
   p->y = ym + r*rmy;
   p->z = zm + r*rmz;

/*   printf(" new arc point %e %e %e\n",p->x,p->y,p->z);    */

   return p;
  }

 
/* ************************************************************
 NewP          allocates new point, presets most fields
 ---------
 return *PT    address of point, nil if failed
************************************************************ */
PT* NewP()
{
   PT *p = (PT*)GetElem(PTELEM);
 
   if (p==nil) return nil;
   p->last = actTriang->lastPoint;
   p->vec = (REAL*)GetElem(PTVEC);
   if ((p->vec)==nil) return nil;
   p->level = actTriang->refLevel;
   p->gP = false;

   if ((actTriang->lastPoint)!=nil) (actTriang->lastPoint)->next = p;
   actTriang->lastPoint = p;
 
   return p;
}
 
/* ************************************************************
 returnP       releases memory of a point
 ---------
 PT *p         address of the point
************************************************************ */
void ReturnP(p)
   PT *p;
{
   if ((p->vec) != nil) ReturnElem(PTVEC,(PTR)p->vec);
   if ((p->next)==nil) (actTriang->lastPoint) = p->last;
     else (p->next)->last = p->last;
   if ((p->last)==nil) (actTriang->firstPoint) = p->next;
     else (p->last)->next = p->next;
 
   ReturnElem(PTELEM, (PTR)p);
   return;
}
 
/* ************************************************************
 NewE          allocates new edge, presets most fields
 ---------
 return *EDG   address of edge, nil if failed
************************************************************ */
EDG* NewE()
{
   EDG *ed = (EDG*)GetElem(EDGELEM);
 
   if (ed==nil) return nil;
   ed->boundP = INTERIOR;
   ed->bound  = nil;
   ed->MidPoint = StdMidPoint;
   ed->level = actTriang->refLevel;
   ed->next = nil;
   ed->vec = (REAL*)GetElem(EDGVEC);
   if ((ed->vec)==nil) return nil;

 
   return ed;
}
 
/* ************************************************************
 returnE       releases memory of an edge
 ---------
 EDG *ed        address of the edge
************************************************************ */
void ReturnE(ed)
   EDG *ed;
{
   INT depth = ed->depth;

   if ((ed->vec) !=nil) ReturnElem(EDGVEC,(PTR)ed->vec);

   if ((actTriang->depthFirsts)[depth] == ed)
   if ((ed->next) == nil)
     (actTriang->depthFirsts)[depth] = nil;
    else if ((ed->next)->depth != depth)
           (actTriang->depthFirsts)[depth] = nil;
          else (actTriang->depthFirsts)[depth] = (ed->next);

   if ((ed->next)==nil) (actTriang->lastEdge) = ed->last;
     else (ed->next)->last = ed->last;
   if ((ed->last)==nil) (actTriang->firstEdge) = ed->next;
     else (ed->last)->next = ed->next;
 
       
   ReturnElem(EDGELEM, (PTR)ed);
   return;
}
 
/* ************************************************************
 NewT           allocates new triangle, presets most fields
 ---------
 return *TR2    address of triangle, nil if failed
************************************************************ */
TR2* NewT()
  {
    TR2 *t = (TR2*)GetElem(TRELEM), *tt;
	INT k;
 
    if (t==nil) return nil; 

    t->boundP = INTERIOR;
    t->bound  = nil;
    t->MidPoint = StdMidPoint;
    t->father = nil;
    t->firstSon = nil;
    t->type = T_WHITE;  
    t->level = actTriang->refLevel; 
/*   
    t->vec = (REAL*)GetElem(TRVEC);
    if ((t->vec)==nil) return nil;
*/

    t->next = nil;
    t->last = actTriang->lastTriangle;
    
    if ((actTriang->firstTriangle)==nil) (actTriang->firstTriangle) = t;
    if ((actTriang->lastTriangle)!=nil) (actTriang->lastTriangle)->next = t;
    actTriang->lastTriangle = t;
 
    return t;
  }
 
/* ************************************************************
 New8T3         allocates new tetrahedron, presets most fields
 ---------
 return *TR3    address of tetrahedron, nil if failed
************************************************************ */
TR3* New8T3()
  {
    TR3 *t8 = (TR3*)GetElem(TR3ELEM), *td;
	INT k;
 
	if (t8==nil) return nil; 
 	for(k = 0; k<8; k++)
	  {
		td = &(t8[k]);
		td->next = &(t8[k+1]);
		td->last = &(t8[k-1]);
		td->firstSon = nil; 
		td->father = nil; 
		td->level = actTriang->refLevel;
                td->gValsP = false;
/*		td->vec = (REAL*)GetElem(TR3VEC);   */
	  }
 	t8[0].last = nil;
	t8[7].next = nil;
 
    return t8;
  }



/* ************************************************************
 returnT       releases memory of a triangle
 ---------
 TR2 *p         address of the triangle
************************************************************ */
void Return4T(t)
  TR2 *t;
  {
	INT k;
    TR2 *tk;
 
/*
 	for(k = 0; k<4; k++)
	  {
		tk = &(t[k]);
		if ((tk->vec)  !=nil) ReturnElem(TRVEC,(PTR)tk->vec);
		if ((tk->user) !=nil) ReturnElem(LOCALSTIFF,(PTR)tk->user);
	  }  
    ReturnElem(TRELEM, (PTR)t);
*/
    return;
  }
 


/* ************************************************************
 returnT       releases memory of a triangle
 ---------
 TR2 *t         address of the triangle
************************************************************ */
void ReturnT(t)
   TR2 *t;
{
   INT k;
   TR2 *tk;
/*
   for (k = 0; k<4; k++)
      {
        tk = &(t[k]);
        if ((tk->vec)  !=nil) ReturnElem(TRVEC,(PTR)tk->vec);
        if ((tk->user) !=nil) ReturnElem(LOCALSTIFF,(PTR)tk->user);
      }  
*/
 
   if ((t->next)==nil) (actTriang->lastTriangle) = t->last;
   else (t->next)->last = t->last;
   if ((t->last)==nil) (actTriang->firstTriangle) = t->next;
   else (t->last)->next = t->next;
 
   ReturnElem(TRELEM, (PTR)t);
 
   return;
}


/* ************************************************************
 returnT3       releases memory of a tetrahedron
 ---------
 TR3 *p         address of the tetrahedron
************************************************************ */
void Return8T3(td)
   TR3 *td;
{
   INT k;
 
   ReturnElem(TR3ELEM, (PTR)td);
 
   return;
}



/* ***************************************************************
   CrTri              create Triangulation
   --------------
   char*  name        Name of the Triangulation
   Return
      TRIANGULATION*  Adress of the Description of a Triangulation
*************************************************************** */    
 
TRIANGULATION *CrTri(name)
  char *name;
  {
    TRIANGULATION *new;
    INT length, k;
 
    new = (TRIANGULATION*) ZIBAlloc ((long)sizeof(TRIANGULATION));
    if (new==nil)
	  { ZIBStdOut("Not enough memory (CrTri)\n"); return nil; }
 
    if (name!=nil)
      {
        length = strlen(name);
        new->name = (char*)ZIBAlloc((long)(length+1));
        if ((new->name)==nil)
		  { ZIBStdOut("Not enough memory (CrTri)\n"); return nil; }
        strcpy(new->name,name);
      }
    else new->name = nil;
    new->fileName = nil;

    new->noOfPoints = 0;
    new->noOfEdges = 0;
    new->noOfTriangles = 0;
    new->noOfTetrahedra = 0;
    new->refLevel = 0;
    new->maxDepth = 0;
    new->noOfInitPoints = 0;
    new->noOfInitEdges = 0;
    new->noOfInitTriangles = 0;
    new->noOfInitTetrahedra = 0;
    new->sparseAvailable = false;

    new->problem = nil;
    new->assemble = nil;
    new->matrix = nil;
    new->solve = nil;
    new->plot = nil;
    new->windows = nil;

    new->NewP = DummyP;
    new->NewE = DummyP;
    new->NewT = DummyP;
    new->ReturnP = DummyP;
    new->ReturnE = DummyP;
    new->ReturnT = DummyP;

    new->firstPoint = nil;
    new->lastPoint = nil;
    new->initPoints = nil;
    new->lastDirectPoint = nil;

    new->firstEdge = nil;
    new->lastEdge = nil;
    new->firstGreenEdge = nil;
    new->initEdges = nil;
    new->depthFirsts = (EDG**)ZIBAlloc(MAX_DEPTH*sizeof(EDG*));
    for (k=0; k<MAX_DEPTH; k++) (new->depthFirsts)[k] = nil;


    new->firstTriangle = nil;
    new->lastTriangle = nil;
    new->firstGreenTriangle = nil;
    new->initTriangles = nil;

    new->firstTetrahedron = nil;
    new->lastTetrahedron = nil;
    new->initTetrahedra = nil;

    new->next = firstTriang;

    actTriang = new;
    firstTriang = new;
    return new;
  }
 
/* **************************************************************
   SelTri                  Select Triangulation
   ---------------------
   TRIANGULATION* trigul   Triangulation descriptor adress
   Return INT              true or false
************************************************************** */    
INT SelTri(trgul)
  TRIANGULATION *trgul;
  {
    actTriang = trgul;
    return true;
  }
 

static void DeleteSons(t)
   TR3 *t;
{
   TR3 *ts = t->firstSon;
   INT k, lng;

   if (ts==nil) return;
   if ( (t->type)==T_RED ) lng = 8;
     else if ( (t->type==T_GREEN_1) || (t->type==T_GREEN_2A) ) lng = 4;
       else lng = 2;
   for (k = 0; k<lng; k++) DeleteSons(&(ts[k]));
   Return8T3(ts);
   countTd += lng;

   return;
}

/* ***************************************************************
   CloseTri                 Close Triangulation
   --------------
   TRIANGULATION** trigul   Adress of triangulation descriptor adress
   Return INT               true or false
*************************************************************** */    
INT CloseTri(trgul)
   TRIANGULATION **trgul;
{
   EDG *ed,*edNext;
   PT  *p,*pNext;
   TR2 *t, *tNext;
   TR3 *td;
   TRIANGULATION *triang;
   INT k;
 
    countP = 0; countE = 0; countT = 0; countTd = 0;
/*
   Return all tetrahedra, top-down from the initial triangulation
*/
   td = (*trgul)->initTetrahedra;
   for (k = 0; k<((*trgul)->noOfInitTetrahedra); k++)
   DeleteSons(&(td[k]));

/*
   Return all triangles
*/
   t = (*trgul)->firstTriangle;
   while (t!=nil)
     {
      tNext = t->next;
      ReturnT(t);
      countT++;
      t = tNext;
     }

/*
   Return all Edges
*/
   ed = (*trgul)->firstEdge;
   while (ed!=nil)
     {
      edNext = ed->next;
      ReturnE(ed);
      countE++;
      ed = edNext;
     }

/*
   Return all Points
*/
   p = (*trgul)->firstPoint;
   while (p!=nil)
     {
      pNext  =p->next;
      ReturnP(p);
      countP++;
      p = pNext;
     }

/*
   Return List Memory
*/
   ZIBFree((PTR)((*trgul)->initTetrahedra));
   countTd += (*trgul)->noOfInitTetrahedra;

   sprintf(globBuf,"Tri: %4d points     released\n",countP);
   ZIBStdOut(globBuf);

   sprintf(globBuf,"Tri: %4d edges      released\n",countE);
   ZIBStdOut(globBuf);

   sprintf(globBuf,"Tri: %4d triangles  released\n",countT);
   ZIBStdOut(globBuf);

   sprintf(globBuf,"Tri: %4d tetrahedra released\n",countTd);
   ZIBStdOut(globBuf);

   if ((*trgul)==firstTriang)
       firstTriang = (*trgul)->next;
     else
        {
          triang = firstTriang;
          while ((triang->next)!=(*trgul)) triang = triang->next;
          triang->next = (*trgul)->next;
        }
   actTriang = firstTriang;
   ZIBFree((PTR)(*trgul));
   return true;
}

static void FreeSons(t)
   TR3 *t;
{
   TR3 *ts = t->firstSon;
   INT k, lng;

   if (ts==nil) 
     {
      return;
     }
   if ( (t->type)==T_RED ) lng = 8;
     else if ( (t->type==T_GREEN_1) || (t->type==T_GREEN_2A) ) lng = 4;
            else lng = 2;
   for (k = 0; k<lng; k++) FreeSons(&(ts[k]));

   if ((ts[0].last)==nil)
      actTriang->firstTetrahedron = t;
     else    
      (ts[0].last)->next = t;
   if ((ts[lng-1].next)==nil) 
      actTriang->lastTetrahedron   = t;
     else    
      (ts[lng-1].next)->last = t;
   t->last = ts[0].last;
   t->next = ts[lng-1].next;

   Return8T3(ts);

   t->firstSon = nil;
   countTd += lng;

   return;
}
 
static INT CountP(p)
   PT * p;
{
   countP++;
   return true;
}

static INT CountE(ed)
   EDG * ed;
{
   countE++;
   return true;
}

static INT CountT(t)
   TR2 *t;
{
   countT++;
   return true;
}

static INT CountTd(td)
   TR3 * td;
{
   countTd++;
   return true;
}


INT ResetTriang()
{   
   EDG *ed,*edNext;
   PT  *p,*pNext;
   TR2 *t, *tNext;
   TR3 *td;
   INT k;
   TRIANGULATION *triang;

 
    countP = 0; countE = 0; countT = 0; countTd = 0;
/*
   Return all tetrahedra, top-down from the initial triangulation
*/
   td = actTriang->initTetrahedra;
   for (k = 0; k<(actTriang->noOfInitTetrahedra); k++)
      {
        FreeSons(&(td[k]));
        td[k].type = T_WHITE;
        (td[k].t1)->t31 = nil; (td[k].t1)->t32 = nil;
        (td[k].t2)->t31 = nil; (td[k].t2)->t32 = nil;
        (td[k].t3)->t31 = nil; (td[k].t3)->t32 = nil;
        (td[k].t4)->t31 = nil; (td[k].t4)->t32 = nil;

      }
   for (k = 0; k<(actTriang->noOfInitTetrahedra); k++)
      {
        SetT3F(&(td[k]));
      }

/*
   Return all triangles
*/
   t = actTriang->firstTriangle;
   while (t!=nil)
     {
      tNext = t->next;
      if (t->level >0)
        {
          ReturnT(t);
          countT++;
        }
      else
        {
          t->firstSon = nil;
        }
      t = tNext;
     }

/*
   Return all Edges
*/
   ed = actTriang->firstEdge;
   while (ed!=nil)
     {
      edNext = ed->next;
      if (ed->level >0) 
        {
          ReturnE(ed);
          countE++;
        }
      else
        {
          ed->firstSon = nil;
          ed->pm = nil;
        }
      ed = edNext;
     }

/*   Return all Points   */
   p = actTriang->firstPoint;
   while (p!=nil)
     {
      pNext  =p->next;
      if (p->level >0) 
        {
          ReturnP(p);
          countP++;
        }
      else
        {
         p->noOfNeighbors = 0;
         p->maxNeighbors = 0;
         p->neighbors = nil;
        }
      p = pNext;
     }


/*   Return List Memory   */

   sprintf(globBuf,"Tri: %4d points     released\n",countP);
   ZIBStdOut(globBuf);

   sprintf(globBuf,"Tri: %4d edges      released\n",countE);
   ZIBStdOut(globBuf);

   sprintf(globBuf,"Tri: %4d triangles  released\n",countT);
   ZIBStdOut(globBuf);

   sprintf(globBuf,"Tri: %4d tetrahedra released\n",countTd);
   ZIBStdOut(globBuf);

   countP = 0; countE = 0; countT = 0; countTd = 0;

   ApplyP(CountP,all);
   ApplyE(CountE,all);
   ApplyT(CountT,all);
   ApplyT3(CountTd,all);

   sprintf(globBuf,"Tri: %4d points    on level 0\n",countP);
   ZIBStdOut(globBuf);
   sprintf(globBuf,"Tri: %4d edges     on level 0\n",countE);
   ZIBStdOut(globBuf);   
   sprintf(globBuf,"Tri: %4d triangles on level 0\n",countT);
   ZIBStdOut(globBuf);
   sprintf(globBuf,"Tri: %4d tetras    on level 0\n",countTd);
   ZIBStdOut(globBuf);

   actTriang->noOfPoints     = countP;   
   actTriang->noOfEdges      = countE;   
   actTriang->noOfTriangles  = countT;   
   actTriang->noOfTetrahedra = countTd;   
   actTriang->refLevel       = 0;
   actTriang->maxDepth       = 0;


   return true;
}




/* ***************************************************************
   ApplyP          Apply a routine to a selection of points
   -----------
   PROC Routine    Routine to apply
   INT to          selection, see "kasktri.h"
   return INT      true or not true, if Routine returned not true
*************************************************************** */
INT ApplyP(Routine,to)
  PROC Routine;
  INT to;
  {
    PT *p,*pNext;
    INT rc, k;
 
    switch (to)
    {
      case boundHier:
      case boundNodal:
        p = actTriang->firstPoint;
        while (p!=nil)
          {
            pNext = p->next;
            if ((p->boundP)!=INTERIOR)
              {
                rc = (*Routine)(p);
                if (rc==false) return rc;
              }
            p = pNext;
          }
        break;
      case nullBound:
        p = actTriang->firstPoint;
        while (p!=nil)
          {
            pNext = p->next;
            if ((p->boundP)==DIRICHLET)
              {
                rc = (*Routine)(p);
                if (rc==false) return rc;
              }
            p = pNext;
          }
        p = actTriang->initPoints;
        for (k = 0; k<(actTriang->noOfInitPoints); k++) 
          {
            if ((p[k].boundP)==NEUMANN)
              {
				rc = (*Routine)(&(p[k]));
				if (rc==false) return rc;
              }
          }
        break;
      case dirichlet:
        p = actTriang->firstPoint;
        while (p!=nil)
          {
            pNext = p->next;
            if ((p->boundP)==DIRICHLET)
              {
                rc = (*Routine)(p);
                if (rc==false) return rc;
              }
            p = pNext;
          }
        break;
      case allNotDirichlet:
        p = actTriang->firstPoint;
        while (p!=nil)
          {
            pNext = p->next;
            if ((p->boundP)!=DIRICHLET)
              {
                rc = (*Routine)(p);
                if (rc==false) return rc;
              }
            p = pNext;
          }
        break;
      case all:
        p = actTriang->firstPoint;
        while (p!=nil)
          {
            pNext = p->next;
            rc = (*Routine)(p);
            if (rc==false) return rc;
            p = pNext;
          }
		break;
      case allBackward:
        p = actTriang->lastPoint;
        while (p!=nil)
          {
            pNext = p->last;
            rc = (*Routine)(p);
            if (rc==false) return rc;
            p = pNext;
          }
		break;
      case nonBound:
        p = actTriang->firstPoint;
        while (p!=nil)
          {
            pNext = p->next;
			if ((p->boundP)==INTERIOR)
			  {
                rc = (*Routine)(p);
                if (rc==false) return rc;
		      }
            p = pNext;
          }
        break;
      case initial:
        p = actTriang->initPoints;
        for (k = 0; k<(actTriang->noOfInitPoints); k++) 
          {
            rc = (*Routine)(&(p[k]));
            if (rc==false) return rc;
          }
        break;
      case initialNotDirichlet:
        p = actTriang->initPoints;
        for (k = 0; k<(actTriang->noOfInitPoints); k++) 
          {
	    if (p[k].boundP != DIRICHLET) 
              {
                rc = (*Routine)(&(p[k]));
                if (rc==false) return rc;
	      }
          }
        break;
      case boundInit:
        p = actTriang->initPoints;
        for (k = 0; k<(actTriang->noOfInitPoints); k++) 
          {
            if ((p[k].boundP)!=INTERIOR)
              {
				rc = (*Routine)(&(p[k]));
				if (rc==false) return rc;
              }
          }
        break;
      case direct:
	    if ((actTriang->lastDirectPoint)==nil) break;
        p = actTriang->firstPoint;
        while (p!=nil)
          {
            pNext = p->next;
            rc = (*Routine)(p);
            if (rc==false) return rc;
			if (p==(actTriang->lastDirectPoint)) break;
            p = pNext;
          }
		break;
      case nonDirect:
        p = actTriang->firstPoint;
		if ((actTriang->lastDirectPoint)!=nil)
		  p = (actTriang->lastDirectPoint)->next;
        while (p!=nil)
          {
            pNext = p->next;
            rc = (*Routine)(p);
            if (rc==false) return rc;
            p = pNext;
          }
		break;
    }
    return true;
  }

/* ***************************************************************
   ApplyE          Apply a routine to a selection of edges
   -----------
   PROC Routine    Routine to apply
   INT to          selection, see "kasktri.h"
   return INT      true or not true, if Routine returned not true
*************************************************************** */
INT ApplyE(Routine,to)
  PROC Routine;
  INT to;
  {
    EDG *ed,*edNext;
    INT rc, k;
 
   switch (to)
    {
      case boundHier:
        ed = actTriang->firstEdge;
        while (ed!=nil)
          {
            edNext = ed->next;
            if ((ed->boundP)!=INTERIOR)
              {
                rc = (*Routine)(ed);
                if (rc==false) return rc;
              }
            ed = edNext;
          }
        break;
      case boundNodal:
        ed = actTriang->firstEdge;
        while (ed!=nil)
          {
            edNext = ed->next;
            if (((ed->boundP)!=INTERIOR)&&((ed->firstSon)==nil))
              {
                rc = (*Routine)(ed);
                if (rc==false) return rc;
              }
            ed = edNext;
          }
        break;
      case dirichlet:
        ed = actTriang->firstEdge;
        while (ed!=nil)
          {
            edNext = ed->next;
            if (((ed->boundP)==DIRICHLET)&&((ed->firstSon)==nil))
              {
                rc = (*Routine)(ed);
                if (rc==false) return rc;
              }
            ed = edNext;
          }
        break;
      case boundInit:
        ed = actTriang->initEdges;
        for (k = 0; k<(actTriang->noOfInitEdges); k++) 
          {
            if ((ed[k].boundP)!=INTERIOR)
              {
				rc = (*Routine)(&(ed[k]));
				if (rc==false) return rc;
              }
          }
        break;
      case all:
        ed = actTriang->firstEdge;
        while (ed!=nil)
          {
            edNext = ed->next;
    	    rc = (*Routine)(ed);
    	    if (rc==false) return rc;
            ed = edNext;
          }
        break;
      case nodal:
        ed = actTriang->firstEdge;
        while (ed!=nil)
          {
            edNext = ed->next;
            if (((ed->firstSon)==nil)&&((ed->type)!=T_NOTHING))
              {
                rc = (*Routine)(ed);
                if (rc==false) return rc;
			  }
            ed = edNext;
          }
        break;
      case allBackward:
        ed = actTriang->lastEdge;
        while (ed!=nil)
          {
            edNext = ed->last;
            rc = (*Routine)(ed);
            if (rc==false) return rc;
            ed = edNext;
          }
        break;
      case allGreens:
        ed = actTriang->firstEdge;
        while (ed!=nil)
          {
            if ((ed->type)!=T_GREEN_1) { ed = ed->next; continue; }
            edNext = ed->next;
            rc = (*Routine)(ed);
            if (rc==false) return rc;
            ed = edNext;
          }
        break;
    }
    return true;
  }
 

/* ***************************************************************
   ApplyEDepth      Apply a routine to edges of a depth dep
   -----------
   PROC Routine    Routine to apply
   INT dep         depth of triangulation
   return INT      true or false
*************************************************************** */
INT ApplyEDepth(Routine, dep)
  PROC Routine;
  INT dep;
  {
	EDG  *ed;
	INT rc;

	ed = actTriang->depthFirsts[dep];   

        while (ed != nil)
	  {
	  if ((ed->depth)!=dep) break;   

		if ((ed->firstSon) != nil)
			  {
				rc = (*Routine)(ed);
				if (!rc) return false;
			  }
	   ed = ed->next;
	  }
  return true;
}





/* ***************************************************************
   ApplyT          Apply a routine to a selection of triangles
   -----------
   PROC Routine    Routine to apply
   INT to          selection, see "kasktri.h"
   return INT      true or not true, if Routine returned not true
*************************************************************** */
INT ApplyT(Routine,to)
   PROC Routine;
   INT to;
{
   TR2 *t, *tNext;
   INT rc, k;

   switch (to)
     {
       case all:
         t = actTriang->firstTriangle;
         while (t!=nil)
           {
             if ((t->firstSon)!=nil)
               {  t = t->next; continue; }
  
             tNextApply = t->next;
             rc = (*Routine)(t);
             if (rc==false) return rc;
             t = tNextApply;
           }
         break;

       case dirichlet:
           t = actTriang->firstTriangle;
           while (t!=nil)
             {
               if ((t->firstSon)!=nil)
                 {  t = t->next; continue; }
               if ((t->boundP)!=DIRICHLET)
                 {  t = t->next; continue; }

               tNextApply = t->next;
               rc = (*Routine)(t);
               if (rc==false) return rc;
               t = tNextApply;
             }
           break;

       case boundNodal:
           t = actTriang->firstTriangle;
           while (t!=nil)
             {
               if ((t->firstSon)!=nil)
                 {  t = t->next; continue; }
               if ((t->boundP)==INTERIOR)
                 {  t = t->next; continue; }

               tNextApply = t->next;
               rc = (*Routine)(t);
               if (rc==false) return rc;
               t = tNextApply;
             }
           break;

      case allReds:
        t = actTriang->firstTriangle;
        while (t!=nil)
          {
            if ((t->firstSon)!=nil)
              {  t = t->next; continue; }
  
            if (((t->type)==T_GREEN_1) || ((t->type)==T_GREEN_2A) ||
                ((t->type)==T_GREEN_3) )
               { t = t->next; continue; }
  
            tNextApply = t->next;
            rc = (*Routine)(t);
            if (rc==false) return rc;
            t = tNextApply;
          }
        break;

      case allGreens:
        t = actTriang->firstTriangle;
        while (t!=nil)
          {
            if ( ((t->type)!=T_GREEN_1) && ((t->type)!=T_GREEN_2A) &&
                 ((t->type)!=T_GREEN_3) )
               { t = t->next; continue; }
            tNext = t->next;
            rc = (*Routine)(t);
            if (rc==false) return rc;
            t = tNext;
          }
        break;
    }
    return true;
  }

 
/* ***************************************************************
   ApplyT3         Apply a routine to a selection of tetrahedra
   -----------
   PROC Routine    Routine to apply
   INT to          selection, see "kasktri.h"
   return INT      true or not true, if Routine returned not true
*************************************************************** */
INT ApplyT3(Routine,to)
  PROC Routine;
  INT to;
  {
    TR3 *t, *tNext;
    INT rc, k;

	switch (to)
    {
      case all:
        t = actTriang->firstTetrahedron;
        while (t!=nil)
          {
            tNext = t->next;
            rc = (*Routine)(t);
            if (rc==false) return rc;
            t = tNext;
          }
        break;
      case allOfClassTumor:
        t = actTriang->firstTetrahedron;
        while (t!=nil)
          {
            if ((t->class)!=TUMOR)
              {  t = t->next; continue; }

            tNext = t->next;
            rc = (*Routine)(t);
            if (rc==false) return rc;
            t = tNext;
          }
        break;
      case initial:
        t = actTriang->initTetrahedra;
        for (k = 0; k<(actTriang->noOfInitTetrahedra); k++) 
          {
            rc = (*Routine)(&(t[k]));
            if (rc==false) return rc;
          }

        break;
    }
    return true;
  }


/* ***************************************************************
   ApplyPLevel     Apply a routine to points of a level
   -----------
   PROC Routine    Routine to apply
   INT lev         level of triangulation
   return INT      true or false
*************************************************************** */
INT ApplyPLevel(Routine, lev)
  PROC Routine;
  INT lev;
  {
    PT *p = actTriang->firstPoint;
    INT rc;
 
    while (p!=nil)
      {
       if ((p->level)<=lev)
         {
          rc = (*Routine)(p);
          if (rc==false) return rc;
         }
       p = p->next;
      }
    return true;
  }


/* ***************************************************************
   ApplyELevel     Apply a routine to edges of a level
   -----------
   PROC Routine    Routine to apply
   INT lev         level of triangulation
   return INT      true or false
*************************************************************** */
INT ApplyELevel(Routine, lev)
  PROC Routine;
  INT lev;
  {
    EDG *ed = actTriang->firstEdge;
    INT rc;

    while (ed!=nil)
     {
      if ((ed->level)<=lev)
        {
         if ((ed->firstSon)==nil)
           {
             if ((ed->type)!=T_NOTHING)
               {
                 rc = (*Routine)(ed);
                 if (rc==false) return rc;
               }
            }
          else
            {
              if (((ed->firstSon)->level)>lev)
                {
                 if ((ed->type)!=T_NOTHING)
                   {
                    rc = (*Routine)(ed);
                    if (rc==false) return rc;
                   }
                }
            }
        }
      ed = ed->next;
     }
    return true;
  }


/* ***************************************************************
   ApplyTLevel      Apply a routine to triangles of a level
   -----------
   PROC Routine    Routine to apply
   INT lev         level of triangulation
   return INT      true or false
*************************************************************** */
static INT InLevel(t)
  TR2 *t;
  {
/*
	if ((t->level)>level) return false;
	if ((t->firstSon)==nil) return true;
	else return (((t->firstSon)->level)>level);
*/
  }

INT GreenRefinedP(t,lev)
  TR2 *t;
  INT lev;
  {
    EDG *ed = t->e1;

	if ((ed->firstSon)!=nil)
	  if (((ed->firstSon)->level)<=lev) return 0;
	ed = t->e2;
	if ((ed->firstSon)!=nil)
	  if (((ed->firstSon)->level)<=lev) return 1;
	ed = t->e3;
	if ((ed->firstSon)!=nil)
	  if (((ed->firstSon)->level)<=lev) return 2;
	return -1;
  }

static INT ApplyTLevRec(Routine, t)
  PROC Routine;
  TR2 *t;
  {
    TR2 *tt;
	INT rc, refEdgNo;
/*
	if (InLevel(t))
	  {
        refEdgNo = GreenRefinedP(t, level);
		if (refEdgNo>=0)
		  {
		    PT **pts = &(t->p1);
			EDG **eds = &(t->e1);
			INT i1 = refEdgNo, i2 = refEdgNo+1 , i3 = refEdgNo+2;

			if (i2>2) i2 -= 3;
			if (i3>2) i3 -= 3;

			greenE.p1 = PM(eds[i1]);
			greenE.p2 = pts[i1];
			greenE.t1 = &greenT1;
			greenE.t2 = &greenT2;
			greenE.level = level;

			greenT1.p1 = pts[i1]; 
			greenT1.p2 = pts[i2];
			greenT1.p3 = PM(eds[i1]);
			greenT1.e2 = &greenE;
			greenT1.e3 = eds[i3];
			greenT1.level = level;
			greenT1.depth = (t->depth)+1;

		    greenT2.p1 = pts[i1];
			greenT2.p2 = PM(eds[i1]);
			greenT2.p3 = pts[i3];
			greenT2.e2 = eds[i2];
			greenT2.e3 = &greenE;
			greenT2.level = level;
			greenT2.depth = (t->depth)+1;

			if (pts[i2]==(eds[i1]->p1))
			  {
			    greenT1.e1 = eds[i1]->firstSon;
				greenT2.e1 = (greenT1.e1)->next;
			  }
			else
			  {
				greenT2.e1 = eds[i1]->firstSon;
				greenT1.e1 = (greenT2.e1)->next;
			  }
			rc = (*Routine)(&greenT1);
			if (!rc) return rc;
			return (*Routine)(&greenT2);
		  }
		else return (*Routine)(t);
	  }
	tt = TRED1SON(t);
	if (!ApplyTLevRec(Routine, tt)) return false;
	tt = TRED2SON(t);
	if (!ApplyTLevRec(Routine, tt)) return false;
	if ((t->type)!=T_RED) return true;
	tt = TRED3SON(t);
	if (!ApplyTLevRec(Routine, tt)) return false;
	tt = TRED4SON(t);
*/
	return ApplyTLevRec(Routine, tt);
  }

INT ApplyTLevel(Routine, lev)
  PROC Routine;
  INT lev;
  {
    TR2 *t = actTriang->initTriangles;
    INT rc, k;

    level = lev;
	for (k = 0; k<(actTriang->noOfInitTriangles); k++) 
      {
        rc = ApplyTLevRec(Routine, &t[k]);
        if (!rc) return rc;
      }
    return true;
  }



/* ***************************************************************
   ApplyT3Level      Apply a routine to tetrahedra of a level
   -----------
   PROC Routine      Routine to apply
   INT lev           level of triangulation
   return INT        true or false
*************************************************************** */

static INT ApplyT3LevRec(Routine, t3)
  PROC Routine;
  TR3 *t3;
  {
    TR2 *tt;
	INT rc, refEdgNo;

	return true;
  }



INT ApplyT3Level(Routine, lev)
  PROC Routine;
  INT lev;
  {
    TR3 *t3 = actTriang->initTetrahedra;
    INT rc, k;

    level = lev;
	for (k = 0; k<(actTriang->noOfInitTetrahedra); k++) 
      {
        rc = ApplyT3LevRec(Routine, &t3[k]);
        if (!rc) return rc;
      }
    return true;
  }
