/*
C
C  _______________________________________________________________
C
C*   Licence
C    =======
C
C    You may use or modify this code for your own non commercial
C    purposes for an unlimited time. 
C    In any case you should not deliver this code without a special 
C    permission of ZIB.
C    In case you intend to use the code commercially, we oblige you
C    to sign an according licence agreement with ZIB.
C
C
C  _______________________________________________________________
C
*/

#include <stdio.h>
#include <math.h>

#include "kask.h"
#include "kasktri.h"
#include "kaskass.h"
#include "kaskcmd.h"
#include "kasksol.h"


#define L_SMALL 4
#define L_MEDIUM 7
#define L_LARGE 30
#define BLOCK_SMALL 100
#define BLOCK_MEDIUM 100
#define BLOCK_LARGE 100


int iluExists= 0;
static int failLevel= 10;
static int AX,AY;
static int testPrintP= true;
static long nCount,mCount;


int ILUPrepNeigh();
int ILUCompNeigh();
int ILUPrepare();
int ILUAssemble();
int ComputeILU();
int ILUaxMul();
int InvertILU();
int ILUIterate();
int CheckILU();

int InitNeighbors();
int StoreNeighborsIndex();
int StoreNeighborsDegree();

static int SetNoOne();
static int CountNeighbors();
static int AllocSparse();
static int SetSparseZero();
static int ILUAddElem();
static int ILUSetBC();
static int OneLineILU();
static int MulLine();
static int StepDown();
static int StepUp();
static void ILUTest();
static int SparseCheck();
static void ILUPrint();
static int ILUPrintElem();

int ILUPrepNeigh()
  { 
     if(!ApplyP(SetNoOne,all))return false;
     if(!ApplyE(CountNeighbors,nodal))return false;
     if(!ApplyP(AllocSparse,all))
       {ZIBStdOut("Generation of sparse matrix failed\n");return false;}

     return true;
  }

int ILUCompNeighIndex()
  { 
     if(!ApplyP(InitNeighbors,all))return false;
     if(!ApplyE(StoreNeighborsIndex,nodal))return false;
     iluExists= actTriang->noOfPoints;

     return true;
  }


int ILUCompNeighDegree()
  { 
     if(!ApplyP(InitNeighbors,all))return false;
     if(!ApplyE(StoreNeighborsDegree,nodal))return false;
     iluExists= actTriang->noOfPoints;

     return true;
  }


int ILUPrepare(order)
   int order;
{ 
  static int first= true;

  if ((actTriang->noOfPoints)!=iluExists)
    {
      if (!ILUPrepNeigh()) return false;
      switch (order)
        {
          case INDEX:
              if (!ILUCompNeighIndex()) return false;
               break;
          case DEGREE:
              if (!ILUCompNeighDegree()) return false;
              break;
         }

    }

  if(!ApplyP(SetSparseZero,all))return false;

return true;
}

int ILULowerStruct()
{
   PT        *first = actTriang->firstPoint, *p, *pm;
   NEIGHBOR  *np, *npm;
   int       i, j, k, l, n, foundIndex;


   p = first;
   while (p != nil)
     {  
       if (p->noOfNeighbors > 1)
         { 
           np = p->neighbors;
           i=0;
           while ( (np[i].p) != p )
              i++;
           if ( i == (p->noOfNeighbors - 1) ) 
               { p = p->next; continue; }
           pm = np[i+1].p;
           npm =pm->neighbors;
           k=0;
           while ( npm[k].p != pm )
              k++;
           l = k+1;

           for (j=i+2; j < (p->noOfNeighbors); j++)
              {
               foundIndex = false;
               while (!foundIndex)
                 {
                   if (l == pm->noOfNeighbors)
                     {
                       foundIndex = true;
                       pm->noOfNeighbors++;
                       AllocLSparse(pm);

                        /*  ranhaengen   */
                       npm = pm->neighbors;
                       npm[l].p = np[j].p;

                 /*      printf("an pm ran: %d %d \n", 
                               pm->indexP,np[j].p->indexP);
                       printf("noOfN,maxN %d %d \n",
                               pm->noOfNeighbors, pm->maxNeighbors);  */
                     }
                   else
                      if ( (npm[l].p->indexP) > (np[j].p->indexP) )
                        {
                          foundIndex = true;
                          pm->noOfNeighbors++;
                          AllocLSparse(pm);
                            /*   einfuegen   */
                          npm = pm->neighbors;
                          for (n=(pm->noOfNeighbors)-1;n>l;n--)
                               npm[n].p = npm[n-1].p;
                          npm[l].p = np[j].p;

                       /*   printf("bei pm rein: %d %d \n",
                                  pm->indexP,np[j].p->indexP);
                          printf("noOfN,maxN %d %d \n",
                                  pm->noOfNeighbors, pm->maxNeighbors);  */
                        }
                      else 
                        if ( (npm[l].p->indexP) == (np[j].p->indexP) )
                           {
                            /*   gleicher Index   */
                            foundIndex = true; 
                           }
                        else l++;
                  }
              }
         }
       p = p->next;
     }


return true;
}



static int SetNoOne(p)
PT*p;
{
p->noOfNeighbors= 1;
p->maxNeighbors= 0;
return true;
}


static int CountNeighbors(ed)
EDG*ed;
{
PT*p1= ed->p1,*p2= ed->p2;

(p1->noOfNeighbors)++;
(p2->noOfNeighbors)++;

return true;
}


static int AllocSparse(p)
PT *p;
{
   NEIGHBOR   *neigh= nil;
   int        no= p->noOfNeighbors;

   if( no <= (p->maxNeighbors)) return true;   

   if ( (p->neighbors) != nil )
     ZIBFree( (PTR)(p->neighbors) );

   neigh = (NEIGHBOR*)ZIBAlloc( (long)(no)*sizeof(NEIGHBOR) );

   if (neigh==nil) 
     {
      printf(" no memory in AllocSparse\n");
      return false;
     }
   p->maxNeighbors= no;
   p->neighbors= neigh;

return true;
}


static int AllocLSparse(p)
PT *p;
{
   NEIGHBOR   *neigh= nil;
   int        no= (p->noOfNeighbors), i;

   if (no <= (p->maxNeighbors)) return true;

   neigh = (NEIGHBOR*)ZIBAlloc( (long)(no+5)*sizeof(NEIGHBOR) );
   if (neigh==nil) {printf(" error in AllocL \n"); return false;}
   for (i=0; i<no-1;i++)
        neigh[i].p = (p->neighbors)[i].p;

   if ( (p->neighbors) != nil )
     ZIBFree( (PTR)(p->neighbors) );


   p->maxNeighbors= no+5;
   p->neighbors= neigh;


   return true;
}



int InitNeighbors(p)
PT*p;
{
	p->counterP = 1;
	(p->neighbors)[0].p = p;
	return true;
}


int StoreNeighborsIndex(ed)
EDG *ed;
{
  PT *p1= ed->p1, *p2= ed->p2;
  NEIGHBOR*nb;
  int i,k,lng;

  lng = p1->counterP;
  for (k= 0;k<lng;k++)
     {
      nb= &((p1->neighbors)[k]);
      if(((nb->p)->indexP)>(p2->indexP))break;
     }
  for (i= lng;i>k;i--)
     (p1->neighbors)[i].p= (p1->neighbors)[i-1].p;
  (p1->neighbors)[k].p= p2;

  (p1->counterP)++;

  lng = p2->counterP;
  for (k= 0;k<lng;k++)
     {
      nb= &((p2->neighbors)[k]);
      if(((nb->p)->indexP)>(p1->indexP))break;
     }
  for (i= lng;i>k;i--)
     (p2->neighbors)[i].p= (p2->neighbors)[i-1].p;
  (p2->neighbors)[k].p= p1;

  (p2->counterP)++;


return true;

}


int StoreNeighborsDegree(ed)
EDG*ed;
{
PT*p1= ed->p1,*p2= ed->p2;
NEIGHBOR*nb;
int i,k,lng;


	lng = p1->counterP;
	for (k = 0; k < lng; k++) {
		nb = &((p1->neighbors)[k]);
		if (((nb->p)->noOfNeighbors) > (p2->noOfNeighbors))
			break;	/* Ordne nach Degree nicht nach Index */
	}
	for (i = lng; i > k; i--)
		(p1->neighbors)[i].p = (p1->neighbors)[i - 1].p;
	(p1->neighbors)[k].p = p2;
	(p1->counterP)++;

	lng = p2->counterP;
	for (k = 0; k < lng; k++) {
		nb = &((p2->neighbors)[k]);
		if (((nb->p)->noOfNeighbors) > (p1->noOfNeighbors))
			break;	/* Ordne nach Degree nicht nach Index */
	}
	for (i = lng; i > k; i--)
		(p2->neighbors)[i].p = (p2->neighbors)[i - 1].p;
	(p2->neighbors)[k].p = p1;
	(p2->counterP)++;

  return true;
}


static int SetSparseZero(p)
  PT*p;
  {
   int k,lng= p->noOfNeighbors;
   NEIGHBOR*n;

   for (k= 0;k<lng;k++)
      {
        n= &((p->neighbors)[k]);
        n->stiff= ZERO;
        n->ilu= ZERO;
      }

   RA(p,R_DIAG)= ZERO;
   RA(p,R_RHS)= ZERO;
   return true;
}



int ILUAssemble()
{
  PT *p;
  NEIGHBOR *nb;
  int i;

  if (actTriang->iluAvailable != true)
  {
    InitNumAss(N_STD);
 
    partP= P_ALL;
    if(!ILUPrepare(INDEX)) return false;
    if(!ApplyT3(ILUAddElem,all)) return false;
   
    if ( !(actSolve->compNormP)) 
      { 
        if(!ApplyP(ILUSetBC,dirichlet))return false; 
      }

    actTriang->iluAvailable= true;
  }

return true;
}


static int ILUAddElem(t)
TR3 *t;
{
  PT        **pts= &(t->p1),*p;
  NEIGHBOR  *n;
  int       i,j,k,lng;

  if ((actProblem->NumAss)(t)==false) return false;
  for (i= 0;i<4;i++)
    {
      p= pts[i];
      RA(p,R_DIAG)+= assA[i][i];
      lng= p->noOfNeighbors;
      n= p->neighbors;
      for (j= 0;j<4;j++)
      for (k= 0;k<lng;k++)
        if (n[k].p==pts[j])
          n[k].stiff += 
            (symP==FULL)?assA[i][j]:(i>j)?assA[i][j]:assA[j][i];
      RA(p,R_RHS)+= assB[i];
    }
  return true;
}


static int ILUSetBCalt(p)
	PT *p;
{
	NEIGHBOR *n= p->neighbors;
	int      k,lng= p->noOfNeighbors;
	int      class = p->class;
	REAL     bxy, dirVals[NO_OF_DIR_VALS];
 
    (actProblem->DirichF)(p->x,p->y,p->z,class,dirVals);
    bxy= dirVals[0];

    for(k= 0;k<lng;k++)
      n[k].stiff= (n[k].p==p)?ONE:ZERO;
    RA(p,R_RHS)= bxy;
    RA(p,R_SOL)= bxy;
    RA(p,R_DIAG)= ONE;
   
    return true;
}


static int ILUSetBC(p)
  PT *p;
  {
   PT *np;
   NEIGHBOR *n = p->neighbors, *nn;
    int i = p->indexP, k, j, lng = p->noOfNeighbors;
    int lngn, l;
	int      class = p->class;
	REAL     bxy, dirVals[NO_OF_DIR_VALS];

	(actProblem->DirichF)(p->x,p->y,p->z,class,dirVals);
	bxy= dirVals[0];

        for (k = 0; k<lng; k++)  
         {
          n=&((p->neighbors)[k]);
          j=(n->p)->indexP;
          if (i!=j)
          {
          n->stiff = ZERO;
          np = n->p;
          if ((np->boundP)!=DIRICHLET)
            {
             lngn = np->noOfNeighbors;
             for (l=0; l<lngn; l++)  
                {
                  nn = &((np->neighbors)[l]);
                  if ((nn->p) == p)
                   { 
                     RA(n->p, R_RHS) -= bxy*(nn->stiff);
                     (nn->stiff) = ZERO;
                   }
                }
             }

          }
        if (i==j)  
          {
            RA(p, R_RHS) = bxy;
            RA(p, R_SOL) = bxy;

            n->stiff = ONE;
          }
        }
    return true;
  }


int ComputeILU()
{
return ApplyP(OneLineILU,allBackward);
}


static int OneLineILU(p)
PT*p;
{
NEIGHBOR*nbk,*nb1,*nb2;
PT*pk,*pj;
double sum,Lij,Rjk;
int r,s,i,k,min,j,lngi= p->noOfNeighbors,lngj;

i= p->indexP;

for(r= 0;r<lngi;r++)
{
nbk= &((p->neighbors)[r]);
pk= nbk->p;
k= pk->indexP;
min= (k<i)?k:i;

sum= ZERO;

for(j= 0;j<lngi;j++)
{
nb1= &((p->neighbors)[j]);
pj= nb1->p;
if((pj->indexP)>=min)break;

Lij= nb1->ilu;
lngj= pj->noOfNeighbors;

for(s= 0;s<lngj;s++)
{
nb2= &((pj->neighbors)[s]);
if(nb2->p==pk)
{
Rjk= nb2->ilu;
break;
}
}
sum= sum+Lij*Rjk;
}

if(k<i)nbk->ilu= ((nbk->stiff)-sum)/RA(pk,R_DIAG);
else nbk->ilu= (nbk->stiff)-sum;
}

return true;
}



int ILUaxMul(x,y)
int x,y;
{
int rc;

AX= x;AY= y;
rc= ApplyP(MulLine,all);

return rc;
}


int InvertILU(x,y)
int x,y;
{
AX= x;AY= y;

SetZeroField(x);
ApplyP(StepDown,all);
ApplyP(StepUp,allBackward);

return true;
}


static int MulLine(p)
  PT*p;
{
  int k,lng= p->noOfNeighbors;
  NEIGHBOR *n;
  double sum= ZERO;

  for (k= 0;k<lng;k++)
    {
      n= &((p->neighbors)[k]);
      sum+= ((p->neighbors)[k].stiff)*RA(n->p,AX);
    }
  RA(p,AY)= sum;

  return true;
}


static int StepDown(p)
PT*p;
{
NEIGHBOR*nb;
double sum= ZERO;
int m;

m= 0;
nb= &((p->neighbors)[m]);

while(nb->p!=p)
{
sum= sum+(nb->ilu)*RA(nb->p,AX);
m= m+1;
nb= &((p->neighbors)[m]);
}
RA(p,AX)= RA(p,AY)-sum;
return true;
}


static int StepUp(p)
PT*p;
{
NEIGHBOR*nb;
double sum= ZERO;
int m,lng= p->noOfNeighbors;

m= lng-1;
nb= &((p->neighbors)[m]);

while(nb->p!=p)
{
sum= sum+(nb->ilu)*RA(nb->p,AX);
m= m-1;
nb= &((p->neighbors)[m]);
}
RA(p,AX)= (RA(p,AX)-sum)/(nb->ilu);

return true;
}




int ILUCheck(cmd)
COMMAND*cmd;
{
if(ParsCheck(cmd,0,0))return false;

if (actTriang==nil)
{ ZIBStdOut("No triangulation\n"); return false; }
testPrintP= (actTriang->noOfPoints)<100;
if(!ILUPrepare(INDEX))return false;
ILUTest();
if(!ILUAssemble())return false;
if(!ComputeILU())return false;
ILUPrint();
return true;
}


static void ILUTest()
{
nCount= 0L;mCount= 0L;
ZIBStdOut("Checking sparse storage\n");
ApplyP(SparseCheck,all);
sprintf(globBuf,"%d elements, %4.1f%%filled\n",nCount,(100.0*nCount)/((REAL)mCount));
ZIBStdOut(globBuf);
return;
}


static int SparseCheck(p)
PT*p;
{
int k,lng= p->noOfNeighbors;
NEIGHBOR*n;

nCount+= lng;
mCount+= (p->maxNeighbors);
if(testPrintP)
{
sprintf(globBuf,"%d:(",p->indexP);ZIBStdOut(globBuf);
for(k= 0;k<lng;k++)
{
n= &((p->neighbors)[k]);
sprintf(globBuf,"%d",(n->p)->indexP);ZIBStdOut(globBuf);
if(k<(lng-1))ZIBStdOut(",");
}
ZIBStdOut(")\n");
}
return true;
}


static void ILUPrint()
{
if(testPrintP)ApplyP(ILUPrintElem,all);
return;
}


static int ILUPrintElem(p)
PT*p;
{
int k,lng= p->noOfNeighbors;
REAL x;

for(k= 0;k<lng;k++)
{
x= (p->neighbors)[k].stiff;
if(x==ZERO)continue;
sprintf(globBuf,"(%d,%d):%10.4e %10.4e\n",p->indexP,
(p->neighbors)[k].p->indexP,x,
(p->neighbors)[k].ilu);
ZIBStdOut(globBuf);
}
return true;
}





