#define R_SOL      0
#define R_RHS      1
#define R_DIAG     2
#define R_STIFF    2
#define R_CGD      3
#define R_CGG      4
#define R_CGH      5
#define R_SOL_LAST 6


#define R_RES      0
#define R_RQ       4


#define E_TPART    4
#define E_THOM     5
#define E_TOPT     6

#define R_MAXPT    7
#define R_MAXEDG   5
#define R_MAXTR    0
#define R_MAXTR3   0

#define R_GX       0
#define R_GY       1
#define R_GNORM    2



#define LISTLIST       0
#define PTVEC          1
#define EDGVEC         2
#define LOCALSTIFF     3
#define SPARSE_SMALL   4
#define SPARSE_MEDIUM  5
#define SPARSE_LARGE   6
#define LISTELEM       7
#define PTELEM         8
#define EDGELEM        9
#define TRELEM        10
#define BOUNDELEM     11
#define TR3ELEM       12
#define MAXLISTS      13
#define NOLIST        -1


struct PTSELEM
  {
    struct PT *p;
    struct PTSELEM *next, *last;
  };
typedef struct PTSELEM  PTSELEM;


struct _ELEM
  {
    PTR pt;
    struct _ELEM *next;
  };
typedef struct _ELEM  PTRELEM;

typedef PTRELEM*      PTRLIST;
typedef INT           LIST;

struct BOUNDARY
  {
    REAL x,y,z;
    struct BOUNDARY *next;
  };
typedef struct BOUNDARY BOUNDARY;

struct NEIGHBOR
  {
    struct PT	*p;
    REAL        stiff, ilu;
  };
typedef struct NEIGHBOR NEIGHBOR;
  
struct PT
  {
    REAL             *vec;
    REAL             x, y, z;
    REAL             g;
    INT              boundP, index, counter, level, mark, class, l;
    INT              gP;
    INT              noOfNeighbors, maxNeighbors;
    struct NEIGHBOR  *neighbors;
    struct PT        *next,*last;
  };
typedef struct PT  PT;


struct EDG
  {
    REAL        *vec;
    INT         boundP, type, level, depth, class, mark;
    struct PT   *p1,*p2,*pm;
    BOUNDARY    *bound;
    PT          *(*MidPoint)();
    struct EDG  *next,*last,*father,*firstSon;
  };
typedef struct EDG EDG;


struct TR2
  {
    struct PT   *p1,*p2,*p3;
    struct EDG  *e1,*e2,*e3;
    struct TR3  *t31,*t32;
    struct TR2  *next,*last,*father,*firstSon;
    BOUNDARY    *bound;
    PT          *(*MidPoint)();
    INT         boundP, type, class, level;
  };
typedef struct TR2 TR2;
 
struct TR3
  {
    struct PT   *p1,*p2,*p3,*p4;
    struct EDG  *e1,*e2,*e3,*e4,*e5,*e6;
    struct TR2  *t1,*t2,*t3,*t4;
    struct TR3  *next,*last,*father,*firstSon;
    INT         counter, type, level, class, depth, mark;
    INT         gValsP;
    REAL	volume, gVals[5];
  };
typedef struct TR3 TR3;



struct TRIANGULATION
  {
    char    *name;
    char    *fileName;
    INT     noOfPoints, noOfEdges, noOfTriangles, noOfTetrahedra,
            refLevel, maxDepth, sparseAvailable,
            noOfInitPoints, noOfInitEdges, noOfInitTriangles,
            noOfInitTetrahedra;
    PTR     problem, assemble, matrix, solve, plot, windows;
    PROC    NewP, NewE, NewT, ReturnP, ReturnE, ReturnT;
    PT      *firstPoint, *lastPoint, *initPoints, *lastDirectPoint;
    EDG     *firstEdge, *lastEdge, *firstGreenEdge, *initEdges, **depthFirsts;
    TR2     *firstTriangle, *lastTriangle, *firstGreenTriangle,
            *initTriangles;
    TR3     *firstTetrahedron, *lastTetrahedron, *initTetrahedra;
    PTRLIST *lists;
    struct TRIANGULATION *next;
  };
typedef struct TRIANGULATION TRIANGULATION;
 
#define allReds -1
#define allGreens -2
#define all -3
#define allBackward -4
#define initial -5
#define boundInit -6
#define boundHier -7
#define boundNodal -8
#define quadraticExtension -9
#define nodal -10
#define nonBound -11
#define previous -12
#define dirichlet -13
#define direct -14
#define nonDirect -15
#define nullBound -16
#define sonsOfRed -17
#define allNotDirichlet -18
#define initialNotDirichlet -19
#define allOfClassTumor -20
 
#define INTERIOR  1
#define DIRICHLET 2
#define NEUMANN   3
#define CAUCHY    4
#define T_WHITE 0
#define T_GREEN_1 1
#define T_RED 2
#define T_GREEN_3 3
#define T_NOTHING 4
#define T_GREEN_2A 5
#define T_GREEN_2B 6

#define EXACT    0
#define TRIANGUL 1
 
#define MAX_DEPTH 100
#define BONE   0
#define BRAIN  1
#define TUMOR  2

extern	INT		InitTri();
extern	TRIANGULATION	*CrTri();
extern	INT     SelTri();
extern	INT     CloseTri();

extern	INT     ApplyP();
extern	INT     ApplyE();
extern	INT     ApplyT();
extern	INT     ApplyPLevel();
extern	INT     ApplyELevel();
extern	INT     ApplyTLevel();
extern	INT     ApplyEDepth();

extern	INT     OpenRef();
extern	INT     RefTr(), RefTrHyper();
extern	INT     CloseRef();
extern	INT     CheckTd();
extern	INT     CheckAngle();

extern	PT      *StdMidPoint();
extern	PT      *ArcMidPoint();
extern	void	SetTP();
extern	void	SetTE();
extern	void	SetT3P();
extern	void	SetT3PBey();
extern	void	SetT3F();
extern	void	SetT3E();

extern	PT	*NewP();
extern	void 	ReturnP();
extern	EDG	*NewE();
extern	void 	ReturnE();
extern	TR2	*NewT();
extern	TR3	*New8T3();
extern	void 	Return4T();
extern	INT     MkGreen();
extern	EDG	*RefineEdg();
extern	INT    CompleteTd();
extern  INT     RefineTr3();
extern  INT     RefineBey();
extern  INT     ResetTriang();
extern  INT     CheckTri();

extern  REAL    alphaMax, alphaMin;

extern	void    PrintTR();

extern TRIANGULATION *actTriang, *firstTriang;

#define TFIRSTSON(t) ((t)->firstSon)
#define TRED1SON(t) ((t)->firstSon)
#define TRED2SON(t) (&(((t)->firstSon)[1]))
#define TRED3SON(t) (&(((t)->firstSon)[2]))
#define TRED4SON(t) (&(((t)->firstSon)[3]))
#define RA(p,ind) (((p)->vec)[ind])
#define RD(p,ind) (((p).vec)[ind])
#define GREENEDGE(t)((((t)->firstSon)->e3))
#define NEIGH_TR(ed,t) (((ed->t1)==t)?(ed->t2):(ed->t1))
#define NEIGHBOR1(t)((((((t)->e1)->t1)==(t))?((t)->e1)->t2:((t)->e1)->t1))
#define NEIGHBOR2(t)((((((t)->e2)->t1)==(t))?((t)->e2)->t2:((t)->e2)->t1))
#define NEIGHBOR3(t)((((((t)->e3)->t1)==(t))?((t)->e3)->t2:((t)->e3)->t1))
#define NEIGHBORK(t,k)((((((&((t)->e1))[k])->t1)==(t))?((&((t)->e1))[k])->t2:((&((t)->e1))[k])->t1))
#define EFIRSTSON(ed) ((ed)->firstSon)
#define ESECONDSON(ed) (((ed)->firstSon)->next)
#define ESONSP(ed) (((ed)->firstSon)!=nil)
#define PM(ed) ((ed)->pm)
#define MID(t) ((((t)->firstSon)->e3)->t2)
#define OPPO_EDG(t,ed) ((((t)->e1)==(ed))?(t)->p1:((((t)->e2)==(ed))?(t)->p2:(t)->p3))
#define OPPO_PT(t,p) ((((t)->p1)==(p))?(t)->e1:((((t)->p2)==(p))?(t)->e2:(t)->e3))
#define SON_AT(ed,p) ((((ed)->p1)==(p))?(ed)->firstSon:((ed)->firstSon)->next)
#define SON_NOT_AT(ed,p) ((((ed)->p1)!=(p))?(ed)->firstSon:((ed)->firstSon)->next)
#define OTHER_EDG(t,ed1,ed2) (((((t)->e1)!=(ed1))&&(((t)->e1)!=(ed2)))?(t)->e1:((((t)->e2)!=(ed1))&&(((t)->e2)!=(ed2)))?(t)->e2:(t)->e3)
#define OTHER_PT(t,p1,p2) (((((t)->p1)!=(p1))&&(((t)->p1)!=(p2)))?(t)->p1:((((t)->p2)!=(p1))&&(((t)->p2)!=(p2)))?(t)->p2:(t)->p3)
#define COMMON_PT(ed1,ed2) (((((ed1)->p1)==((ed2)->p1))||(((ed1)->p1)==((ed2)->p2)))?(ed1)->p1:(ed1)->p2)
#define PT_IN_EDG(p,ed) (((ed->p1)==p)||((ed->p2)==p))
#define PT_IN_TR(p,t) (((t->p1)==p)||((t->p2)==p)||((t->p3)==p))
#define EDG_IN_TR(ed,t) (((t->e1)==ed)||((t->e2)==ed)||((t->e3)==ed))
