/*
C
C  _______________________________________________________________
C
C*   Licence
C    =======
C
C    You may use or modify this code for your own non commercial
C    purposes for an unlimited time. 
C    In any case you should not deliver this code without a special 
C    permission of ZIB.
C    In case you intend to use the code commercially, we oblige you
C    to sign an according licence agreement with ZIB.
C
C
C  _______________________________________________________________
C
*/


#include <math.h>
 
#include "kask.h"
#include "kasktri.h"
#include "kaskass.h"
#include "kasksol.h"
  
INTEGDATA *stdIData=nil, *qStdIData=nil, *lStdIData=nil, *eStdIData=nil;

/*
    Integration points and weights for integration on tetrahedra
    Definition in shape.c
*/
REAL *iPointX, *iPointY, *iPointZ, *iWeight,
     **shape, **shapeX, **shapeY, **shapeZ,
     **shapeXX, **shapeXY, **shapeYY;

/*
    Integration points and weights for integration on faces of tetrahedra
    being used for the Cauchy boundary values
*/
/*
int  noOfAreaIPs = 3;
REAL areaIPsX[3] = { HALF, HALF, ZERO };
REAL areaIPsY[3] = { ZERO, HALF, HALF };
REAL areaIW[3]  = { SIXTH, SIXTH, SIXTH };
REAL areaI[10][3];
REAL xixy[3],sigmaxy[3];
*/
/*
int  noOfAreaIPs = 3;
REAL areaIPsX[3] = { SIXTH, TWOTHIRD, SIXTH };
REAL areaIPsY[3] = { SIXTH, SIXTH, TWOTHIRD };
REAL areaIW[3]  = { SIXTH, SIXTH, SIXTH };
REAL areaI[10][3];
REAL xixy[3],sigmaxy[3];
*/

int  noOfAreaIPs = 7;
REAL areaIPsX[7] = { 3.333333333333333e-01, 4.701420641051151e-01,
                     5.971587178976981e-02, 4.701420641051151e-01,
                     1.012865073234563e-01, 7.974269853530872e-01,
                     1.012865073234563e-01 };
REAL areaIPsY[7] = { 3.333333333333333e-01, 4.701420641051151e-01,
                     4.701420641051151e-01, 5.971587178976981e-02, 
                     1.012865073234563e-01, 1.012865073234563e-01,
                     7.974269853530872e-01 };
REAL areaIW[7]  = { 1.125000000000000e-01, 6.619707639425308e-02, 
                    6.619707639425308e-02, 6.619707639425308e-02,
                    6.296959027241358e-02, 6.296959027241358e-02, 
                    6.296959027241358e-02 };
REAL areaI[10][7];
REAL xixy[7],sigmaxy[7];



INT noOfIPoints, noOfShapeFunc,
    iFrom, iTo, kFrom, kTo,
    symP, curAss, partP;


REAL *assAdr[10];

VOIDPROC Shape;

static INT ellipMsg = false, qNegMsg = false;

INT InitNumAss(type)
  INT type;
  {
    INT i, j,k;
    REAL dum1,dum2,dum3,dum4,dum5,dum6, z;

    if (stdIData==nil)
      { 
       stdIData = NewIData(HYPERIP,4, StdShape, actProblem->symP);  
/*       stdIData = NewIData(USERIP,4, StdShape, actProblem->symP);  */
       if (stdIData==nil) return false;
      }
    if (lStdIData==nil)
      { 
        lStdIData = NewIData(HYPERIP, 10, StdShape, actProblem->symP);
        if (lStdIData==nil) return false;
      }
    if (qStdIData==nil)
	  { 
	    qStdIData = NewIData(HYPERIP, 10, StdQShape, actProblem->symP);
	    if (qStdIData==nil) return false;
	  }


    curAss = type;
    assA = assAG;
    assB = assBG;
    switch (type)
     {
      case N_STD:
        assA = assAdr;
        for(i=0; i<10; i++) assA[i] = assAL[i];
        noOfIPoints = stdIData->noOfIPoints;
        noOfShapeFunc = stdIData->noOfShapeFunc;
        Shape =  StdShape;
        iFrom = 0;
        iTo = stdIData->noOfShapeFunc;
        kFrom = 0;
        kTo = stdIData->noOfShapeFunc;
        symP = actProblem->symP;
        iPointX = stdIData->integPointX;
        iPointY = stdIData->integPointY;
        iPointZ = stdIData->integPointZ;
        iWeight = stdIData->integWeight;
        shape   = stdIData->shape;
        shapeX  = stdIData->shapeX;
        shapeY  = stdIData->shapeY;
        shapeZ  = stdIData->shapeZ;
        shapeXX = stdIData->shapeXX;
        shapeXY = stdIData->shapeXY;
        shapeYY = stdIData->shapeYY;
        break;

      case N_QQ:
        noOfIPoints = lStdIData->noOfIPoints;
        noOfShapeFunc = 10;
        Shape =  StdShape;
        iFrom = 4;
        iTo = 10;
        kFrom = 4;
        kTo = 10;
        symP = DIAGONAL;
        iPointX = lStdIData->integPointX;
        iPointY = lStdIData->integPointY;
        iPointZ = lStdIData->integPointZ;
        iWeight = lStdIData->integWeight;
        shape   = lStdIData->shape;
        shapeX  = lStdIData->shapeX;
        shapeY  = lStdIData->shapeY;
        shapeZ  = lStdIData->shapeZ;
        shapeXX = lStdIData->shapeXX;
        shapeXY = lStdIData->shapeXY;
        shapeYY = lStdIData->shapeYY;
        break;

        case N_LQ:
        case N_QL:
        noOfIPoints = lStdIData->noOfIPoints;
        noOfShapeFunc = 10;
        Shape =  StdShape;
        iFrom = 4;
        iTo = 10;
        kFrom = 0;
        kTo = 4;
        symP = FULL;
        iPointX = lStdIData->integPointX;
        iPointY = lStdIData->integPointY;
        iPointZ = lStdIData->integPointZ;
        iWeight = lStdIData->integWeight;
        shape   = lStdIData->shape;
        shapeX  = lStdIData->shapeX;
        shapeY  = lStdIData->shapeY;
        shapeZ  = lStdIData->shapeZ;
        shapeXX = lStdIData->shapeXX;
        shapeXY = lStdIData->shapeXY;
        shapeYY = lStdIData->shapeYY;
        break;

      default:
        ZIBStdOut("Unknown parameter for InitNumAss\n");
        return false;
     }

   for (i = 0; i<10; i++)
      {
         assB[i] = ZERO;
         for (k = 0; k<10; k++) assA[i][k] = ZERO;
      }



    /*
            fuer Cauchy-Rdbedg.:
            Tabellierung der Werte der phi[i] 
            in den Integrationspunkten des Standarddreiecks   
    */

   for (i=0;i<10;i++)
      {
            for (j=0;j<noOfAreaIPs;j++)
              { 
                z = ZERO;
		(*Shape)(areaIPsX[j],areaIPsY[j],z,i,&areaI[i][j],
			    &dum1,&dum2,&dum3,&dum4,&dum5,&dum6);   
              } 
      }


   return true;
}


REAL NumAssSigma(i,kk)
   INT i, kk;
{
   REAL hilfe = ZERO;
   INT m;

   for (m = 0; m < noOfAreaIPs; m++)
        hilfe += areaIW[m]*sigmaxy[m]*areaI[i][m]*areaI[kk][m];

   return hilfe;
}



REAL NumAssXi(i)
   INT i;
{
   REAL hilfe = ZERO;
   INT m;

   for (m = 0; m < noOfAreaIPs; m++)
       hilfe += areaIW[m]*xixy[m]*areaI[i][m];

   return hilfe;
}



INT NumAss(t)
   TR3 *t;
{
   PT   *p1 = t->p1,*p2 = t->p2,*p3 = t->p3,*p4 = t->p4;
   PT  **pts = &(t->p1), *pt1, *pt2, *pt3, *pt4;
   REAL fVals[NO_OF_F_VALS], cauVals[NO_OF_CAU_VALS];
   REAL f11, f12, f13, f21, f22, f23, f31, f32, f33;
   REAL e11, e12, e13, e21, e22, e23, e31, e32, e33;
   REAL hilfe,qHelp,XX,XY,XZ,YX,YY,YZ,ZX,ZY,ZZ;
   REAL sXX,sXY,sXZ,sYX,sYY,sYZ,sZX,sZY,sZZ;
   REAL xm, ym, zm, pX[16], pY[16], pZ[16], qxy[16], gxy[16];
   REAL d, dAbs, s;
   INT  i,k,kk,m;
   TR2  **ts;

   f11 = (p2->x)-(p1->x);
   f12 = (p3->x)-(p1->x);
   f13 = (p4->x)-(p1->x);
   f21 = (p2->y)-(p1->y);
   f22 = (p3->y)-(p1->y);
   f23 = (p4->y)-(p1->y);
   f31 = (p2->z)-(p1->z);
   f32 = (p3->z)-(p1->z);
   f33 = (p4->z)-(p1->z);

/*
    d =  f11*(f22*f33-f23*f32)
       - f21*(f12*f33-f32*f13)  
       + f31*(f12*f23-f13*f22);

    dAbs = fabs(d);
*/

   dAbs = t->volume*6.0;

   e11 = f22*f33 - f23*f32;
   e12 = f13*f32 - f12*f33;
   e13 = f12*f23 - f13*f22;
   e21 = f23*f31 - f21*f33;
   e22 = f11*f33 - f13*f31;
   e23 = f13*f21 - f11*f23;
   e31 = f21*f32 - f22*f31;
   e32 = f12*f31 - f11*f32;
   e33 = f11*f22 - f12*f21;

   for (m = 0; m<noOfIPoints; m++)
     {
      xm = f11*iPointX[m]+f12*iPointY[m]+f13*iPointZ[m]+(p1->x);
      ym = f21*iPointX[m]+f22*iPointY[m]+f23*iPointZ[m]+(p1->y);
      zm = f31*iPointX[m]+f32*iPointY[m]+f33*iPointZ[m]+(p1->z);

      (actProblem->F)(xm, ym, zm, t->class,fVals,partP);

      pX[m]  = fVals[0];
      pY[m]  = fVals[1];
      pZ[m]  = fVals[2];
      qxy[m] = fVals[3];
      gxy[m] = fVals[4];
     }

   for (i = iFrom; i<iTo; i++)
     {
       if ( (partP==P_STIFF)||(partP==P_ALL) )
       for (k = kFrom; k<kTo; k++)
        {
          if ((symP==SYMMETRIC)&&(k>i)) continue;
          if ((symP==DIAGONAL)&&(k!=i)) continue;
          assA[i][k] = ZERO;
          for (m = 0; m<noOfIPoints; m++)
           {

            sXX = shapeX[i][m]*shapeX[k][m];
            sXY = shapeX[i][m]*shapeY[k][m];
            sXZ = shapeX[i][m]*shapeZ[k][m];
            sYX = shapeY[i][m]*shapeX[k][m];
            sYY = shapeY[i][m]*shapeY[k][m];
            sYZ = shapeY[i][m]*shapeZ[k][m];
            sZX = shapeZ[i][m]*shapeX[k][m];
            sZY = shapeZ[i][m]*shapeY[k][m];
            sZZ = shapeZ[i][m]*shapeZ[k][m];

            hilfe = iWeight[m]*((1.0/dAbs)*
                 (pX[m]*
                    (e11*e11*sXX+e11*e21*sXY+e11*e31*sXZ   +
                     e21*e11*sYX+e21*e21*sYY+e21*e31*sYZ   +
                     e31*e11*sZX+e31*e21*sZY+e31*e31*sZZ)  +
                  pY[m]* 
                    (e12*e12*sXX+e12*e22*sXY+e12*e32*sXZ   +
                     e22*e12*sYX+e22*e22*sYY+e22*e32*sYZ   +
                     e32*e12*sZX+e32*e22*sZY+e32*e32*sZZ)  +
                  pZ[m]*  
                    (e13*e13*sXX+e13*e23*sXY+e13*e33*sXZ   +
                     e23*e13*sYX+e23*e23*sYY+e23*e33*sYZ   +
                     e33*e13*sZX+e33*e23*sZY+e33*e33*sZZ)) +

                  dAbs*qxy[m]*shape[i][m]*shape[k][m]
                );


           assA[i][k] += hilfe;
         }
     }

   if ( (partP==P_RHS)||(partP==P_ALL) )
     {
       assB[i] = ZERO;
       for (m = 0; m<noOfIPoints; m++)
          {
            hilfe = iWeight[m]*gxy[m]*shape[i][m];
            assB[i] += hilfe;
          }
       assB[i] = assB[i]*dAbs;
     }
   }

/*      Behandlung der Cauchy-Randbedingungen     */

   ts = &(t->t1);
   for (k=0;k<4;k++)
      { 
        if (ts[k]->boundP == CAUCHY)
          {
            pt1 = pts[(k+1)%4]; 
            pt2 = pts[(k+2)%4];
            pt3 = pts[(k+3)%4];
            pt4 = pts[k];

            s = AreaT(pt1,pt2,pt3);
            
            f11 = (pt2->x)-(pt1->x);
            f12 = (pt3->x)-(pt1->x);
            f21 = (pt2->y)-(pt1->y);
            f22 = (pt3->y)-(pt1->y);
            f31 = (pt2->z)-(pt1->z);
            f32 = (pt3->z)-(pt1->z);

            for (m=0; m<noOfAreaIPs; m++)
               {
     		 xm = f11*areaIPsX[m]+f12*areaIPsY[m]+(pt1->x);
     		 ym = f21*areaIPsX[m]+f22*areaIPsY[m]+(pt1->y);
     		 zm = f31*areaIPsX[m]+f32*areaIPsY[m]+(pt1->z);

                 (actProblem->CauchyF)(xm,ym,zm,ts[k]->class,cauVals,0);
                 sigmaxy[m] = cauVals[0];
                 xixy[m]    = cauVals[1];
               }


            if ( (partP==P_STIFF)||(partP==P_ALL) )
              {
                switch (curAss)
                  {
                    case N_STD:
                       switch (k)
                         {
                          case 0:
                            assA[1][1] += s*NumAssSigma(0,0);
                            assA[2][1] += s*NumAssSigma(1,0);
                            assA[2][2] += s*NumAssSigma(1,1);
                            assA[3][1] += s*NumAssSigma(2,0);
                            assA[3][2] += s*NumAssSigma(2,1);
                            assA[3][3] += s*NumAssSigma(2,2);
                            break;
                          case 1:
                            assA[0][0] += s*NumAssSigma(2,2);
                            assA[2][0] += s*NumAssSigma(0,2);
                            assA[2][2] += s*NumAssSigma(0,0);
                            assA[3][0] += s*NumAssSigma(1,2);
                            assA[3][2] += s*NumAssSigma(1,0);
                            assA[3][3] += s*NumAssSigma(1,1);
                            break;
                          case 2:
                            assA[0][0] += s*NumAssSigma(1,1);
                            assA[1][0] += s*NumAssSigma(2,1);
                            assA[1][1] += s*NumAssSigma(2,2);
                            assA[3][0] += s*NumAssSigma(0,1);
                            assA[3][1] += s*NumAssSigma(0,2);
                            assA[3][3] += s*NumAssSigma(0,0);
                            break;
                          case 3:
                            assA[0][0] += s*NumAssSigma(0,0);
                            assA[1][0] += s*NumAssSigma(1,0);
                            assA[1][1] += s*NumAssSigma(1,1);
                            assA[2][0] += s*NumAssSigma(2,0);
                            assA[2][1] += s*NumAssSigma(2,1);
                            assA[2][2] += s*NumAssSigma(2,2);
                            break;
                         }
                       break;
                     case N_LQ:
                       switch (k)
                         {
                          case 0:
                            assA[5][1] += s*NumAssSigma(6,0);
                            assA[5][2] += s*NumAssSigma(6,1);
                            assA[5][3] += s*NumAssSigma(6,2);
                            assA[7][1] += s*NumAssSigma(5,0);
                            assA[7][2] += s*NumAssSigma(5,1);
                            assA[7][3] += s*NumAssSigma(5,2);
                            assA[8][1] += s*NumAssSigma(4,0);
                            assA[8][2] += s*NumAssSigma(4,1);
                            assA[8][3] += s*NumAssSigma(4,2);
                            break;
                          case 1:
                            assA[4][0] += s*NumAssSigma(4,2);
                            assA[4][2] += s*NumAssSigma(4,0);
                            assA[4][3] += s*NumAssSigma(4,1);
                            assA[7][0] += s*NumAssSigma(6,2);
                            assA[7][2] += s*NumAssSigma(6,0);
                            assA[7][3] += s*NumAssSigma(6,1);
                            assA[9][0] += s*NumAssSigma(5,2);
                            assA[9][2] += s*NumAssSigma(5,0);
                            assA[9][3] += s*NumAssSigma(5,1);
                            break;
                          case 2:
                            assA[6][0] += s*NumAssSigma(5,1);
                            assA[6][1] += s*NumAssSigma(5,2);
                            assA[6][3] += s*NumAssSigma(5,0);
                            assA[8][0] += s*NumAssSigma(4,1);
                            assA[8][1] += s*NumAssSigma(4,2);
                            assA[8][3] += s*NumAssSigma(4,0);
                            assA[9][0] += s*NumAssSigma(6,1);
                            assA[9][1] += s*NumAssSigma(6,2);
                            assA[9][3] += s*NumAssSigma(6,0);
                            break;
                          case 3:
                            assA[4][0] += s*NumAssSigma(4,0);
                            assA[4][1] += s*NumAssSigma(4,1);
                            assA[4][2] += s*NumAssSigma(4,2);
                            assA[5][0] += s*NumAssSigma(5,0);
                            assA[5][1] += s*NumAssSigma(5,1);
                            assA[5][2] += s*NumAssSigma(5,2);
                            assA[6][0] += s*NumAssSigma(6,0);
                            assA[6][1] += s*NumAssSigma(6,1);
                            assA[6][2] += s*NumAssSigma(6,2);
                            break;
                         }
                       break;
                     case N_QQ:
                       switch (k)
                         {
                          case 0:
                            assA[5][5] += s*NumAssSigma(6,6);
                            assA[7][7] += s*NumAssSigma(5,5);
                            assA[8][8] += s*NumAssSigma(4,4);
                            break;
                          case 1:
                            assA[4][4] += s*NumAssSigma(4,4);
                            assA[7][7] += s*NumAssSigma(6,6);
                            assA[9][9] += s*NumAssSigma(5,5);
                            break;
                          case 2:
                            assA[6][6] += s*NumAssSigma(5,5);
                            assA[8][8] += s*NumAssSigma(4,4);
                            assA[9][9] += s*NumAssSigma(6,6);
                            break;
                          case 3:
                            assA[4][4] += s*NumAssSigma(4,4);
                            assA[5][5] += s*NumAssSigma(5,5);
                            assA[6][6] += s*NumAssSigma(6,6);
                            break;
                         }
                       break;
                   }
               }


            if ( (partP==P_RHS)||(partP==P_ALL) )
              {
                switch (curAss)
                  {
                    case N_STD:
                      for (i= iFrom; i< iTo; i++)
                         {
                           hilfe = s*NumAssXi(i);
                           assB[(k+1+i)%4]  += hilfe; 
                         }
                       break;
                     case N_LQ:
                     case N_QQ:
                       switch (k)
                         {
                          case 0:
                            assB[5] += s*NumAssXi(6);
                            assB[7] += s*NumAssXi(5);
                            assB[8] += s*NumAssXi(4);
                            break;
                          case 1:
                            assB[4] += s*NumAssXi(4);
                            assB[7] += s*NumAssXi(6);
                            assB[9] += s*NumAssXi(5);
                            break;
                          case 2:
                            assB[6] += s*NumAssXi(5);
                            assB[8] += s*NumAssXi(4);
                            assB[9] += s*NumAssXi(6);
                            break;
                          case 3:
                            assB[4] += s*NumAssXi(4);
                            assB[5] += s*NumAssXi(5);
                            assB[6] += s*NumAssXi(6);
                            break;
                         }
                       break;
                  }
              }
          }
     }

    return true;
}


