/*
C
C  _______________________________________________________________
C
C*   Licence
C    =======
C
C    You may use or modify this code for your own non commercial
C    purposes for an unlimited time. 
C    In any case you should not deliver this code without a special 
C    permission of ZIB.
C    In case you intend to use the code commercially, we oblige you
C    to sign an according licence agreement with ZIB.
C
C
C  _______________________________________________________________
C
*/


#include <stdio.h>
#include <fcntl.h>
#include <strings.h>
#include <math.h>
 
#include "kask.h"
#include "kaskcmd.h"
#include "kasktri.h"
#include "kaskgraph.h"

#define GETLINE() line=tp;lineNo++;while(((*tp)!='\n')&&((*tp)!=NUL))tp++;if ((*tp)!=NUL){*tp=NUL;tp++;}
 
#ifdef PROTO
static void CleanUp(char*);
#else
static void CleanUp();
#endif

static PT *PArray   = nil;
static EDG *EArray  = nil;
static TR2 *TArray  = nil;
static TR3 *T3Array = nil;
static BOUNDARY *BArray = nil;
static INT ptdim, edgdim, trdim, tr3dim, bounddim;
static TRIANGULATION *rdTriang = nil;
static char *text = nil;
static INT countP, countE, countT, countT3;
static EDG ***edsOfPts;
static TR2 ***trsOfPts;
static INT domain[20], CLASST3;
static REAL xLine[8], yLine[8];

static INT counter, counterBd, counter1, counter2, counter3;
static FILE *outFile = stdout, *outFileR = stdout, *inFile;

extern void prepare_rhs_();

 
static INT CountEd(ed)
  EDG *ed;
  {
    counter++;
	return true;
  }

static void CleanUp(fname)
  char *fname;
  {
    ZIBFree((PTR)text);
    if (countP!=ptdim)
      {
        ZIBStdOut("ReadTri: not enough points\n");
        if (PArray!=nil) ZIBFree((PTR)PArray);
        return;
      }
    if (countT3!=tr3dim)
      {
        ZIBStdOut("ReadTri: not enough tetrahedra\n");
        if (T3Array!=nil) ZIBFree((PTR)T3Array);
        return;
      }
    sprintf(globBuf, 
    "ReadTri: %d points, %d edges, %d triangles %d tetrahedra read from %s\n",
              countP, countE, countT, countT3,(fname==nil)?"input":fname);
    ZIBStdOut(globBuf);
    rdTriang->noOfInitPoints = ptdim;
    rdTriang->initPoints = PArray;
    rdTriang->noOfInitEdges = edgdim;
    rdTriang->initEdges = EArray;
    rdTriang->noOfInitTriangles = trdim;
    rdTriang->initTriangles = TArray;
    rdTriang->noOfInitTetrahedra = tr3dim;
    rdTriang->initTetrahedra = T3Array;

    return;
  }




static INT CompleteBd(td)
  TR3 *td;
  {
    INT i;
    EDG *ed, **eds;
    TR2 *t, **ts;
    TR3 *t31, *t32;
    
    ts = &(td->t1);
    for (i=0; i<4; i++)
       {
         t = ts[i];
         if (t->firstSon != nil)
           { printf("t in td \n");
             return true;
           }
         t31 = t->t31; t32 = t->t32;
         if ( (t31 == nil) || (t32 == nil) )
           {
             t->boundP = DIRICHLET;
             (t->p1)->boundP = DIRICHLET;
             (t->p2)->boundP = DIRICHLET;
             (t->p3)->boundP = DIRICHLET;
             (t->e1)->boundP = DIRICHLET;
             (t->e2)->boundP = DIRICHLET;
             (t->e3)->boundP = DIRICHLET;
           }
       }

    return true;
  }




static EDG* FindEdgByPts(p1,p2)
  PT  *p1, *p2;
  { 
    EDG *ed, *edNew;
    INT index1 = p1->index, index2 = p2->index;
    INT edgNo=0, depth=0;

    while (edsOfPts[index1][edgNo] != nil)
      {
        ed = edsOfPts[index1][edgNo];
        if( PT_IN_EDG(p1,ed) && PT_IN_EDG(p2,ed) )
          return ed;
        edgNo++;
      }
  
    edNew = NewE();
    edNew->p1 = p1; edNew->p2 = p2;
    edNew->depth = depth;
    if (((actTriang->depthFirsts)[depth])==nil)
      {
        (actTriang->depthFirsts)[depth] = edNew;
        edNew->last = actTriang->lastEdge;
        if ((actTriang->firstEdge)==nil) (actTriang->firstEdge) = edNew;
        if ((actTriang->lastEdge)!=nil) (actTriang->lastEdge)->next = edNew;
        actTriang->lastEdge = edNew;
      }
    else 
      if (((actTriang->depthFirsts)[depth+1])==nil)
        {
          (actTriang->lastEdge)->next = edNew;
          edNew->last = (actTriang->lastEdge);
          (actTriang->lastEdge) = edNew;
        }
      else 
        {
          (((actTriang->depthFirsts)[depth+1])->last)->next = edNew;
          edNew->last = ((actTriang->depthFirsts)[depth+1])->last;
          (actTriang->depthFirsts)[depth+1]->last = edNew;
          edNew->next = (actTriang->depthFirsts)[depth+1];
        }


    edsOfPts[index1][edgNo] = edNew;

    edgNo = 0;
    while (edsOfPts[index2][edgNo] != nil)
        edgNo++;
    edsOfPts[index2][edgNo] = edNew;

    (actTriang->noOfEdges)++;

    return edNew;
  }



static TR2* FindTr2ByEds(p1,p2,p3,e1,e2,e3)
  PT   *p1, *p2, *p3;
  EDG  *e1, *e2, *e3;
  { 
    TR2 *t, *tNew;
    INT index1=p1->index, index2=p2->index, index3=p3->index;
    INT trNo=0;

    while (trsOfPts[index1][trNo] != nil)
      {
        t = trsOfPts[index1][trNo];
        if( EDG_IN_TR(e1,t) && EDG_IN_TR(e2,t) && EDG_IN_TR(e3,t) ) 
          return t;
        trNo++;
      }
  
    tNew = NewT();
    tNew->e1 = e1; tNew->e2 = e2; tNew->e3 = e3;
    SetTP(tNew,(TR2*)nil); 
    trsOfPts[index1][trNo] = tNew;

    trNo = 0;
    while (trsOfPts[index2][trNo] != nil)
        trNo++;
    trsOfPts[index2][trNo] = tNew;

    trNo = 0;
    while (trsOfPts[index3][trNo] != nil)
        trNo++;
    trsOfPts[index3][trNo] = tNew;


    (actTriang->noOfTriangles)++;

    return tNew;
  }



/* ***************************************************************************
   CompleteTd(td)    set the edges and faces of a tetrahedron, computed from a 
                     computed from the points
   ----------
   TR3 *td           Adress of the tetrahedron
*************************************************************************** */
static INT CompleteTd(td)
  TR3 *td;
  { 
   PT  *p1 = td->p1, *p2 = td->p2, *p3 = td->p3, *p4 = td->p4;
   EDG *e1, *e2, *e3, *e4, *e5, *e6;

   td->e1 = FindEdgByPts(p1,p3);
   td->e2 = FindEdgByPts(p2,p3);
   td->e3 = FindEdgByPts(p1,p2);
   td->e4 = FindEdgByPts(p3,p4);
   td->e5 = FindEdgByPts(p2,p4);
   td->e6 = FindEdgByPts(p1,p4);
 
   e1 = td->e1; e2 = td->e2; e3 = td->e3;
   e4 = td->e4; e5 = td->e5; e6 = td->e6;

   td->t1 = FindTr2ByEds(p2,p3,p4,e2,e4,e5);
   td->t2 = FindTr2ByEds(p1,p3,p4,e1,e4,e6);
   td->t3 = FindTr2ByEds(p1,p2,p4,e3,e5,e6);
   td->t4 = FindTr2ByEds(p1,p2,p3,e1,e2,e3);

   SetT3F(td);

   return true;
  }

INT EdgHorMarkHyperF(ed)
   EDG *ed;
{
   PT **pts;
   INT i, countPt=0;
   REAL x, y, r;

   ed->mark = false;

   pts = &(ed->p1);
   for (i=0; i<2; i++)
      {
        if ( fabs( (pts[i]->z) - 0.5 ) > 1.e-5) return true;
        x = pts[i]->x;
        y = pts[i]->y;
        r = sqrt( x*x + y*y);

        if ( fabs(r - 0.3 ) < 1.e-5)
            countPt++;
      }

    if (countPt == 2) ed->mark = true;

    return true;
  }


INT EdgVerMarkHyperF(ed)
   EDG *ed;
{
   PT **pts;
   INT i, j, countPt=0;
   REAL x, y, r;

   ed->mark = false;

   pts = &(ed->p1);
   for (i=0; i<2; i++)
      {
        x = pts[i]->x;
        y = pts[i]->y;
        r = sqrt( x*x + y*y);

        if ( fabs( r - 0.3 ) > 1.e-5) return true;

        for (j=0; j<8; j++)
           {
              if ( fabs(x - xLine[j]) < 1.e-5)
                if ( fabs(y - yLine[j]) < 1.e-5)
                  {
                     countPt++; 
                     break;
                  }
           }

      }

    if (countPt == 2) ed->mark = true;

    return true;
  }




INT RdTri(cmd)
  COMMAND *cmd;
  {
    double x,y,z, fx,fy,fz;
    INT rc, i, k, k1, k2, k3, k4, k5, length, ii,
        l1,l2,l3,l4,
	    lineNo = 0, class, partner;
    INT no, beyP= false;
    char ch, *line, *tp, *line1, *name;
    PT *p, *p1, *p2, *p3;
    EDG *ed;
    EDG *e1,*e2,*e3,*e4,*e5,*e6;
    TR2 *t;
    TR3 *td;
    INT   dirPTdim, dirTR2dim, trNo, classT3, classT2, boundclass;
    REAL timeC;

    if (ParsCheck(cmd,1,2)) return false;

    if (!ZIBReadFile((cmd->pars)[1],&text))
      {
       sprintf(globBuf, "ReadTri: could not open file '%s'\n", (cmd->pars)[1]);
       ZIBStdOut(globBuf);
       return false;
      }
    tp = text;

    if ((cmd->noOfPars) == 2)
      {
       rc = 0;
       rc = sscanf((cmd->pars)[2],"%d",&no);
       if (rc!=1)
         {
           sprintf(globBuf, "ReadTri: wrong parameter '%s'\n", (cmd->pars)[1]);
           ZIBStdOut(globBuf);
           return false;
         }

       if (no == 1) beyP = true; 
         else beyP = false;
      }


    GETLINE();
    line1=  line;

    GETLINE();
    bounddim = 0;
    rc = sscanf(line,"Dimension:(%d,%d,%d,%d)",
                      &ptdim,&tr3dim,&dirTR2dim,&bounddim);
    if (rc==0)
      {
        sprintf(globBuf, "ReadTri: missing dimension, read: '%10.10s'...\n",
                          line);
        ZIBStdOut(globBuf);
        ZIBFree((PTR)text);
        return false;
      }



    rdTriang = CrTri(line1);

    PArray = (PT*) ZIBAlloc((long)(ptdim*(sizeof(PT))));
    if (PArray==nil) { CleanUp((cmd->pars)[1]); return false; }

    T3Array = (TR3*) ZIBAlloc((long)(tr3dim*(sizeof(TR3))));
    if (T3Array==nil) { CleanUp((cmd->pars)[1]); return false; }
    
    if (bounddim!=0)
      {
       BArray = (BOUNDARY*) ZIBAlloc((long)(bounddim*(sizeof(BOUNDARY))));
       if (BArray==nil) { CleanUp((cmd->pars)[1]); return false; }
      }

    countP = 0; countE = 0; countT = 0; countT3 = 0;

    while (true)
      {
        GETLINE();
        ch = ' ';
        rc = sscanf(line, "%d:(%le,%le,%le),%c,%le,%le,%le", &i, &x, &y, &z,
		  			  &ch,&fx,&fy,&fz);

        if (rc==0) break;

        p = &PArray[i];
        p->x = x;
        p->y = y;
        p->z = z;
        p->index = countP;
        p->level = 0;
        p->class = 0;
        p->mark  = false;
        p->gP    = false;

        if ((ch=='D')||(ch=='d')) p->boundP = DIRICHLET;
        else if ((ch=='I')||(ch=='i')) p->boundP = INTERIOR;
          else if ((ch=='N')||(ch=='n')) p->boundP = NEUMANN;
            else if ((ch=='C')||(ch=='c')) p->boundP = CAUCHY;
              else  p->boundP = INTERIOR;
        p->next = &PArray[i+1];
        p->last = &PArray[i-1];
        p->noOfNeighbors = 0;
        p->maxNeighbors  = 0;
        p->neighbors = nil;
        p->vec = (REAL*)GetElem(PTVEC);
        if ((rc==6))
          {
                RA(p, R_SOL) = fx;
           }
         else if (rc==8)
              {
                RA(p, R_SOL) = fx;
                RA(p, R_CGD) = fy;
                RA(p, R_CGG) = fz;
              }
            else RA(p, R_SOL) = ONE;
 
        RA(p,R_RHS) = ZERO;
        countP++;
      }

    if (strncmp(line,"END",3)!=0)
      { 
        sprintf(globBuf, "ReadTri: syntax error reading points (line %d)\n",
                          lineNo);
        ZIBStdOut(globBuf);
        CleanUp((cmd->pars)[1]);
        return false;
      }

    rdTriang->noOfPoints=countP;
    rdTriang->firstPoint = &PArray[0];
    PArray[0].last = nil;
    rdTriang->lastPoint = &PArray[ptdim-1];
    PArray[ptdim-1].next = nil;


    while (true)
      {
        GETLINE();
        rc = sscanf(line,"%d:(%d,%d,%d,%d),%d%s",
                          &i,&k1,&k2,&k3,&k4,&classT3,globBuf);
        if (rc==0) break;
        if ((k1>=ptdim)||(k2>=ptdim)||(k3>=ptdim)||(k4>=ptdim))
          {
            sprintf(globBuf,
                 "ReadTri: bound check error reading tetrahedron (line %d)\n",
                  lineNo);
            ZIBStdOut(globBuf);
            CleanUp((cmd->pars)[1]);
            return false;
          }
        countT3++;
        td = &T3Array[i];
 

        td->p1 = &PArray[k1];
        td->p2 = &PArray[k2];
        td->p3 = &PArray[k3];
        td->p4 = &PArray[k4];

        PArray[k1].noOfNeighbors +=3;
        PArray[k2].noOfNeighbors +=3;
        PArray[k3].noOfNeighbors +=3;
        PArray[k4].noOfNeighbors +=3;

        td->next = &T3Array[i+1];
        td->last = &T3Array[i-1];
        td->father = nil;
        td->firstSon = nil;
        td->mark = false;
        td->type = T_WHITE;
        td->level = 0;
        td->depth = 0;
        td->class = classT3;
        td->gValsP = false;

        if (!beyP)
           SetTPts(td,(TR3*)nil); 
        else   
           SetTPtsBey(td,(TR3*)nil); 

     }

    if (strncmp(line,"END",3)!=0)
      { 
        sprintf(globBuf, "ReadTri: syntax error reading tetrahedra (line %d)\n",
                          lineNo);
        ZIBStdOut(globBuf);
        CleanUp((cmd->pars)[1]);
        return false;
      }

    rdTriang->noOfTetrahedra=countT3;  
    rdTriang->firstTetrahedron = &T3Array[0];
    T3Array[0].last = nil;
    rdTriang->lastTetrahedron = &T3Array[tr3dim-1];   
     T3Array[tr3dim-1].next = nil;


    edsOfPts = (EDG***)ZIBAlloc( (long)ptdim*sizeof(EDG**) );
    for (k = 0;k<ptdim;k++)
      {
       edsOfPts[k]=
            (EDG**)ZIBAlloc( (long)(PArray[k].noOfNeighbors)*sizeof(EDG*) );
       if (edsOfPts[k] != nil)
         for (i=0; i<PArray[k].noOfNeighbors; i++) edsOfPts[k][i] = nil;
      }

    trsOfPts = (TR2***)ZIBAlloc( (long)ptdim*sizeof(TR2**) );
    for (k = 0;k<ptdim;k++)
      {
       trsOfPts[k]=
            (TR2**)ZIBAlloc( (long)(PArray[k].noOfNeighbors)*sizeof(TR2*) );
       if (trsOfPts[k] != nil)
         for (i=0; i<PArray[k].noOfNeighbors; i++) trsOfPts[k][i] = nil;
      }



   if (ApplyT3(CompleteTd,all)==false)
       {
        ZIBStdOut("CompleteTd failed \n");
        return false;
       }

   edgdim = actTriang->noOfEdges;
   trdim  = actTriang->noOfTriangles;


    if (dirTR2dim!=0)
      {
        for (k = 0;k<dirTR2dim;k++)
          {
            GETLINE();
            ch = ' ';
            classT2 = -1;
            rc = sscanf(line,"%d:(%d,%d,%d),%c,%d,%d,(%le,%le,%le)", 
                              &i,&k1,&k2,&k3,&ch,&classT2,
                              &boundclass,&x,&y,&z);
            if (rc==0) break;
            p1 = &(PArray[k1]);
            p2 = &(PArray[k2]);
            p3 = &(PArray[k3]);
            trNo = 0;
            while (trsOfPts[k1][trNo] != nil)
              {
                t = trsOfPts[k1][trNo];
                if( PT_IN_TR(p1,t) && PT_IN_TR(p2,t) && PT_IN_TR(p3,t) ) 
                break;
                trNo++;
              }

            if ( (ch == 'd') || (ch == 'D')) 
              {
                t->boundP = DIRICHLET;
                (t->e1)->boundP = DIRICHLET;
                (t->e2)->boundP = DIRICHLET;
                (t->e3)->boundP = DIRICHLET;
              }
            else if ( (ch == 'c') || (ch == 'C'))
                   {
                     t->boundP = CAUCHY;
                     if ( ((t->e1)->boundP) != DIRICHLET)
                        (t->e1)->boundP = CAUCHY;
                     if ( ((t->e2)->boundP) != DIRICHLET)
                        (t->e2)->boundP = CAUCHY;
                     if ( ((t->e3)->boundP) != DIRICHLET)
                        (t->e3)->boundP = CAUCHY;
                   }
                else if ( (ch == 'n') || (ch == 'N'))
                       {
                         t->boundP = NEUMANN;
                         if ( ((t->e1)->boundP) != DIRICHLET)
                            (t->e1)->boundP = NEUMANN;
                         if ( ((t->e2)->boundP) != DIRICHLET)
                            (t->e2)->boundP = NEUMANN;
                         if ( ((t->e3)->boundP) != DIRICHLET)
                            (t->e3)->boundP = NEUMANN;
                       }

             t->class = classT2;
             if ( rc>6  )
               { 
                 BArray[boundclass-1].x = x;
                 BArray[boundclass-1].y = y;
                 BArray[boundclass-1].z = z;
                 t->bound = &BArray[boundclass-1];
                 t->MidPoint = ArcMidPoint;
                 t->e1->bound = t->bound;
                 t->e2->bound = t->bound;
                 t->e3->bound = t->bound;
                 t->e1->MidPoint = ArcMidPoint;
                 t->e2->MidPoint = ArcMidPoint;
                 t->e3->MidPoint = ArcMidPoint;
               }
/*
   printf("i= %d : ch %c t->class %d boundP %d \n",i,ch,t->class,t->boundP);   
*/
           }
  

        GETLINE();
        if (strncmp(line,"END",3)!=0)
	      { 
            sprintf(globBuf, "ReadTri: syntax error reading bounds (line %d)\n",
                              lineNo);
            ZIBStdOut(globBuf);
            CleanUp((cmd->pars)[1]);
            return false;
          }
      }


   for (k = 0;k<ptdim;k++)
      {
        ZIBFree((PTR)edsOfPts[k]);
        ZIBFree((PTR)trsOfPts[k]);
      }
    ZIBFree((PTR)edsOfPts);
    ZIBFree((PTR)trsOfPts);


    if ((cmd->pars)[1]!=nil)
      {
        length  =strlen((cmd->pars)[1]);
        rdTriang->fileName = (char*)ZIBAlloc((long)(length+1));
        if ((rdTriang->fileName)!=nil)
          strcpy(rdTriang->fileName,(cmd->pars)[1]);
      }

    CleanUp((cmd->pars)[1]);

 /*     Check Triangulaton   */

     if(!CompAngles()) return false;

     if (ApplyT3(CheckTd,all)==false)
       {
        ZIBStdOut("CheckTri: failed making initial triangulation\n");
        return false;
       }
     else
       {
        ZIBStdOut("CheckTri: ok (RdTri)\n");
       }


    return true;
  }





/*
   *****   Write triangulation data, different interfaces   *******
*/


static INT CountPt(p)
  PT *p;
  {
    counter++;
    return true;
  }

static INT CountPtMark(p)
  PT *p;
  {
    if (p->l)
      {
        counter++;
        p->counter = counter;
      }
    return true;
  }


static INT CountPtMovie(p)
  PT *p;
  {
    counter++;
    if (p->boundP == DIRICHLET) counterBd++;
     return true;
  }

static INT CountEdg(ed)
  EDG *ed;
  {
    counter++;
	return true;
  }

static INT CountTr(t)
  TR2 *t;
  {
    counter++;
	return true;
  }

static INT CountTr3(td)
  TR3 *td;
{
   counter++;
   td->counter = counter;

   return true;
}

static INT CountTr3Movie(td)
   TR3 *td;
{
   counter++;
   td->counter = counter;
   domain[td->class]++;

   return true;
  }

static INT WritePt(p)
  PT *p;
  {
    p->counter = counter;

    fprintf(outFile, "%d:(%e,%e,%e),%c%d,%e\n",
            p->counter, p->x, p->y, p->z,
            ((p->boundP)==INTERIOR)?'I':
            (((p->boundP)==DIRICHLET)?'D':
            (((p->boundP)==NEUMANN)?'N':'C')),
            p->class,
            RA(p, R_SOL));
    counter++;
    return true;
  }





/*    
 ******     data for Grape interface     ******
*/

static INT WritePoint(p)
   PT *p;
{
   fprintf(outFile, "%e %e %e\n", p->x, p->y, p->z);

   return true;
}



static INT Write1ValAtPoint(p)
   PT *p;
{
   fprintf(outFile, "%e\n", RA(p,R_SOL));
   return true;
}


static INT Write3ValAtPoint(p)
   PT *p;
{
   fprintf(outFile, "%e %e %e\n", RA(p,R_SOL), RA(p,R_CGD), RA(p,R_CGG) );   
/*   fprintf(outFile, "%e %e %e\n", p->x, p->y, p->z );   */

   return true;
}


static INT WritePtsMovie(p)
   PT *p;
{
   fprintf(outFile,  "%+e %+e %+e   %+e\n", p->x, p->y, p->z, RA(p,R_SOL));
  
   fprintf(outFileR, "%+e %+e %+e      \n", p->x, p->y, p->z);  
   return true;
}

static INT WriteBdPtsMovie(p)
   PT *p;
{
   if (p->boundP == DIRICHLET)
      fprintf(outFileR, "%5d     \n", p->index+1);  
   return true;
}




static INT WriteTds(td)
   TR3 *td;
{
   fprintf(outFile, "%d %d %d %d %d\n",
                    (td->p1)->index+1, (td->p2)->index+1,
                    (td->p3)->index+1, (td->p4)->index+1, td->class );
 
   return true;
}




static INT WriteTdDomain(td)
  TR3 *td;
  {
    if (td->class == CLASST3)
      fprintf(outFileR, "%5d %5d %5d %5d\n",
            (td->p1)->index+1, (td->p2)->index+1,
            (td->p3)->index+1, (td->p4)->index+1 );
 
	return true;
  }


static INT CountBdOutTrs(t)
   TR2 *t;
{
   counter++;
   return true;

}


static INT WriteBdOutTrs(t)
   TR2 *t;
{
   fprintf(outFileR, "%5d %5d %5d\n",
            (t->p1)->index+1, (t->p2)->index+1,
            (t->p3)->index+1);
   return true;

}


static INT CountBdInTrs(t)
   TR2 *t;
{
   if ( (t->t31 == nil) || (t->t32 == nil) )     return true;
   if ( ((t->t31)->class) == ((t->t32)->class) ) return true;

   counter++;

   return true;
}


static INT WriteBdInTrs(t)
   TR2 *t;
{
   if ( (t->t31 == nil) || (t->t32 == nil) )     return true;
   if ( ((t->t31)->class) == ((t->t32)->class) ) return true;

   fprintf(outFileR, "%5d %5d %5d\n",
            (t->p1)->index+1, (t->p2)->index+1,
            (t->p3)->index+1);

   return true;
}




static INT WriteT3Neigh(td)
  TR3 *td;
{
   TR2 *t;
   INT i, k[4];

   t = td->t1;
   if ( (t->t31 == nil) || (t->t32 == nil) )
      k[0] = -1;
    else
      if ( (t->t31 == td) )  k[0] = (t->t32)->counter;
         else k[0] = (t->t31)->counter;

   t = td->t2;
   if ( (t->t31 == nil) || (t->t32 == nil) )
      k[1] = -1;
    else
      if ( (t->t31 == td) )  k[1] = (t->t32)->counter;
         else k[1] = (t->t31)->counter;

   t = td->t3;
   if ( (t->t31 == nil) || (t->t32 == nil) )
      k[2] = -1;
    else
      if ( (t->t31 == td) )  k[2] = (t->t32)->counter;
         else k[2] = (t->t31)->counter;

   t = td->t4;
   if ( (t->t31 == nil) || (t->t32 == nil) )
      k[3] = -1;
    else
      if ( (t->t31 == td) )  k[3] = (t->t32)->counter;
         else k[3] = (t->t31)->counter;
    fprintf(outFile, "%d %d %d %d\n",
            k[0], k[1], k[2], k[3]);
 
    return true;
}




INT Grape(cmd)
  COMMAND *cmd;
{
	INT noOfLines = 2, lev, dep, index, rc;
	INT noOfPts, noOfEdgs, noOfTrs, noOfT3rs, noOfValues=1;
	char *tp;
        REAL t;
        INT  noOfAntennae = 0, no, i;
        REAL antBegX[10], antBegY[10], antBegZ[10];
        REAL antEndX[10], antEndY[10], antEndZ[10];

	if (ParsCheck(cmd,0,3)) return false;

	if (actTriang==nil)
	  { ZIBStdOut("WriteTri: no triangulation to write\n"); 
            return false; 
          }
	lev = actTriang->refLevel;

	outFile = fopen((cmd->pars)[1], "w");
	if (outFile==nil)
	  {
	    sprintf(globBuf, "WriteTri: could not open file '%s'\n",
                (cmd->pars)[1]);
		ZIBStdOut(globBuf);
		return false;
	  }
        
    if ((cmd->noOfPars) >= 2)
      {
       rc = 0;
       rc = sscanf((cmd->pars)[2],"%d",&noOfValues);
       if (rc!=1)
         {
           sprintf(globBuf, "ReadTri: wrong parameter '%s'\n", (cmd->pars)[1]);
           ZIBStdOut(globBuf);
           return false;
         }

      }

    if ((cmd->noOfPars) == 3)
      {
       rc = 0;
       rc = sscanf((cmd->pars)[3],"%d",&noOfAntennae);
       if (rc!=1)
         {
           sprintf(globBuf, "ReadTri: wrong parameter '%s'\n", (cmd->pars)[1]);
           ZIBStdOut(globBuf);
           return false;
         }

      }


	index = 0;
	switch (index)
	  {
	  case -1:  
            sprintf(globBuf, "WriteTri: unknown parameter '%s'\n",
                  (cmd->pars)[2]);
            ZIBStdOut(globBuf);
            fclose(outFile);
            return false;
	  case  0:

	    counter = 0;
	    ApplyP(CountPt, all);
	    noOfPts = counter;
	    counter = 0;
	    ApplyT3(CountTr3, all);
	    noOfT3rs = counter;

	    fprintf(outFile, "%d\n", noOfT3rs);
	    fprintf(outFile, "%d\n", noOfPts);

	    ApplyP(WritePoint, all);

	    ApplyT3(WriteTds, all);

 	    ApplyT3(WriteT3Neigh, all);    


	    fprintf(outFile, "%d\n", noOfValues);
            
            if (noOfValues == 1)
	      ApplyP(Write1ValAtPoint, all);
            else if (noOfValues == 3)
                    ApplyP(Write3ValAtPoint, all);

            if (noOfAntennae > 0) 
             {
	       inFile = fopen("antennae", "r");

	       if (inFile==nil)
	         {
	          sprintf(globBuf, "WriteTri: could not open file antennae \n");
		  ZIBStdOut(globBuf);
		  return false;
	         }

               fprintf(outFile, "%d\n", noOfAntennae);
               for(i=0; i<noOfAntennae;i++) 
                {
                 fscanf(inFile,"%le %le %le\n",&antBegX[i], &antBegY[i], &antBegZ[i]);
                 fprintf(outFile, "%e %e %e\n",antBegX[i], antBegY[i], antBegZ[i]);
                }
               for(i=0; i<noOfAntennae;i++) 
                {
                 fscanf(inFile,"%le %le %le\n",&antEndX[i], &antEndY[i], &antEndZ[i]);
                 fprintf(outFile, "%e %e %e\n",antEndX[i], antEndY[i], antEndZ[i]);
                }

               fclose(inFile);
             }
              

	    fclose(outFile);

	    sprintf(globBuf, "GRAPE: Data  ok. \n");
		ZIBStdOut(globBuf);
	    break;

	  case  1:
	    break;
	  }
	return true;
  }




INT Movie(cmd)
   COMMAND *cmd;
{
   INT noOfLines = 2, lev, dep, index, rc;
   INT noOfPts, noOfBdPts, noOfEdgs, noOfTrs, noOfOutTrs, noOfInTrs, noOfT3rs;
   INT noOfTdBone, noOfTdBrain, noOfTdTumor;
   INT noOfDomains, i;
   char *tp;

   if (ParsCheck(cmd,0,1)) return false;

   if (actTriang==nil)
     { ZIBStdOut("WriteTri: no triangulation to write\n"); return false; }
   lev = actTriang->refLevel;

   outFile = fopen((cmd->pars)[1], "w");
   if (outFile==nil)
     {
       sprintf(globBuf, "WriteTri: could not open file '%s'\n",
               (cmd->pars)[1]);
       ZIBStdOut(globBuf);
       return false;
     }

   outFileR = fopen("GRID_R", "w");
   if (outFileR==nil)
     {
       sprintf(globBuf, "WriteTri: could not open file GRID_RETOUR \n");
       ZIBStdOut(globBuf);
       return false;
     }


   index = 0;
   switch (index)
     {
       case -1:  
         sprintf(globBuf, "WriteTri: unknown parameter '%s'\n",
                 (cmd->pars)[2]);
         ZIBStdOut(globBuf);
         fclose(outFile);
         return false;

       case  0:
         counter   = 0;
         counterBd = 0;
         ApplyP(CountPtMovie, all);
         noOfPts     = counter;
         noOfBdPts   = counterBd;
         counter     = 0;
         for (i=0; i<20; i++)
             domain[i] = 0;
         ApplyT3(CountTr3Movie, all);
         noOfT3rs = counter;

         noOfDomains = 0;
         for (i=0; i<20; i++)
            {
             if ( domain[i] != 0) noOfDomains++;
            }

         noOfInTrs  = 0;
         counter = 0;
         ApplyT(CountBdOutTrs, dirichlet);
         noOfOutTrs = counter;
 
         counter = 0;
         ApplyT(CountBdInTrs, all);
         noOfInTrs = counter;

         fprintf(outFile, "%d\n", noOfPts);
         fprintf(outFileR, "%5d %5d %5d %5d %5d %5d\n",
           noOfPts, noOfBdPts, noOfT3rs, noOfDomains, noOfOutTrs, noOfInTrs);

         for (i=0; i<noOfDomains; i++)
            {
             fprintf(outFileR, "%5d \n", domain[i]);
            }

         ApplyP(WritePtsMovie, all);
         ApplyP(WriteBdPtsMovie, all);

         for (i=0; i < noOfDomains; i++)
            {
              CLASST3 = i;
              ApplyT3(WriteTdDomain, all);
            }

         ApplyT(WriteBdOutTrs, dirichlet);
         ApplyT(WriteBdInTrs, all);

         fclose(outFile);
         fclose(outFileR);

         sprintf(globBuf, "Movie ok. \n");
         ZIBStdOut(globBuf);
         break;

       case  1:
          break;
     }

   return true;
}



/* ****************   Write Triangulation    ************************ */    

int WrTri(cmd)
  COMMAND *cmd;
  {
    int noOfLines = 2, lev, dep, index, rc;
    int noOfPts, noOfEdgs, noOfTrs, noOfTds;
    char *tp;

    if (ParsCheck(cmd,2,3)) return false;

    if (actTriang==nil)
      { ZIBStdOut("WriteTri: no triangulation to write\n"); return false; }
    lev = actTriang->refLevel;
    dep = actTriang->maxDepth;

    outFile = fopen((cmd->pars)[1], "w");
    if (outFile==nil)
      {
        sprintf(globBuf, "WriteTri: could not open file '%s'\n",
                (cmd->pars)[1]);
        ZIBStdOut(globBuf);
        return false;
      }

    tp = (cmd->pars)[2];
    index = CheckName(&tp, cmd->names, nameClass);

    switch (index)
         {
           case -1:  
             sprintf(globBuf, "WriteTri: unknown parameter '%s'\n",
                     (cmd->pars)[2]);
             ZIBStdOut(globBuf);
             fclose(outFile);
             return false;
	   case  0:
             if ((cmd->noOfPars)==2) lev = 0;
               else
                 {
                   rc = sscanf((cmd->pars)[3], "%d", &lev);
                   if (rc!=1)
                     {
                      sprintf(globBuf, "WriteTri: '%s' not a level number\n",
                              (cmd->pars)[2]);
                      ZIBStdOut(globBuf);
                      fclose(outFile);
                      return false;
                     }
                   if ((lev>(actTriang->refLevel))||(lev<0))
                     {
                      sprintf(globBuf, "WriteTri: level %d does not exist\n",
                              lev);
                      ZIBStdOut(globBuf);
                      fclose(outFile);
                      return false;
                     }
                 }

             counter = 0;
             ApplyPLevel(CountPt, lev);
             noOfPts = counter;
printf(" level, counter %d %d \n",lev, counter);

/*
             counter = 0;
             ApplyELevel(CountEdg, lev);
             noOfEdgs = counter;
             counter = 0;
             ApplyTLevel(CountTr, lev);
             noOfTrs = counter;
*/

             fprintf(outFile, "%s\n", actTriang->name);
             fprintf(outFile, "Dimension:(%d)\n", noOfPts);

             counter = 0;
             ApplyPLevel(WritePt, lev);
             fprintf(outFile, "END\n");
             noOfLines += noOfPts+1;

/*
             counter = 0;
             ApplyELevel(WriteEdg, lev);
             fprintf(outFile, "END\n");
             noOfLines += noOfEdgs+1;

             counter = 0;
             ApplyTLevel(WriteTr, lev);
             fprintf(outFile, "END\n");
             noOfLines += noOfTrs+1;
*/

             fclose(outFile);
             sprintf(globBuf, 
               "WriteTri: %d lines written to '%s' for level %d\n",
                noOfLines, (cmd->pars)[1], lev);
             ZIBStdOut(globBuf);
             break;
 
           case  1:
             if ((cmd->noOfPars)==2) dep = 0;
               else
                 {
                   rc = sscanf((cmd->pars)[3], "%d", &dep);
                   if (rc!=1)
                     {
                      sprintf(globBuf, "WriteTri: '%s' not a depth number\n",
                              (cmd->pars)[2]);
                      ZIBStdOut(globBuf);
                      fclose(outFile);
                      return false;
                     }
                   if ((dep>(actTriang->maxDepth))||(dep<0))
                     {
                       sprintf(globBuf, "WriteTri: level %d does not exist\n",
                               lev);
                       ZIBStdOut(globBuf);
                       fclose(outFile);
                       return false;
                     }
                  }

/*
             counter = 0;
             ApplyPDepth(CountPt, dep, TRIANGUL, all);
             noOfPts = counter;
             counter = 0;
             ApplyEDepth(CountEdg, dep, TRIANGUL, all);
             noOfEdgs = counter;
             counter = 0;
             ApplyTDepth(CountTr, dep, TRIANGUL, all);
             noOfTrs = counter;

             fprintf(outFile, "%s\n", actTriang->name);
             fprintf(outFile, "Dimension:(%d,%d,%d,%d)\n", 
                     noOfPts, noOfEdgs, noOfTrs, 0);

             counter = 0;
             ApplyPDepth(WritePt, dep, TRIANGUL, all);
             fprintf(outFile, "END\n");
             noOfLines += noOfPts+1;

             counter = 0;
             ApplyEDepth(WriteEdg, dep, TRIANGUL, all);
             fprintf(outFile, "END\n");
             noOfLines += noOfEdgs+1;

             counter = 0;
             ApplyTDepth(WriteTr, dep, TRIANGUL, all);
             fprintf(outFile, "END\n");
             noOfLines += noOfTrs+1;
*/

             fclose(outFile);
             sprintf(globBuf,
                     "WriteTri: %d lines written to '%s' for depth %d\n",
                     noOfLines, (cmd->pars)[1], lev);
             ZIBStdOut(globBuf);

             break;
       }
       return true;
  }
