/*
C
C  _______________________________________________________________
C
C*   Licence
C    =======
C
C    You may use or modify this code for your own non commercial
C    purposes for an unlimited time. 
C    In any case you should not deliver this code without a special 
C    permission of ZIB.
C    In case you intend to use the code commercially, we oblige you
C    to sign an according licence agreement with ZIB.
C
C
C  _______________________________________________________________
C
*/

#include <math.h>

#include "kask.h"
#include "kaskass.h"
#include "kaskcmd.h"

#define REZSQRT3 0.577350269
#define CW1 0.133333333333333
#define CW2 0.075000000000000
#define QW1 0.041666666666666
#define QA 0.5854102000000
#define QB 0.1381966000000


#define UC1 0.0919710781
#define UC2 0.7240867657
#define UC3 0.0563508327
#define UC4 0.4436491673
#define UC5 0.3197936278
#define UC6 0.0406191166
#define UW1 0.01975308641
#define UW2 0.01198951396
#define UW3 0.00881834215
#define UW4 0.01151136786

#define HW1 0.133333333333333
#define HW2 0.008333333333333

/*
    Integration points and weights
*/
static REAL cubicIPX[] = { QUARTER, SIXTH, HALF,  SIXTH, SIXTH };
static REAL cubicIPY[] = { QUARTER, SIXTH, SIXTH, HALF,  SIXTH };
static REAL cubicIPZ[] = { QUARTER, SIXTH, SIXTH, SIXTH, HALF  };
static REAL cubicIW[]  = { -CW1,    CW2,    CW2,   CW2,  CW2   };

static REAL linIPX[] = { QUARTER };
static REAL linIPY[] = { QUARTER };
static REAL linIPZ[] = { QUARTER };
static REAL linIW[]  = { SIXTH   };

static REAL quadIPX[] = { QB,  QA,  QB,  QB };
static REAL quadIPY[] = { QB,  QB,  QA,  QB };
static REAL quadIPZ[] = { QB,  QB,  QB,  QA };
static REAL quadIW[]  = { QW1, QW1, QW1, QW1};

static REAL hyperIPX[] = { QUARTER,  ZERO,  ONE ,  ZERO,  ZERO };
static REAL hyperIPY[] = { QUARTER,  ZERO,  ZERO,  ONE,   ZERO };
static REAL hyperIPZ[] = { QUARTER,  ZERO,  ZERO,  ZERO,  ONE  };
static REAL hyperIW[]  = { HW1,      HW2,   HW2,   HW2,   HW2 };
INT hyperNoIP = 5;

static REAL userIPX[] = { QUARTER, UC1, UC2, UC1, UC1,
                          UC4, UC3, UC3, UC3, UC4, UC4,
                          UC5, UC6, UC5, UC5 };
static REAL userIPY[] = { QUARTER, UC1, UC1, UC2, UC1,
                          UC3, UC4, UC3, UC4, UC3, UC4,
                          UC5, UC5, UC6, UC5 };
static REAL userIPZ[] = { QUARTER, UC1, UC1, UC1, UC2,  
                          UC3, UC3, UC4, UC4, UC4, UC3,
                          UC5, UC5, UC5, UC6 };
static REAL userIW[]  = { UW1, UW2,UW2,UW2,UW2, UW3,UW3,UW3,UW3,UW3,UW3, 
                          UW4,UW4,UW4,UW4 };
INT userNoIP = 15;


REAL **assA = nil, *assB = nil, **assAG = nil, **assAL = nil,
	 *assBG = nil;
 
/*
    Values of shape functions
	
        phi[1]=1-x-y-z; phi[2]=x; phi[3]=y; phi[4]=z;
		qPhi[1]= ;
		qPhi[2]= ;
		qPhi[3]= ;
		qPhi[4]= ;
		qPhi[5]= ;
		qPhi[6]= ;
		qPhi[7]= ;
		qPhi[8]= ;
		qPhi[9]= ;
		qPhi[10]= ;
		
    at the integration points [Schwarz, p.134f]
*/

void StdShape(x,y,z,no,f,fx,fy,fz,fxx,fxy,fyy)
  REAL x, y, z, *f, *fx, *fy, *fz, *fxx, *fxy, *fyy;
  INT no;
  {
    switch (no)
	{
	  case 0:				/* 1-x-y-z */
	  	*f   = ONE-x-y-z;
		*fx  = -ONE;
		*fy  = -ONE;
		*fz  = -ONE;
		*fxx = ZERO;
		*fxy = ZERO;
		*fyy = ZERO;
		break;
	  case 1:				/* x */
	    *f	 = x;
		*fx  = ONE;
		*fy  = ZERO;
		*fz  = ZERO;
		*fxx = ZERO;
		*fxy = ZERO;
		*fyy = ZERO;
		break;
	  case 2:				/* y */
	    *f   = y;
		*fx  = ZERO;
		*fy  = ONE;
		*fz  = ZERO;
		*fxx = ZERO;
		*fxy = ZERO;
		*fyy = ZERO;
		break;
	  case 3:				/* z */
	    *f   = z;
		*fx  = ZERO;
		*fy  = ZERO;
		*fz  = ONE;
		*fxx = ZERO;
		*fxy = ZERO;
		*fyy = ZERO;
		break;
	  case 4:				/* 4y(1-x-y-z) */
                *f   =  FOUR*y*(ONE-x-y-z);
		*fx  = -FOUR*y;
		*fy  =  FOUR*(1-x-TWO*y-z);
		*fz  = -FOUR*y;
		*fxx =  ZERO;
		*fxy = -FOUR;
		*fyy = -EIGHT;
		break;
	  case 5:				/* 4xy */
                *f   = FOUR*x*y;
		*fx  = FOUR*y;
		*fy  = FOUR*x;
		*fz  = ZERO;
		*fxx = ZERO;
		*fxy = FOUR;
		*fyy = ZERO;
		break;
	  case 6:				/* 4x(1-x-y-z) */
                *f   =  FOUR*x*(ONE-x-y-z);
		*fx  =  FOUR*(1-TWO*x-y-z);
		*fy  = -FOUR*x;
		*fz  = -FOUR*x;
		*fxx = -EIGHT;
		*fxy = -FOUR;
		*fyy =  ZERO;
		break;
	  case 7:				/* 4yz */
                *f   = FOUR*y*z;
		*fx  = ZERO;
		*fy  = FOUR*z;
		*fz  = FOUR*y;
		*fxx = ZERO;
		*fxy = ZERO;
		*fyy = ZERO;
		break;
	  case 8:				/* 4xz */
                *f   = FOUR*x*z;
		*fx  = FOUR*z;
		*fy  = ZERO;
		*fz  = FOUR*x;
		*fxx = ZERO;
		*fxy = ZERO;
		*fyy = ZERO;
		break;
	  case 9:				/* 4z(1-x-y-z) */
                *f   =  FOUR*z*(ONE-x-y-z);
		*fx  = -FOUR*z;
		*fy  = -FOUR*z;
		*fz  =  FOUR*(1-x-y-TWO*z);
		*fxx =  ZERO;
		*fxy =  ZERO;
		*fyy =  ZERO;
		break;
	}
	return;
  }

void StdQShape(x,y,z,no,f,fx,fy,fz,fxx,fxy,fyy)
  REAL x, y, z, *f, *fx, *fy, *fz, *fxx, *fxy, *fyy;
  INT no;
  {
    switch (no)
	{
	  case 0:				/* 2(1-x-y)(1/2-x-y) */	
	    *f   = TWO*(ONE-x-y)*(HALF-x-y);
		*fx  = FOUR*(x+y)-THREE;
		*fy  = FOUR*(x+y)-THREE;
		*fxx = FOUR;
		*fxy = FOUR;
		*fyy = FOUR;
		break;
	  case 1:				/* x(2x-1) */
	    *f   = x*(TWO*x-ONE);
		*fx  = FOUR*x-ONE;
		*fy  = ZERO;
		*fxx = FOUR;
		*fxy = ZERO;
		*fyy = ZERO;
		break;
	  case 2:				/* y(2y-1) */
	    *f   = y*(TWO*y-ONE);
		*fx  = ZERO;
		*fy  = FOUR*y-ONE;
		*fxx = ZERO;
		*fxy = ZERO;
		*fyy = FOUR;
		break;
	  case 3:				/* 4xy */
	    *f   = FOUR*x*y;
		*fx  = FOUR*y;
		*fy  = FOUR*x;
		*fxx = ZERO;
		*fxy = FOUR;
		*fyy = ZERO;
		break;
	  case 4:				/* 4y(1-x-y) */
	    *f   = FOUR*y*(ONE-x-y);
		*fx  = -FOUR*y;
		*fy  = FOUR*(1-x-TWO*y);
		*fxx = ZERO;
		*fxy = -FOUR;
		*fyy = -EIGHT;
		break;
	  case 5:				/* 4x(1-x-y) */
	    *f   = FOUR*x*(ONE-x-y);
		*fx  = FOUR*(1-TWO*x-y);
		*fy  = -FOUR*x;
		*fxx = -EIGHT;
		*fxy = -FOUR;
		*fyy = ZERO;
		break;
	}
	return;
  }

INT CompShapeVals(iData,ShapeF)
  INTEGDATA *iData;
  VOIDPROC ShapeF;
  {
    INT noIP = iData->noOfIPoints, noSF = iData->noOfShapeFunc;
	INT i, k;
	long lng;
	REAL **buffer, *buf;
	if (assAG==nil) assAG = ExpMatFull(assAG, 0, 10, false);
	if (assBG==nil) assBG = ExpVecFull(assBG, 0, 10);
	if (assAL==nil) assAL = ExpMatFull(assAL, 0, 10, false);
	assA = assAG;
	assB = assBG;

	lng = 7*((noIP*noSF)*sizeof(REAL)+noSF*sizeof(REAL*));

	buffer = (REAL**) ZIBAlloc(lng);
	if (buffer==nil)
	  {
	    ZIBStdOut("Assemble: not enough memory (CompShapeVals)\n");
		return false;
	  }

	iData->shape   = buffer; buffer += noSF;
	iData->shapeX  = buffer; buffer += noSF;
	iData->shapeY  = buffer; buffer += noSF;
	iData->shapeZ  = buffer; buffer += noSF;
	iData->shapeXX = buffer; buffer += noSF;
	iData->shapeXY = buffer; buffer += noSF;
	iData->shapeYY = buffer; buf = (REAL*)(buffer+noSF);
	for (k = 0; k<noSF; k++)
	  {
        (iData->shape)[k]   = buf; buf += noIP;
        (iData->shapeX)[k]  = buf; buf += noIP;
        (iData->shapeY)[k]  = buf; buf += noIP;
        (iData->shapeZ)[k]  = buf; buf += noIP;
        (iData->shapeXX)[k] = buf; buf += noIP;
        (iData->shapeXY)[k] = buf; buf += noIP;
        (iData->shapeYY)[k] = buf; buf += noIP;
        for (i = 0; i<noIP; i++)
            (*ShapeF)((iData->integPointX)[i],
                      (iData->integPointY)[i],
                      (iData->integPointZ)[i],
                      k,
                      &((iData->shape)[k][i]),
                      &((iData->shapeX)[k][i]),
                      &((iData->shapeY)[k][i]),
                      &((iData->shapeZ)[k][i]),
                      &((iData->shapeXX)[k][i]),
                      &((iData->shapeXY)[k][i]),
                      &((iData->shapeYY)[k][i])
                     );
	  }
	return true;
  }

INT UpdateIData(iData, iFormula, SF, ShapeF, symP)
  INTEGDATA *iData;
  INT iFormula, SF, symP;
  VOIDPROC ShapeF;
  {
	switch (iFormula)
	{
	  case CUBICIP:
		iData->integPointX = cubicIPX;
		iData->integPointY = cubicIPY;
		iData->integPointZ = cubicIPZ;
		iData->integWeight = cubicIW;
		iData->noOfIPoints = 5;
		break;
	  case LINIP:
		iData->integPointX = linIPX;
		iData->integPointY = linIPY;
		iData->integPointZ = linIPZ;
		iData->integWeight = linIW;
		iData->noOfIPoints = 1;
		break;
	  case QUADIP:
		iData->integPointX = quadIPX;
		iData->integPointY = quadIPY;
		iData->integPointZ = quadIPZ;
		iData->integWeight = quadIW;
		iData->noOfIPoints = 4;
		break;
	  case USERIP:
		iData->integPointX = userIPX;
		iData->integPointY = userIPY;
		iData->integPointZ = userIPZ;
		iData->integWeight = userIW;
		iData->noOfIPoints = 15;
		break;
	  case HYPERIP:
		iData->integPointX = hyperIPX;
		iData->integPointY = hyperIPY;
		iData->integPointZ = hyperIPZ;
		iData->integWeight = hyperIW;
		iData->noOfIPoints = 5;
		break;
	  default:
	    sprintf(globBuf, "Assemble: integration formula %d not known\n",
				iFormula);
		ZIBStdOut(globBuf);

		ZIBFree((PTR)iData);
		return false;
	}
	iData->symP = symP;
	iData->noOfShapeFunc = SF;

	CompShapeVals(iData, ShapeF);

	return true;
  }

INTEGDATA *NewIData(iFormula, SF, ShapeF, symP)
  INT iFormula, SF, symP;
  VOIDPROC ShapeF;
  {
    INTEGDATA *iData;

	iData = (INTEGDATA*)ZIBAlloc((long)sizeof(INTEGDATA));
	if (iData==nil)
	  { ZIBStdOut("Assemble: not enough memory (NewIData)\n"); return nil; }

	UpdateIData(iData, iFormula, SF, ShapeF, symP);

	return iData;
  }
