/*
C
C  _______________________________________________________________
C
C*   Licence
C    =======
C
C    You may use or modify this code for your own non commercial
C    purposes for an unlimited time. 
C    In any case you should not deliver this code without a special 
C    permission of ZIB.
C    In case you intend to use the code commercially, we oblige you
C    to sign an according licence agreement with ZIB.
C
C
C  _______________________________________________________________
C
*/

#include <stdio.h>
#include <math.h>

#include "kask.h"
#include "kaskcmd.h"
#include "kasktri.h"
#include "kaskass.h"
#include "kasksol.h"

SOLVE *actSolve = nil;

REAL mbar, Theta, newFactor = 1.0;
REAL discrErr, discrMaxErr;
REAL maxDiff, maxDiffRestr, xMax, yMax, zMax;
REAL eNormU = ZERO, eNormUUl=ZERO;
REAL totalTime=ZERO, solTime=ZERO, estiTime=ZERO, refTime=ZERO;
REAL iteTime=ZERO, dirTime=ZERO, assTime=ZERO, numTime=ZERO, drawTime;
REAL ares, res;
INT  noOfFCalls=0;

static int curSteps, breakReason;


static INT CompDiscrErrP(p)
  PT *p;
  {
    REAL adiff, diff;
    REAL solVals[NO_OF_SOL_VALS];
    INT class = p->class;

    (actProblem->Sol)(p->x,p->y,p->z, class,solVals,-1);
    diff = RA(p,R_SOL)-solVals[0];
 
 
    discrErr += diff*diff;
    adiff = fabs(diff);
    if (discrMaxErr < adiff) 
    discrMaxErr = adiff;

    return true;
  }




static void CompMeanVals(verboseP)
  INT verboseP;
  {
    InitNumAss(N_LQ);
    (actSolve->compNormP) = false;
    if (!axMul(R_SOL, R_RQ))
      { actSolve->estiFail = true; return; }
	AssRSideDiag();
	mbar = CompRes(R_RES, R_RQ, R_DIAG, R_RHS, &(actSolve->globEps));
	actSolve->globEps += actSolve->iteEps;

	InitNumAss(N_STD);
	if (verboseP)
	  {
            sprintf(globBuf, "Solve: estiEps=%9.3e(%e), mbar=%9.3e (CompEVals)\n",
                 actSolve->globEps, sqrt(actSolve->globEps),mbar);
		ZIBStdOut(globBuf);
	  }
	return;
  }


void MeanValue(verboseP)
  INT verboseP;
  {
    REAL localEstiTime;	
    REAL eNormUh=ZERO;

    localEstiTime = ZIBSeconds();

/*     energy norm of discret solution for relative error	*/

    (actSolve->compNormP) = true;
/*
    if (actSolve->sparseP)
      if (!SparseAssemble())
        { actSolve->iteFail = true; return; }

    InitNumAss(N_STD);
    if (!(actSolve->MatrixMul)(R_SOL, R_CGH))
      { actSolve->estiFail = true; return; }
	
    eNormUh = scalprod(R_SOL,R_CGH);
*/

    eNormUh = ONE;
    (actSolve->eNorm) = sqrt(eNormUh);
    (actSolve->compNormP) = false;

    sprintf(globBuf," energy norm of Uh    %e \n", (actSolve->eNorm));
    ZIBStdOut(globBuf);


/*	estimation of error in energy norm	*/
    CompMeanVals(verboseP);
    if ((actSolve->globEps)<ZERO)
      {
        ZIBStdOut("Estimate: stiffness matrix maybe not positive definite\n");
        actSolve->estiFail = true;
        return;
      }
    if (verboseP) 
       ZIBStdOut("\n step   level depth  dim    estiErr  iter#    iteErr    trueErr on grid         time\n");
	sprintf(globBuf,"Estimate  %3d   %3d  %5d  %9.3e\n",
		    	actTriang->refLevel, actTriang->maxDepth,
				actTriang->noOfPoints, sqrt(actSolve->globEps));
	ZIBStdOut(globBuf);
	actSolve->estiP = true;
	actSolve->lastStep = ESTIMATE;
	actSolve->estiFail = false;

	localEstiTime = ZIBSeconds() - localEstiTime;
	estiTime  += localEstiTime;
	totalTime += localEstiTime;

	return;  }



static void CompExtrVals(verboseP)
  INT verboseP;
  {

    InitNumAss(N_LQ);
    (actSolve->compNormP) = false;


    if (!axMul(R_SOL, R_RQ))
      { actSolve->estiFail = true; return; }

    AssRSideDiag();

    mbar = CompResExtrapol(R_RES, R_RQ, R_DIAG, R_RHS, &(actSolve->globEps));
    actSolve->globEps += actSolve->iteEps;

    InitNumAss(N_STD);
    if (verboseP)
      {
        sprintf(globBuf, "Solve: estiEps=%9.3e(%e), mbar=%9.3e (CompEVals)\n",
                        actSolve->globEps, sqrt(actSolve->globEps),mbar);
        ZIBStdOut(globBuf);
      }
   return;
  }



void ExtraPol(verboseP)
  INT verboseP;
  {
    REAL localEstiTime;
    REAL eNormUh=ZERO;

    localEstiTime = ZIBSeconds();

/*     energy norm of discret solution for relative error	*/

    (actSolve->compNormP) = true;
/*
    if (actSolve->sparseP)
    if (!SparseAssemble())
      { actSolve->iteFail = true; return; }

    InitNumAss(N_STD);
    if (!(actSolve->MatrixMul)(R_SOL, R_CGH))
      { actSolve->estiFail = true; return; }
	
    eNormUh = scalprod(R_SOL,R_CGH);
*/ 

    InitNumAss(N_STD);
    if (!(axMul)(R_SOL, R_CGH))
      { actSolve->estiFail = true; return; }
	
    eNormUh = scalprod(R_SOL,R_CGH);

    (actSolve->eNorm) = sqrt(eNormUh);
    printf(" norm %e \n",actSolve->eNorm);   
    (actSolve->compNormP) = false;
/*
    sprintf(globBuf," energy norm of Uh    %e \n", (actSolve->eNorm));
    ZIBStdOut(globBuf);
*/

/*	estimation of error in energy norm	*/
 

    CompExtrVals(verboseP);
    if ((actSolve->globEps)<ZERO)
      {
       ZIBStdOut("Estimate: stiffness matrix maybe not positive definite\n");
       actSolve->estiFail = true;
       return;
      }
    if (verboseP) 
      ZIBStdOut("\n step   level depth  dim    estiErr  iter#    iteErr    trueErr on grid         time\n");
      sprintf(globBuf,"Estimate  %3d   %3d  %5d  %9.3e\n",
                      actTriang->refLevel, actTriang->maxDepth,
                      actTriang->noOfPoints, sqrt(actSolve->globEps));
      ZIBStdOut(globBuf);
      actSolve->estiP = true;
      actSolve->lastStep = ESTIMATE;
      actSolve->estiFail = false;

      localEstiTime = ZIBSeconds() - localEstiTime;
      estiTime  += localEstiTime;
      totalTime += localEstiTime;

    return;
  }

static void SortExtrVals(verboseP)
  INT verboseP;
  {

    InitNumAss(N_LQ);
    (actSolve->compNormP) = false;


    if (!axMul(R_SOL, R_RQ))
      { actSolve->estiFail = true; return; }

    AssRSideDiag();

    mbar = SortResExtrapol(R_RES, R_RQ, R_DIAG, R_RHS, &(actSolve->globEps));
    actSolve->globEps += actSolve->iteEps;

    InitNumAss(N_STD);
    if (verboseP)
      {
        sprintf(globBuf, "Solve: estiEps=%9.3e(%e), mbar=%9.3e (CompEVals)\n",
                        actSolve->globEps, sqrt(actSolve->globEps),mbar);
        ZIBStdOut(globBuf);
      }
   return;
  }


void ExtraSort(verboseP)
  INT verboseP;
  {
    REAL localEstiTime;
    REAL eNormUh=ZERO;

    localEstiTime = ZIBSeconds();

/*     energy norm of discret solution for relative error	*/

    (actSolve->compNormP) = true;
/*
    if (actSolve->sparseP)
    if (!SparseAssemble())
      { actSolve->iteFail = true; return; }

    InitNumAss(N_STD);
    if (!(actSolve->MatrixMul)(R_SOL, R_CGH))
      { actSolve->estiFail = true; return; }
	
    eNormUh = scalprod(R_SOL,R_CGH);
*/ 

    eNormUh = ONE;  

    (actSolve->eNorm) = sqrt(eNormUh);
    printf(" norm %e \n",actSolve->eNorm);   
    (actSolve->compNormP) = false;
/*
    sprintf(globBuf," energy norm of Uh    %e \n", (actSolve->eNorm));
    ZIBStdOut(globBuf);
*/

/*	estimation of error in energy norm	*/
 

    SortExtrVals(verboseP);
    if ((actSolve->globEps)<ZERO)
      {
       ZIBStdOut("Estimate: stiffness matrix maybe not positive definite\n");
       actSolve->estiFail = true;
       return;
      }
    if (verboseP) 
      ZIBStdOut("\n step   level depth  dim    estiErr  iter#    iteErr    trueErr on grid         time\n");
      sprintf(globBuf,"Estimate  %3d   %3d  %5d  %9.3e\n",
                      actTriang->refLevel, actTriang->maxDepth,
                      actTriang->noOfPoints, sqrt(actSolve->globEps));
      ZIBStdOut(globBuf);
      actSolve->estiP = true;
      actSolve->lastStep = ESTIMATE;
      actSolve->estiFail = false;

      localEstiTime = ZIBSeconds() - localEstiTime;
      estiTime  += localEstiTime;
      totalTime += localEstiTime;

    return;
  }


static INT RefInitMark(td)
   TR3 *td;
{
   td->mark = false;
   if ( (td->father) != nil)
          (td->father)->mark = false;

   return true;
}


static INT RefIfLessTheta(td)
   TR3 *td;
{	
   PT *p1=td->p1, *p2, *p3, *p4;
   EDG *e1=td->e1, *e2=td->e2, *e3=td->e3, *e4=td->e4, *e5=td->e5, *e6=td->e6;
    

   if ( (RA(e1, R_RES)>=Theta) || (RA(e2, R_RES)>=Theta) ||
        (RA(e3, R_RES)>=Theta) || (RA(e4, R_RES)>=Theta) ||
        (RA(e5, R_RES)>=Theta) || (RA(e6, R_RES)>=Theta) )  
      { 
        if ( ((td->type) != T_GREEN_1)  && ((td->type) != T_GREEN_2A)
          && ((td->type) != T_GREEN_2B) && ((td->type) != T_GREEN_3) )
           {
            RefTr(td);
           }
        else
          {
            RefTr(td->father);
          }
      }
/*

    if ( (td->res) >= Theta )  
      { 
        RefTr(td);
      }
*/
    return true;
  }




void Refine(verboseP)
   INT verboseP;
{
   INT initNoOfPoints = actTriang->noOfPoints, levCount = 0;
   REAL localRefTime;

   if (!(actSolve->estiP))
     {
       ZIBStdOut("Refine: no data from the error estimator\n");
       if ((actSolve->Estimate) != nil)
         {  actSolve->refFail = true; return; }
     }

   if (verboseP)
     {
       sprintf(globBuf, "Solve: start adaptive refinement, sigma=%.3f, s=%.1f, %d points\n", 
              actSolve->sigma, actSolve->sFactor, initNoOfPoints);
       ZIBStdOut(globBuf);
     }

   actSolve->refFail = false;
   if ((actSolve->Estimate) == nil)
     {
       localRefTime = ZIBSeconds();
       OpenRef();
       ApplyT3(RefTr,all);
       CloseRef(true);
       sprintf(globBuf,"Refine    %3d   %3d  %5d\n",
               actTriang->refLevel, actTriang->maxDepth, actTriang->noOfPoints);
       ZIBStdOut(globBuf);

       localRefTime = ZIBSeconds() - localRefTime;
       refTime   += localRefTime;
       totalTime += localRefTime;
     }
   else
     while (true)
       {
         localRefTime = ZIBSeconds();
         Theta = (actSolve->sigma)*mbar;

         ApplyT3(RefInitMark, all);
         ApplyT3(RefIfLessTheta, all);
         OpenRef();

         if (!CloseRef(verboseP))
           {
             actSolve->refFail = true;
             if (initNoOfPoints==(actTriang->noOfPoints))
                ZIBStdOut("Solve: solution identical to zero\n");
             break;
           }
         levCount++;
         actSolve->estiP = false;
         if (verboseP)
            ZIBStdOut("\n step   level depth  dim    estiErr  iter#    iteErr    trueErr on grid         time\n");
         sprintf(globBuf,"Refine    %3d   %3d  %5d\n",
                 actTriang->refLevel, actTriang->maxDepth, actTriang->noOfPoints);
         ZIBStdOut(globBuf);

         localRefTime = ZIBSeconds() - localRefTime;
         refTime   += localRefTime;
         totalTime += localRefTime;

/*         break;    */


         if (initNoOfPoints*(actSolve->sFactor)<=(actTriang->noOfPoints)) break;
         if ((actSolve->breakDim)<(actTriang->noOfPoints)) break;
         if ((actSolve->breakLevel)<(actTriang->refLevel)) break;
         if ((actSolve->breakDepth)<(actTriang->maxDepth)) break;
	
         (actSolve->Estimate)(verboseP);
         if ((actSolve->globEps)<ZERO)
           {
             ZIBStdOut("Estimate: stiffness matrix maybe not positive definite\n");
             actSolve->estiFail = true;
             return;
           }
      }

   if (verboseP)
     {
       sprintf(globBuf, "Solve: end adaptive refinement, %d new points, %d level(s)\n",
               (actTriang->noOfPoints)-initNoOfPoints,levCount);
       ZIBStdOut(globBuf);
     }

   newFactor = initNoOfPoints;
   newFactor = newFactor/(actTriang->noOfPoints);
   newFactor = pow(newFactor,TWOTHIRD);
   actSolve->lastStep = REFINE;
   actSolve->refFail = false;

   return;
  }



void DirectEnvelope(verboseP)
  INT verboseP;
  {
    INT      rc, dim= actTriang->noOfPoints;
    INT      i, j;
    REAL     localDrawTime, localDirTime, localAssTime, localSortTime;
    REAL     loc0CholTime, loc1CholTime;
    PT       *p;
    NEIGHBOR *np;

    (actTriang->sparseAvailable) = false;

    localSortTime = ZIBSeconds();
    ReverseCuthillMcKee();  
/*    NoSorting();   */   
    localSortTime = ZIBSeconds() - localSortTime;
    printf(" Sort.. %8.4f \n",localSortTime);


   i=0;
   j=0;
   p= actTriang->firstPoint;
   while (p!=nil)
     {
       np = p->neighbors;
       i += ( (p->index) - (np[0].p->index) );
       j += p->noOfNeighbors;
       p = p->next;
     }
   printf(" sum neighbors A %d \n",j);
   printf(" env(A)          %d \n\n",i);


    localAssTime = ZIBSeconds();
    if(!SparseAssemble())
	  { actSolve->dirSolP = false; actSolve->dirFail = true; return; }


    if (!AssembleEnv())
      { actSolve->dirSolP = false; actSolve->dirFail = true; return; }

    localAssTime = ZIBSeconds() - localAssTime;
    assTime   += localAssTime;
    totalTime += localAssTime;


    localDirTime = ZIBSeconds();

    if (verboseP)
      {
	    sprintf(globBuf, "Solve: start of direct cholesky, n=%d\n", dim);
		ZIBStdOut(globBuf);
	  }

    loc0CholTime = ZIBSeconds();
    rc = CholDecEnvelope(actSolve->diagonal, actSolve->env, 
                         actSolve->decomp, dim);

    loc0CholTime = ZIBSeconds() - loc0CholTime;

    if (!rc)
      {
        ZIBStdOut("Solve: direct cholesky failed (CholDec)\n");
        actSolve->dirSolP = false;
		actSolve->dirFail = true;
        return;
      }

    loc1CholTime = ZIBSeconds();

    StoreFieldToVec(R_RHS,actSolve->rhs);
    rc = actSolve->DirectSol(actSolve->env, actSolve->decomp, 
                   actSolve->diagonal, actSolve->rhs,actSolve->solution, dim);

    loc1CholTime = ZIBSeconds() - loc1CholTime;
    printf(" CholDec  Time %8.4f \n",loc0CholTime);
    printf(" CholSol  Time %8.4f \n",loc1CholTime);

    if (!rc)
      {
        ZIBStdOut("Solve: direct cholesky failed (CholSol)\n");
        actSolve->dirSolP = false;
		actSolve->dirFail = true;
        return;
      }

    StoreVecToField(actSolve->solution,R_SOL);

    sparseExists = 0;
    actSolve->dirSolP = true;
    actSolve->dirFail = false;
    actSolve->dirLevel = actTriang->refLevel;
    actSolve->lastStep = DIRECT;
    actSolve->iteEps = ZERO;
    if (verboseP) ZIBStdOut("Solve: end of direct cholesky\n");
    actTriang->lastDirectPoint = actTriang->lastPoint;

    localDirTime = ZIBSeconds()-localDirTime;
    dirTime   += localDirTime + localSortTime;
    totalTime += localDirTime + localSortTime;


    if (verboseP)
       ZIBStdOut("\n step   level depth  dim    estiErr  iter#    iteErr    trueErr on grid         time\n");

    if ((actProblem->Sol)!=nil)
      {
        discrErr  = ZERO;
        discrMaxErr = ZERO;
        ApplyP(CompDiscrErrP,all);	 
        sprintf(globBuf,"Direct    %3d   %3d  %5d                                %9.3e          ",
		          actTriang->refLevel, actTriang->maxDepth,
				  actTriang->noOfPoints, discrMaxErr);
       }
    else
       sprintf(globBuf,"Direct    %3d   %3d  %5d                                                   ",
		          actTriang->refLevel, actTriang->maxDepth,actTriang->noOfPoints);
             
       ZIBStdOut(globBuf);

       if (verboseP)
         {
          sprintf(globBuf,"\n");
          ZIBStdOut(globBuf);
         }
 
   return;
  }

void DirectFullSparse(verboseP)
  INT verboseP;
  {
    INT   rc, dim= actTriang->noOfPoints;
    INT   i, j;
    REAL  localDrawTime, localDirTime, localAssTime;
    REAL  localSortTime, localStructTime;
    REAL  loc0CholTime, loc1CholTime;
    PT    *p;
    NEIGHBOR *nb=nil;

    (actTriang->sparseAvailable) = false;


    localSortTime = ZIBSeconds();
    NestedDissection();            
/*     NoSorting();  */                 
    localSortTime = ZIBSeconds() - localSortTime;
    printf(" Sort.. %8.4f \n",localSortTime);

 
   i=0;
   p= actTriang->firstPoint;
   while (p!=nil)
     {
       i += p->noOfNeighbors;
       p = p->next;
     }
   printf(" sum neighbors A %d \n",i);


    localAssTime = ZIBSeconds();
    if(!SparseAssemble())
	  { actSolve->dirSolP = false; actSolve->dirFail = true; return; }

    localAssTime = ZIBSeconds() - localAssTime;
    assTime   += localAssTime;
    totalTime += localAssTime;

    localStructTime = ZIBSeconds();
    if(!SparseLowerStruct())
      { actSolve->dirSolP = false; actSolve->dirFail = true; return; };
    localStructTime = ZIBSeconds() - localStructTime;
    printf(" Symb.Factorization %8.4f \n",localStructTime);

/*
   p= actTriang->firstPoint;
   while (p!=nil)
     {
      nb = p->neighbors;
      for (j=0; j<p->noOfNeighbors;j++)
        {
          printf("j %d, stiff %e, ilu %e \n",j,nb[j].stiff, nb[j].ilu);
        }
       printf("\n");
       p=p->next;
      }
*/


   j=0;
   p= actTriang->firstPoint;
   while (p!=nil)
     {
       j += p->noOfNeighbors;
       p = p->next;
     }
   printf(" fill in %d \n\n",j-i);


    localAssTime = ZIBSeconds();
    if (!AssembleSparse())
      { actSolve->dirSolP = false; actSolve->dirFail = true; return; }
 
    localAssTime = ZIBSeconds() - localAssTime;
    assTime   += localAssTime;
    totalTime += localAssTime;


    localDirTime = ZIBSeconds();

    if (verboseP)
      {
	    sprintf(globBuf, "Solve: start of direct cholesky, n=%d\n", dim);
		ZIBStdOut(globBuf);
	  }
 
    loc0CholTime = ZIBSeconds();
 
    rc = CholDecSparse(actSolve->env, actSolve->decomp, dim);   
    actTriang->lastDirectPoint = actTriang->lastPoint;


    loc0CholTime = ZIBSeconds() - loc0CholTime;

    if (!rc)
      {
        ZIBStdOut("Solve: direct cholesky failed (CholDec)\n");
        actSolve->dirSolP = false;
		actSolve->dirFail = true;
        return;
      }



    loc1CholTime = ZIBSeconds();
    StoreFieldToVec(R_RHS,actSolve->rhs);

    rc = actSolve->DirectSol(actSolve->env, actSolve->decomp,
                   actSolve->diagonal, actSolve->rhs,actSolve->solution, dim);

    loc1CholTime = ZIBSeconds() - loc1CholTime;
    printf(" CholDec  Time %8.4f \n",loc0CholTime);
    printf(" CholSol  Time %8.4f \n",loc1CholTime);

    if (!rc)
      {
        ZIBStdOut("Solve: direct cholesky failed (CholSol)\n");
        actSolve->dirSolP = false;
		actSolve->dirFail = true;
        return;
      }

    sparseExists = 0;
    actSolve->dirSolP = true;
    actSolve->dirFail = false;
    actSolve->dirLevel = actTriang->refLevel;
    actSolve->lastStep = DIRECT;
    actSolve->iteEps = ZERO;
    if (verboseP) ZIBStdOut("Solve: end of direct cholesky\n");
    StoreVecToField(actSolve->solution,R_SOL);
/*    StoreVecToField(actSolve->rhs,R_RHS);   */

    localDirTime = ZIBSeconds()-localDirTime;
    dirTime   += localDirTime + localSortTime + localStructTime;
    totalTime += localDirTime + localSortTime + localStructTime;


    if (verboseP)
       ZIBStdOut("\n step   level depth  dim    estiErr  iter#    iteErr    trueErr on grid         time\n");

    if ((actProblem->Sol)!=nil)
      {
        discrErr  = ZERO;
        discrMaxErr = ZERO;
        ApplyP(CompDiscrErrP,all);	 
        sprintf(globBuf,"Direct    %3d   %3d  %5d                                %9.3e          ",
		          actTriang->refLevel, actTriang->maxDepth,
				  actTriang->noOfPoints, discrMaxErr);
       }
    else
       sprintf(globBuf,"Direct    %3d   %3d  %5d                                                   ",
		          actTriang->refLevel, actTriang->maxDepth,actTriang->noOfPoints);
             
       ZIBStdOut(globBuf);

       if (verboseP)
         {
          sprintf(globBuf,"\n");
          ZIBStdOut(globBuf);
         }
 
   return;
  }


static INT CompDiff(p)
   PT *p;
{
   REAL dev;

   dev = RA(p,R_SOL) - RA(p,R_SOL_LAST);
   dev = fabs(dev);
   if (dev > maxDiff) 
     {
       maxDiff = dev;
       xMax = p->x; yMax = p->y; zMax = p->z;
     }
    
   if ( (p->level) < (actTriang->refLevel) )
     {
       if (dev > maxDiffRestr) maxDiffRestr = dev;
     }
   return true;
}




void Iterate(verboseP)
  INT verboseP;
  {
    INT rc, iterations = actSolve->maxIteSteps;
    INT countU;
    REAL eps =(actSolve->globEps)*(actSolve->iteFactor)*newFactor ;
    REAL localIteTime, localAssTime;
    PT *p;

/*  Last approximate solution   */
/*
    countU = assign(R_SOL_LAST,R_SOL);
*/

    localAssTime = ZIBSeconds();
    InitNumAss(N_STD);

/*  for test of iterative solver   */

/*
	eps = 1.0e-12;
	SetZeroField(R_SOL);  
*/ 

    rc = AssRSide();

    if (rc==false)
	  { actSolve->iteFail = true; return; }


    if (actSolve->sparseP)
    if (!SparseAssemble())
      { actSolve->iteFail = true; return; }
   
    localAssTime = ZIBSeconds() - localAssTime;
    assTime   += localAssTime;
    totalTime += localAssTime;


    localIteTime = ZIBSeconds();

    actSolve->iteSteps = 1;
    rc = CGFirst(actSolve->MatrixMul, eps, &res,
		actSolve->precondP, actSolve->Precond, verboseP);

    if (verboseP)
      {
        sprintf(globBuf, "Solve: first CG step: res=%9.3e(%9.3e)\n",
                            res, sqrt(res));
        ZIBStdOut(globBuf);
      }

    if ((res<ZERO)||(rc==false))
      { actSolve->iteFail = true; return; }

	



    if (res>eps)
      { 
        rc = CGStep(&iterations, &res, eps);
        actSolve->iteSteps = iterations;
        if (verboseP)
          {
           sprintf(globBuf, "Solve: after %d iteration steps: res=%e\n",
                             iterations, res);
           ZIBStdOut(globBuf);
          }
        if (rc==false) {
        actSolve->iteFail = true; return; }
      }

    actSolve->iteFail = (res<eps)?false:true;
    actSolve->lastStep = ITERATE;
    actSolve->iteEps = res;
    localIteTime = ZIBSeconds() - localIteTime;
    iteTime   += localIteTime;
    totalTime += localIteTime;

   /* difference between new and old approximation   */
/*
    maxDiff=ZERO;
    maxDiffRestr=ZERO;
    if (!ApplyP(CompDiff,all)) 
      {
        sprintf(globBuf, "Error in Iterate\n");
        ZIBStdOut(globBuf);
        return;
      }
    sprintf(globBuf, 
            "\n max. deviation from last level  = %e %e \n in %e %e %e \n",
                     maxDiff, maxDiffRestr, xMax,yMax,zMax);
    ZIBStdOut(globBuf);
*/

    if (verboseP)
      ZIBStdOut("\n step   level depth  dim    estiErr  iter#    iteErr    trueErr on grid         time\n");

    if ((actProblem->Sol)!=nil)
      {
        discrErr  = ZERO;
        discrMaxErr = ZERO;
        ApplyP(CompDiscrErrP,all);	

        sprintf(globBuf,"Iterate   %3d   %3d  %5d             %3d   %9.3e    %9.3e          ",
               actTriang->refLevel, actTriang->maxDepth,
               actTriang->noOfPoints, actSolve->iteSteps,
               sqrt(actSolve->iteEps),discrMaxErr);
       }
     else
       sprintf(globBuf,"Iterate   %3d   %3d  %5d             %3d   %9.3e                       ",
                        actTriang->refLevel, actTriang->maxDepth,
                        actTriang->noOfPoints, actSolve->iteSteps,
                        sqrt(actSolve->iteEps));
     ZIBStdOut(globBuf);

     if (verboseP) 
       {
         sprintf(globBuf,"\n");
         ZIBStdOut(globBuf);
       }
  return;
}

static INT StoreSolPt(p)
   PT   *p;
{
   RA(p,R_SOL_LAST) = RA(p,R_SOL);
   return true;
}


INT StoreSol(cmd)
   COMMAND *cmd;
{
   if (ParsCheck(cmd, 0, 0)) return false;

   ApplyP(StoreSolPt,all);
   actSolve->keepP = true;
   return true;
}


static INT ErrorSolPt(p)
   PT   *p;
{
   if (p->boundP == DIRICHLET)
     RA(p,R_SOL_LAST) = ZERO;
   else
     RA(p,R_SOL_LAST) = RA(p,R_SOL_LAST) - RA(p,R_SOL);
   return true;
}


INT ErrorSol(cmd)
   COMMAND *cmd;
{
   REAL eNormUh;

   if (ParsCheck(cmd, 0, 0)) return false;

   ApplyP(ErrorSolPt,all);
   InitNumAss(N_STD);
   if (!(axMul)(R_SOL_LAST, R_CGH))
      { actSolve->estiFail = true; return; }
	
   eNormUh = scalprod(R_SOL_LAST,R_CGH);

   printf(" error(energy norm): %e \n", sqrt(eNormUh));
   return true;
}


void StatisticsTime()
  {
   printf(" \ntime-statistics: \n");
   printf("process      CPU-time[s]  %% of accumul. time \n\n");
   printf("direct       %8.4f        %6.2f\n",dirTime,dirTime/totalTime*100);
   printf("estimate     %8.4f        %6.2f\n",estiTime,estiTime/totalTime*100);
   printf("refine       %8.4f        %6.2f\n",refTime,refTime/totalTime*100);
   printf("iterate      %8.4f        %6.2f\n",iteTime,iteTime/totalTime*100);
   printf("integrate    %8.4f        %6.2f\n\n",assTime,assTime/totalTime*100);
   printf("accumulated  %8.4f        %6.2f\n\n",totalTime,100.);
   printf("solve-cmd    %8.4f        \n\n",solTime);
   printf("noOfFCalls   %d           \n\n",noOfFCalls);
   return;
  }


INT Statistics(cmd)
  COMMAND *cmd;
  {
    if (ParsCheck(cmd, 0, 0)) return false;
    if (!CheckPreSets()) return false;
    StatisticsTime();
    return true;
  }



INT CheckPreSets()
  {
    SOLVE *sol = nil;
	
    if (actTriang==nil)
      {
        ZIBStdOut("Solve: no triangulation - nothing to solve\n");
        return false; 
      }

    if ((actTriang->problem)==nil) actTriang->problem = (PTR)actProblem;
    if (actProblem==nil)
      { ZIBStdOut("Solve: no problem - nothing to solve\n"); return false; }

    if ((actTriang->solve)==nil)
      {
        sol = (SOLVE*) ZIBAlloc((long)sizeof(SOLVE));
        if (sol==nil) return false;
        actTriang->solve = (PTR)sol;
        sol->Direct = DirectEnvelope;
        sol->DirectSol = CholSolEnvelope;
        sol->Estimate = MeanValue;
        sol->Precond = nil;
        sol->MatrixMul = SparseAxMul;
        sol->Iterate = Iterate;
        sol->Refine = Refine;
        sol->RedRefine = RefineTr3;
      }
    else return true;

    actSolve = (SOLVE*)actTriang->solve;
    actSolve->stiff = nil;
    actSolve->rhs = nil;
    actSolve->env = nil;
    actSolve->decomp = nil;
    actSolve->diagonal = nil;
    actSolve->solution = nil;
    actSolve->dirSolP = false;
    actSolve->dirLevel = 0;
    actSolve->dirDim = 0;
    actSolve->dirFail = false;
    actSolve->estiFail = false;
    actSolve->iteFail = false;
    actSolve->lastStep = START;
    actSolve->maxIteSteps = 1000;
    actSolve->precondP = false;
    actSolve->preCondName = "(no preconditioning)";
    actSolve->estiP = false;
    actSolve->estiPosition = 1;
    actSolve->keepP = false;
    actSolve->estiType = 1;
    actSolve->estiThreshold = 0.0;
    actSolve->estiQuadraticP = true;
    actSolve->constCoeff = false;
    actSolve->estiName = "(Estimator: meanValue)";
    actSolve->eNorm = ZERO;
    actSolve->sparseP = true;
    actSolve->compNormP = false;
    actSolve->globEps = ONE;
    actSolve->iteEps = ZERO;
    actSolve->breakEps = 1.0e-4;
    actSolve->breakDim = 20000;
    actSolve->breakLevel = 20;
    actSolve->breakDepth = 20;
    actSolve->breakSteps = 20;
    actSolve->iteFactor = 1.e-6;
    actSolve->sigma = 1.0;
    actSolve->sFactor = 2.0;
    actSolve->omega = 0.6;
    actSolve->layer = all;
    actSolve->broydenDelta = 10;
    actSolve->renumber = false;

    return true;
  }




INT RefCmd(cmd)
  COMMAND *cmd;
  {
   INT index;
   char *par1 = (cmd->pars)[1];

   if (ParsCheck(cmd,0,1)) return false;
   if (!CheckPreSets()) return false;    

   if ((cmd->noOfPars)==0)
     { 
       (actSolve->Refine)(true);    
       return true; 
     }

   index = CheckName(&par1,cmd->names, nameClass);
   if (index==-1)
     {
      sprintf(globBuf, "Solve: unknown parameter name '%s' for '%s'\n",
                      (cmd->pars)[1], (cmd->pars)[0]);
      ZIBStdOut(globBuf);
      return false;
     }

   if (!OpenRef()) return false;
   switch (index)
     {
       case 0:
         if (!ApplyT3(RefTr, all)) return false; 
         break;
       case 1:
         if (!RefTr(actTriang->firstTetrahedron)) return false;
         break;
       case 2:
         if (!RefTr(actTriang->lastTetrahedron)) return false;
         break;
       case 3:
         if (!ApplyT3(RefTrHyper,all)) return false;
         break;
     }
   if (!CloseRef(true))
     {
       actSolve->refFail = true;
       return false;
     }
 
   actSolve->globEps = ONE;
   actSolve->lastStep = REFINE;
   actSolve->refFail = false;
 
   return true;
  }



INT SelDirect(cmd)
  COMMAND *cmd;
  {
    INT index = 0;
	char *tp;

	if (ParsCheck(cmd,0,1)) return false;

	if ((cmd->noOfPars)==1)
	  {
	    tp = (cmd->pars)[1];
		index = CheckName(&tp, cmd->names, nameClass);
	  }
	if (index==-1)
	  {
	    sprintf(globBuf, "Parameter %s of %s not allowed\n",
				(cmd->pars)[1], (cmd->pars)[0]);
	    ZIBStdOut(globBuf);
		return false;
	  }

	if (!CheckPreSets()) return false;
    switch (index)
	{ 
          case 0: actSolve->Direct    = DirectEnvelope;   /*   envelope   */
                  actSolve->DirectSol = CholSolEnvelope;
                  break;
          case 1: actSolve->Direct = DirectFullSparse;   /*   fully sparse  */
                  actSolve->DirectSol = CholSolSparse;
                  break;
        } 
    return true;
  }


INT RedRefine(cmd)
  COMMAND *cmd;
  {
    INT index = 0;
    char *tp;

    if (ParsCheck(cmd,0,1)) return false;

    if ((cmd->noOfPars)==1)
      {
        tp = (cmd->pars)[1];
        index = CheckName(&tp, cmd->names, nameClass);
      }
    if (index==-1)
      {
        sprintf(globBuf, "Parameter %s of %s not allowed\n",
                         (cmd->pars)[1], (cmd->pars)[0]);
        ZIBStdOut(globBuf);
        return false;
      }

    if (!CheckPreSets()) return false;
    switch (index)
      { 
        case 0: actSolve->RedRefine = RefineTr3;    /*   shortest diagonal   */
                break;
        case 1:	actSolve->RedRefine = RefineBey;    /*   refinement of Bey   */
                break;
     } 
   return true;
  }



INT ErrEsti(cmd)
  COMMAND *cmd;
  {
    if (ParsCheck(cmd, 0, 0)) return false;
    if (!CheckPreSets()) return false;
    if ( (actSolve->Estimate) != nil)
       (actSolve->Estimate)(true);
	
    return true;
  }



INT RefStrategy(cmd)
  COMMAND *cmd;
  {
    INT index = 0;
    char *tp;

    if (ParsCheck(cmd,0,1)) return false;

    if ((cmd->noOfPars)==1)
      {
        tp = (cmd->pars)[1];
        index = CheckName(&tp, cmd->names, nameClass);
      }
    if (index==-1)
      {
        sprintf(globBuf, "Parameter %s of %s not allowed\n",
                         (cmd->pars)[1], (cmd->pars)[0]);
        ZIBStdOut(globBuf);
        return false;
      }

    if (!CheckPreSets()) return false;
    switch (index)
      { 
        case 0: actSolve->Estimate = MeanValue;    /*   std   */
                actSolve->estiName = "standard";
                actSolve->estiQuadraticP = true;
                break;
        case 1:	actSolve->Estimate = MeanValue;	  /*   MeanValue   */
                actSolve->estiName = "meanvalue";
                actSolve->estiQuadraticP = true;
                actSolve->sigma = 1.0;
                actSolve->sFactor = 2.0;
                break;
        case 2: actSolve->Estimate = ExtraPol;   /*   ExtraPolation the values on the edges   */
                actSolve->estiName = "extrapolation on edges";
                actSolve->estiQuadraticP = false;
                actSolve->sigma = 1.0;
                actSolve->sFactor = 1.1;
                break;
        case 3: actSolve->Estimate = ExtraSort;   /*   Sorting and ExtraPolation the values on the edges   */
                actSolve->estiName = "extrapolation on tetrahedra";
                actSolve->estiQuadraticP = false;
                actSolve->sigma = 1.0;
                actSolve->sFactor = 1.0;
                break;
        case 4:	actSolve->Estimate = nil;   /*   none   */
                actSolve->estiName = "none";
                actSolve->estiQuadraticP = false;
                actSolve->estiP = true;	 
                actSolve->iteFactor = 1.0e-14;
                break;
     } 

   sprintf(globBuf,"\n\n actual refinement strategy: %s \n\n",actSolve->estiName);
   ZIBStdOut(globBuf);

   return true;
  }


int SetBreak(cmd)
  COMMAND *cmd;
{
  INT index = 0, no, k, rc;
  double deps;
  REAL eps;
  char *tp;

  if (ParsCheck(cmd, 0, 2)) return false;
  if (!CheckPreSets()) return false;

  if ((cmd->noOfPars)==0)
    {
     ZIBStdOut("Solve: state of solution process\n");
     sprintf(globBuf, "  level     %5d - %5d\n",
                     actTriang->refLevel, actSolve->breakLevel);
     ZIBStdOut( globBuf);
     sprintf(globBuf, "  depth     %5d - %5d\n",
                      actTriang->maxDepth, actSolve->breakDepth);
     ZIBStdOut( globBuf);
     sprintf(globBuf, "  dimension %5d - %5d\n",
                      actTriang->noOfPoints, actSolve->breakDim);
     ZIBStdOut( globBuf);
     sprintf(globBuf, "  global error %9.3e - %9.3e\n", sqrt(actSolve->globEps),
                       sqrt(actSolve->breakEps));
     ZIBStdOut( globBuf);

     sprintf(globBuf, "  s-factor: %9.3e\n ", actSolve->sFactor);
     ZIBStdOut( globBuf);
     sprintf(globBuf, " sigma    : %9.3e\n ", actSolve->sigma);
     ZIBStdOut( globBuf); 

     sprintf(globBuf, " max steps for Iterate: %d\n ", actSolve->maxIteSteps);
     ZIBStdOut( globBuf);

     sprintf(globBuf, " iteration factor %9.3e\n", actSolve->iteFactor);
     ZIBStdOut( globBuf);

     sprintf(globBuf, "  last iteration error: %9.3e\n", sqrt(actSolve->iteEps));
     ZIBStdOut( globBuf);

     if (actSolve->dirSolP)
       {
        sprintf(globBuf, "  direct solution for level %d available\n",
                         actSolve->dirLevel);
        ZIBStdOut(globBuf);
       }
     else ZIBStdOut("  no direct solution available\n");

     if (actSolve->dirFail)
        ZIBStdOut("  Direct failed\n");
     else if (actSolve->estiFail)
            ZIBStdOut("  Estimate failed\n");
          else if (actSolve->iteFail)
                 ZIBStdOut("  Iterate failed\n");
			      else ZIBStdOut("  no failures yet\n");
    return true;
   }

    for (k = 1; k<=(cmd->noOfPars); k++)
       {
         tp = (cmd->pars)[k];
         index = CheckName(&tp, cmd->names, nameClass);
         if (index==-1)
           {
            sprintf(globBuf, "Solve: '%s' unknown parameter to %s\n",
                             (cmd->pars)[k], (cmd->pars)[0]);
            ZIBStdOut(globBuf);
            return false;
        }
    rc = 0;
    switch (index)
       {
         case 0:
                k++;
                if (k<=(cmd->noOfPars))
                  rc = sscanf((cmd->pars)[k],"%d",&no);
                if (rc!=1)
                  {
                    sprintf(globBuf, "Solve: '%s' should be followed by a number\n",
                                     (cmd->pars)[k-1]);
                    ZIBStdOut(globBuf);
                    break;
                   }
                 actSolve->breakDepth = no;
                 break;
          case 1:
                 k++;
                 if (k<=(cmd->noOfPars))
                   rc = sscanf((cmd->pars)[k],"%d",&no);
                 if (rc!=1)
                   {
                     sprintf(globBuf, "Solve: '%s' should be followed by a number\n",
                                      (cmd->pars)[k-1]);
                     ZIBStdOut(globBuf);
                     break;
                    }
                 actSolve->breakLevel = no;
                 break;
          case 2:
                 k++;
                 if (k<=(cmd->noOfPars))
                 rc = sscanf((cmd->pars)[k],"%d",&no);
                 if (rc!=1)
                   {
                    sprintf(globBuf, "Solve: '%s' should be followed by a number\n",
                           (cmd->pars)[k-1]);
                    ZIBStdOut(globBuf);
                    break;
                   }
                 actSolve->breakDim = no;
                 break;
          case 3:
                 k++;
                 if (k<=(cmd->noOfPars))
                 rc = sscanf((cmd->pars)[k],"%d",&no);
                 if (rc!=1)
                   {
                    sprintf(globBuf, "Solve: '%s' should be followed by a number\n",
                                     (cmd->pars)[k-1]);
                    ZIBStdOut(globBuf);
                    break;
                   }
                 actSolve->breakSteps = no;
                 break;
          case 4:                                  /* globalEps */
                 k++;
                 if (k<=(cmd->noOfPars))
                   {
                    rc = sscanf((cmd->pars)[k],"%le",&deps);
                    eps=deps;
                   }
                 if (rc!=1)
                   {
                    sprintf(globBuf, "Solve: '%s' should be followed by a number\n",
                                     (cmd->pars)[k-1]);
                    ZIBStdOut(globBuf);
                    break;
                   }
                 actSolve->breakEps = eps*eps;
                 break;
          case 5:
                 k++;
                 if (k<=(cmd->noOfPars))
                   rc = sscanf((cmd->pars)[k],"%d",&no);
                 if (rc!=1)
                   {
                    sprintf(globBuf, "Solve: '%s' should be followed by a number\n",
                                     (cmd->pars)[k-1]);
                    ZIBStdOut(globBuf);
                    break;
                   }
                 actSolve->maxIteSteps = no;
                 break;
          case 6:
                 k++;
                 if (k<=(cmd->noOfPars))
                   rc = sscanf((cmd->pars)[k],"%d",&no);
                 if (rc!=1)
                   {
                    sprintf(globBuf, "Solve: '%s' should be followed by a number\n",
                                     (cmd->pars)[k-1]);
                    ZIBStdOut(globBuf);
                    break;
                   }
                 actSolve->precondP = (no!=0)?true:false;
                 break;
          case 7:                              /*   iteFactor   */
                 k++;
                 if (k<=(cmd->noOfPars))
                   {
                    rc = sscanf((cmd->pars)[k],"%le",&deps);
                    eps = deps;
                   }
                 if (rc!=1)
                   {
                    sprintf(globBuf, "Solve: '%s' should be followed by a number\n",
                                     (cmd->pars)[k-1]);
                    ZIBStdOut(globBuf);
                    break;
                   }
                 actSolve->iteFactor = eps;
                 break;
          case 8:                            /*   sFactor   */
                 k++;
                 if (k<=(cmd->noOfPars))
                   {
                    rc = sscanf((cmd->pars)[k],"%le",&deps);
                    eps = deps;
                   }
                 if (rc!=1)
                   {
                    sprintf(globBuf, "Solve: '%s' should be followed by a number\n",
                                     (cmd->pars)[k-1]);
                    ZIBStdOut(globBuf);
                    break;
                   }
                 actSolve->sFactor = eps;
                 break;
          case 9:                             /*   sigma   */
                 k++;
                 if (k<=(cmd->noOfPars))
                   {
                     rc = sscanf((cmd->pars)[k],"%le",&deps);
                     eps = deps;
                   }
                 if (rc!=1)
                   {
                    sprintf(globBuf, "Solve: '%s' should be followed by a number\n",
                                     (cmd->pars)[k-1]);
                    ZIBStdOut(globBuf);
                    break;
                   }
                 actSolve->sigma = eps;
                 break;
          case 10:    /* ilu */
                  k++;
                  if (k<=(cmd->noOfPars))
                    rc = sscanf((cmd->pars)[k],"%d",&no);
                  if (rc!=1)
                    {
                     sprintf(globBuf, "Solve: '%s' should be followed by a number\n",
                                      (cmd->pars)[k-1]);
                     ZIBStdOut(globBuf);
                     break;
                    }
                  actSolve->sparseP = (no!=0)?true:false;
                  actSolve->MatrixMul = (no!=0)?SparseAxMul:axMul;
                  break;
         }
      }
    return true;
}



INT SolveDir(cmd)
  COMMAND *cmd;
  {
    if (ParsCheck(cmd, 0, 0)) return false;
	if (!CheckPreSets()) return false;
	(actSolve->Direct)(true);

    return true;
  }


INT PreCond(cmd)
  COMMAND *cmd;
  {    
	INT index = 0;
	char *tp;

	if (ParsCheck(cmd,0,1)) return false;

	if ((cmd->noOfPars)==1)
	  {
	    tp = (cmd->pars)[1];
		index = CheckName(&tp, cmd->names, nameClass);
	  }
	if (index==-1)
	  {
	    sprintf(globBuf, "Parameter %s of %s not allowed\n",
				(cmd->pars)[1], (cmd->pars)[0]);
	    ZIBStdOut(globBuf);
		return false;
	  }
	if (!CheckPreSets()) return false;
    switch (index)
	{ 
		case 0: actSolve->Precond = nil;			/*	std				*/
				actSolve->precondP = false;
				actSolve->preCondName = "(not preconditioned)";
				break;
		case 1:	actSolve->Precond = pcxMul;			/*	diagonal		*/
				actSolve->precondP = true;
				actSolve->preCondName = "(preconditioned by diagonal)";
				break;
		case 2: actSolve->Precond = nil;		/*	lowertriangle	*/
				actSolve->precondP = false;
				actSolve->preCondName = "(preconditioned by lower triangle)";
				break;
		case 3:	actSolve->Precond = nil;	/*	hierarch. bases	*/
				actSolve->precondP = false;
				actSolve->preCondName = "(preconditioned by hierarch. bases)";
				break;
		case 4:	actSolve->Precond = bpxPcxMul;		/*   bpx:bpxPcxMul   */
				actSolve->precondP = true;
				actSolve->preCondName = "(preconditioned by bpx)";
				InitBpx();
				break;
		case 5:	actSolve->Precond = nil;			/*	none			*/
				actSolve->precondP = false;
				actSolve->preCondName = "(not preconditioned)";
				break;

	 } 
	return true;
  }


INT SolveIte(cmd)
  COMMAND *cmd;
{
    if (ParsCheck(cmd, 0, 0)) return false;
    if (!CheckPreSets()) return false;
    (actSolve->Iterate)(true);
    return true;
}

int Angles(cmd)
   COMMAND *cmd;
{
   if (ParsCheck(cmd,0,0)) return false;

   if (actTriang==nil)
     { ZIBStdOut("WriteTri: no triangulation to write\n"); return false; }

   CompAngles();

   return true;
}


static int BreakCond()
{
   REAL eNorm = (actSolve->eNorm);
   if (actSolve->dirFail) { breakReason = 5; return true; }
   if (actSolve->estiFail) { breakReason = 6; return true; }
   if (actSolve->iteFail) { breakReason = 7; return true; }

   if (eNorm >= 1.0e-6) 
     {
       if ( (actSolve->breakEps) >= (actSolve->globEps)/(eNorm*eNorm) )
         { breakReason = 4; return true; }
     }
   else
     if ( (actSolve->breakEps) >= (actSolve->globEps)/(1.e-12) )
       { breakReason = 4; return true; }

   if ((actSolve->breakDepth)<=(actTriang->maxDepth))
     { breakReason = 3; return true; }
   if ((actSolve->breakLevel)<=(actTriang->refLevel))
     { breakReason = 2; return true; }
   if ((actSolve->breakDim)<=(actTriang->noOfPoints))
     { breakReason = 1; return true; }
   if ((actSolve->breakSteps)<=curSteps)
     { breakReason = 0; return true; }
   breakReason = -1;

   return false;
}


int Solve(cmd)
   COMMAND *cmd;
{
   INT saveMaxSteps, index, lastNoOfPoints;
   char *tp;
   REAL lastTime, accTime = 0.0, localSolTime, localRefTime;

   localSolTime = ZIBSeconds();

   if (ParsCheck(cmd, 0, 1)) return false;
   if (!CheckPreSets()) return false;
   saveMaxSteps = actSolve->breakSteps;

   if ((cmd->noOfPars)==1)
     {
       tp = (cmd->pars)[1];
       index = CheckName(&tp, cmd->names, nameClass);
       if (index==-1)
         {
           sprintf(globBuf, "Solve: '%s' unknown parameter to %s\n",
                   (cmd->pars)[1], (cmd->pars)[0]);
           ZIBStdOut(globBuf);
           return false;
         }
       actSolve->breakSteps = 1;
     }

   ZIBStdOut("\n step   level depth  dim    estiErr  iter#    iteErr    trueErr on grid         time\n");

   curSteps = 0;

   if ((actSolve->lastStep)==START)
     {
       lastTime = ZIBSeconds();
       (actSolve->Direct)(false);
       lastTime = ZIBSeconds()-lastTime;
       accTime += lastTime;
       curSteps++;
       sprintf(globBuf," %6.2f\n\n", 100.0*accTime/(actTriang->noOfPoints));
       ZIBStdOut(globBuf);
     }


   while (!BreakCond())
     {
       if ((actSolve->lastStep)!=ESTIMATE)
         {
           lastTime = ZIBSeconds();
           drawTime = 0.0;
           lastNoOfPoints = actTriang->noOfPoints;
           if ((actSolve->Estimate) == nil)
             {
               localRefTime = ZIBSeconds();
               OpenRef();
               ApplyT3(RefTr,all);
               CloseRef(true);
               sprintf(globBuf,"Refine    %3d   %3d  %5d\n",
               actTriang->refLevel, actTriang->maxDepth, actTriang->noOfPoints);
               ZIBStdOut(globBuf);
               (actSolve->lastStep) = REFINE;
               localRefTime  = ZIBSeconds() - localRefTime;
               refTime   += localRefTime;
               totalTime += localRefTime;
               sprintf(globBuf,"Refine    %3d   %3d  %5d  \n",
                       actTriang->refLevel, actTriang->maxDepth,
                       actTriang->noOfPoints);
               ZIBStdOut(globBuf);
             }
           else
             {
               (actSolve->Estimate)(false);
               if (BreakCond()) break;
               (actSolve->Refine)(false);
             }
         }
       if (!(actSolve->estiFail))
         {
           if ((actSolve->lastStep)!=REFINE)
             (actSolve->Refine)(false);
           (actSolve->Iterate)(false);
           lastNoOfPoints = (actTriang->noOfPoints)-lastNoOfPoints;
           sprintf(globBuf," %6.2f\n\n", 100.0*accTime/(actTriang->noOfPoints));
           ZIBStdOut(globBuf);
         }
       curSteps++;
     }

   actSolve->breakSteps = saveMaxSteps;
   switch (breakReason)
     {
       case 0:
         ZIBStdOut("BreakReason: number of steps reached\n");
         break;
       case 1:
         ZIBStdOut("BreakReason: number of points reached\n");
         break;
       case 2:
         ZIBStdOut("BreakReason: refinement level reached\n");
         break;
       case 3:
         ZIBStdOut("BreakReason: depth reached\n");
         break;
       case 4:
         ZIBStdOut("BreakReason: global (estimated) error reached\n");
         break;
       case 5:
         ZIBStdOut("BreakReason: direct solver failed\n");
         break;
       case 6:
         ZIBStdOut("BreakReason: error estimator failed\n");
         break;
       case 7:
         ZIBStdOut("BreakReason: iterative solver failed\n");
         break;
       default:
         sprintf(globBuf, "Unknown BreakReason: %d\n", breakReason);
         ZIBStdOut(globBuf);
         break;
     }

   localSolTime = ZIBSeconds() - localSolTime;
   solTime += localSolTime;
   return true;
}
