#include "kask.h"
#include "kasktri.h"
#include "kaskass.h"
#include "kaskcmd.h"
#include "kaskgraph.h"
#include "kasksol.h"

/*
 * This file contains routines for a nested dissection and a reverse
 * Cuthill-McKee ordering of the nodal points. 
 */

static INT      levelNo, treeSize, compLevel, minDeg, lastNo, more, count;
static INT      treeDepth, levelAt; 
static PT       *minPoint;
static PT       **pointOfNumber;
static PT       ***subGraph;
static INT      structNextNo, structLastNo, subGraphNo[100];
static PT       **levelStructNext, **levelStructLast, **levelStruct;

INT 
ApplyToNeighbors(Routine, to, of, data)
    PROC            Routine;
    PT             *of;
    INT             to, data;
{
    PT             *p;
    INT             i, rc;

    switch (to) {
    case all:
       for (i = 0; i < of->noOfNeighbors; i++)
         {
           p = (of->neighbors)[i].p;
           rc = (*Routine) (p, data);
           if (!rc)
              break;
         }
        break;
    }

   return rc;
}

static INT 
SetLevel(p, level)
	PT             *p;
	INT             level;
{
	if ((p->mark) && (p->counter == treeDepth))
     {
		p->mark = false;
		p->l = level;
		treeSize++;
		more = true;
        structNextNo++;
        levelStructNext[structNextNo] = p;
     }
	return true;
}

static INT 
SetLevelRCM(p, level)
	PT             *p;
	INT             level;
{
	if ( p->mark )
     {
		p->mark = false;
		p->l = level;
		treeSize++;
		more = true;
        structNextNo++;
        levelStructNext[structNextNo] = p;
     }
	return true;
}



static INT 
ClearLevel(p)
	PT             *p;
{
	p->l = 0;

	return true;
}


static INT 
SetTreeDepthZero(p)
	PT             *p;
{
	p->counter = 0;

	return true;
}



static INT 
Mark(p)
	PT             *p;
{
	p->mark = true;

	return true;
}

static INT 
UnMark(p)
	PT             *p;
{
	p->mark = false;

	return true;
}

static INT 
ResetMarkOnLS(p)
	PT             *p;
{
	if ( p->l > 0  )
		p->mark = true;

	return true;
}

static INT 
RootLS(root)
    PT             *root;
{
    PT             *p;
    INT            i;

    levelNo = 0;
    treeSize = 0;

/*    ApplyP(ClearLevel, all);   */
    for (i=0; i<subGraphNo[treeDepth]; i++)
       {
         p = subGraph[treeDepth][i];
         p->l = 0;
       }

    SetLevel(root, levelNo + 1);
    while (more)
     {
        levelNo++;
        more = false;
        levelStruct     = levelStructNext;
        levelStructNext = levelStructLast;
        levelStructLast = levelStruct;
        structLastNo    = structNextNo;
        structNextNo    = -1;

        for (i=0; i<=structLastNo; i++)
          {
			p = levelStructLast[i];
            ApplyToNeighbors(SetLevel, all, p, levelNo + 1);
          }
     }

/*    ApplyP(ResetMarkOnLS, all);   */
    for (i=0; i< subGraphNo[treeDepth]; i++)
       {
         p = subGraph[treeDepth][i];
         ResetMarkOnLS(p);
       }

	return levelNo;
}


static INT 
RootLsRCM(root)
    PT             *root;
{
    PT             *p;
    INT            i;

    levelNo = 0;
    treeSize = 0;

    ApplyP(ClearLevel, all);   

    SetLevelRCM(root, levelNo + 1);
    while (more)
     {
        levelNo++;
		more = false;
        levelStruct     = levelStructNext;
        levelStructNext = levelStructLast;
        levelStructLast = levelStruct;
        structLastNo    = structNextNo;
        structNextNo    = -1;

        for (i=0; i<=structLastNo; i++)
          {
			p = levelStructLast[i];
            ApplyToNeighbors(SetLevelRCM, all, p, levelNo + 1);
          }
     }

    ApplyP(ResetMarkOnLS, all);

	return levelNo;
}



static INT 
CompDeg(p)
	PT             *p;
{
	if ( p->l != compLevel )
		return true;

	if ((p->noOfNeighbors) <= minDeg)
      {
        minDeg = p->noOfNeighbors;
        minPoint = p;
      }
   return true;
}


static PT      *
MinDegOnLevel(level)
	INT             level;
{
    INT i;
    PT  *p;

	minDeg = actTriang->noOfPoints;
	compLevel = level;
	minPoint = nil;

/*   ApplyP(CompDeg, all);   */
    for (i=0; i<subGraphNo[treeDepth]; i++)
       {
         p = subGraph[treeDepth][i];
         CompDeg(p);
       }

	return minPoint;
}

static PT      *
MinDegOnLevelRCM(level)
	INT             level;
{
	minDeg = actTriang->noOfPoints;
	compLevel = level;
	minPoint = nil;

    ApplyP(CompDeg, all);   

	return minPoint;
}


static PT      *
FindRootRCM(root)
	PT             *root;
{
    INT             noOfLevels, newNoOfLevels;

    newNoOfLevels = RootLsRCM(root);
    if (treeSize == newNoOfLevels)
       return root;

    do {
        noOfLevels = newNoOfLevels;
        root = MinDegOnLevelRCM(noOfLevels);
        newNoOfLevels = RootLsRCM(root);

        if (treeSize == newNoOfLevels)
           return root;

     } while (newNoOfLevels > noOfLevels);
 
    return root;
}

static PT      *
FindRootND(root)
	PT             *root;
{
    INT             noOfLevels, newNoOfLevels;

    newNoOfLevels = RootLS(root);
    if (treeSize == newNoOfLevels)
       return root;

    do {
        noOfLevels = newNoOfLevels;
        root = MinDegOnLevel(noOfLevels);
        newNoOfLevels = RootLS(root);

        if (treeSize == newNoOfLevels)
           return root;

     } while (newNoOfLevels > noOfLevels);
 
    return root;
}



static void 
ChangeIndex(p1, p2)
	PT             *p1, *p2;
{
	INT             index1, index2;

	index1 = p1->index;
	index2 = p2->index;
	p1->index = index2;
	p2->index = index1;
    pointOfNumber[index1] = p2;
    pointOfNumber[index2] = p1;

	return;
}

static INT 
TestNeighbors(p, level)
	PT             *p;
	INT             level;
{
	INT             i;

	for (i = 0; i < (p->noOfNeighbors); i++)
		if (((p->neighbors)[i].p)->l == level)
			return true;

	return false;
}

static PT      *
TakePoint(levelAt, indexAdr)
	INT            levelAt, *indexAdr;
{
    INT   index = *indexAdr;
	PT             *p;

	if (levelAt <= 2)
		while (index < subGraphNo[treeDepth-1])
          {
            p = subGraph[treeDepth-1][index];
			if ( p->l > 0 )
              {
				p->l = 0;
                *indexAdr = index;
				return p;
              }
			index++;
          }
	else
		while (index < subGraphNo[treeDepth-1]) 
          {
            p = subGraph[treeDepth-1][index];
			if ( p->l == levelAt )
				if (TestNeighbors(p, levelAt + 1)) 
                  {
					p->l = 0;
                    *indexAdr = index;
					return p;
                  }
			index++;
         }

    *indexAdr = index;
	return nil;

}




static void 
NumberThePoint(p)
	PT             *p;
{
	PT             *pChange;

	pChange = pointOfNumber[lastNo];
	ChangeIndex(p, pChange);
	lastNo--;
	p->mark = false;

	return;
}

static INT 
Seperate(root)
   PT             *root;
{
   PT       *p;
   INT      noOfPoints = actTriang->noOfPoints, sep, no, i;

   sep = 0;    
   root = FindRootND(root);

   levelAt = (levelNo + 1) / 2;

   if (levelAt <= 2)
     {
      for (i=0; i<subGraphNo[treeDepth]; i++)
        {
          p = subGraph[treeDepth][i];
          if ( p->l > 0 )
            {
              p->l = 0;
              NumberThePoint(p);
              sep++;
            }
         }
 /*      printf("end path, sep = %d\n",sep);  */
      }
      else
      {
        treeDepth++;
        subGraph[treeDepth] = (PT**)ZIBAlloc( (long)noOfPoints*sizeof(PT*) );
        if (subGraph[treeDepth] == nil ) printf("subGraph = nil");

        no = -1;
        for (i=0;i<subGraphNo[treeDepth-1];i++)
          {  
            p = subGraph[treeDepth-1][i];
            if ( p->l >= levelAt )
              {
               no++;
               subGraph[treeDepth][no] = p;
               (p->counter) = treeDepth;

               if ( p->l == levelAt )
                 if (TestNeighbors(p, levelAt + 1)) 
                   {
                     p->l = 0;
                     NumberThePoint(p);
                     sep++;
                   }
               }
          }
          subGraphNo[treeDepth] = no+1;
  /*        printf(" in path, sep %d\n",sep);    */
       }
	return sep;
}



INT 
NestedDissection()
{
	INT             noOfPoints, sep, i;
	PT              *p, *root;
 
    if (!SparsePrepare(DEGREE))
         return false;

    noOfPoints = actTriang->noOfPoints;

    pointOfNumber = (PT**)ZIBAlloc( (long)noOfPoints*sizeof(PT*) );
    if (pointOfNumber == nil ) printf("points = nil");
    p = actTriang->firstPoint;
    while (p != nil)
      {
        pointOfNumber[p->index] = p;
        p =p->next;
      }

    subGraph = (PT***)ZIBAlloc( (long)noOfPoints*sizeof(PTR**) );
    if (subGraph == nil ) printf("subGraph = nil");
    for (i=0; i<noOfPoints; i++)
        subGraph[i] = nil;


    levelStructNext = (PT**)ZIBAlloc( (long)noOfPoints*sizeof(PT*) );
    if (levelStructNext == nil ) printf("levelStructNext = nil");
    structNextNo = -1;

    levelStructLast = (PT**)ZIBAlloc( (long)noOfPoints*sizeof(PT*) );
    if (levelStructLast == nil ) printf("levelStructLast = nil");
    structLastNo = -1;

    ApplyP(Mark, all);
    sep = 0;

    treeDepth  = 0;
    subGraphNo[0] = noOfPoints;
    subGraph[0] = (PT**)ZIBAlloc( (long)noOfPoints*sizeof(PT*) );
    if (subGraph[0] == nil ) printf("subGraph = nil");

    p = actTriang->firstPoint;
    while (p != nil)
      {
        subGraph[0][p->index] = p;
        p =p->next;
      }

    ApplyP(SetTreeDepthZero,all);

    noOfPoints = actTriang->noOfPoints;
	lastNo = noOfPoints - 1;

    
	while (sep < noOfPoints)
      {
        root = nil;
        while (root == nil)
          {
           for (i=0; i<subGraphNo[treeDepth]; i++)
             {
               p = subGraph[treeDepth][i];
               if ( p->mark ) 
                  {
                   root = p;
                   break;
                  }
              }
           if (root==nil) 
             {
              if (subGraph[treeDepth] != nil) ZIBFree((PTR)subGraph[treeDepth]);
              subGraph[treeDepth] = nil;
              treeDepth--;
             }
          }  
        sep += Seperate(root);
      }


   if (subGraph!=nil)        ZIBFree((PTR)subGraph);
   if (pointOfNumber!=nil)   ZIBFree((PTR)pointOfNumber);
   if (levelStructLast!=nil) ZIBFree((PTR)levelStructLast);
   if (levelStructNext!=nil) ZIBFree((PTR)levelStructNext);


    /*  Verlistung der Punkte entsprechend neuer Indizierung   */
    if(!RenumberPts())
       printf(" error in Renumber Points \n");

    /*   Sortieren der Nachbarn nach index   */
    if (!SparseCompNeighIndex()) return false;


	return true;
}

static INT 
NumberTheNeighbors(p)
	PT             *p;
{
	INT             i, num;
	PT             *pNeigh;

	num = 0;
	for (i = 0; i < (p->noOfNeighbors); i++) {
		pNeigh = (p->neighbors[i]).p;
		if (pNeigh->mark) {
			NumberThePoint(pNeigh);
			num++;
		}
	}

	return num;
}

INT RenumberPts()
{
   INT   noOfPoints=actTriang->noOfPoints,
         i;
   PT    *first = actTriang->firstPoint, *p;
   PT    **pp;

   pp = (PT**)ZIBAlloc( (long)noOfPoints*sizeof(PT*) );
   if (pp == nil ) printf("pp = nil");

   p = first;
   while (p != nil)
     {
      pp[p->index] = p;
      p =p->next;
     }


   actTriang->firstPoint = pp[0];
   p = pp[0];
   p->next = pp[1];
   p->last = nil;
   actTriang->lastPoint  = pp[noOfPoints-1];
   p = pp[noOfPoints-1];
   p->next = nil;
   p->last = pp[noOfPoints-2];
 
   for (i=1; i<noOfPoints-1; i++)
      {
       p = pp[i];
       p->next = pp[i+1];
       p->last = pp[i-1];
      }

   actTriang->lastDirectPoint  = nil;

   if (pp!=nil) ZIBFree((PTR)pp);


   return true;
}



INT 
ReverseCuthillMcKee()
{
   INT       noOfPoints, num, i;
   PT        *p, *root;
   PT        *first = actTriang->firstPoint;
   NEIGHBOR  *np;

   if (!SparsePrepare(DEGREE))
		return false;

   noOfPoints = actTriang->noOfPoints;
   pointOfNumber = (PT**)ZIBAlloc( (long)noOfPoints*sizeof(PT*) );
   if (pointOfNumber == nil ) printf("points = nil");
 
   p = first;
   while (p != nil)
     {
      pointOfNumber[p->index] = p;
/*  
      printf("index %d \n",p->index);    
      np =p->neighbors;
      for (i=0;i<p->noOfNeighbors;i++)
         printf("         %d \n",np[i].p->index);
*/
      p =p->next;
     }
  
    levelStructNext = (PT**)ZIBAlloc( (long)noOfPoints*sizeof(PTR) );
    if (levelStructNext == nil ) printf("levelStructNext = nil");
    structNextNo = -1;

    levelStructLast = (PT**)ZIBAlloc( (long)noOfPoints*sizeof(PTR) );
    if (levelStructLast == nil ) printf("levelStructLast = nil");
    structLastNo = -1;


    ApplyP(Mark, all);
    noOfPoints = actTriang->noOfPoints;
    if (noOfPoints <= 1)
       return true;
    lastNo = noOfPoints - 1;
    p = first;
    root = FindRootRCM(p);
	NumberThePoint(root);
    num = 1;
	for (i = noOfPoints - 1; i >= 0; i--) 
      {
		p = pointOfNumber[i];
		num += NumberTheNeighbors(p);
		if (num >= noOfPoints)
			break;
      }

    if (pointOfNumber!=nil)   ZIBFree((PTR)pointOfNumber);
    if (levelStructLast!=nil) ZIBFree((PTR)levelStructLast);
    if (levelStructNext!=nil) ZIBFree((PTR)levelStructNext);


    /*  Verlistung der Punkte entsprechend neuer Indizierung   */
     if(!RenumberPts())
      printf(" error in Renumber Points \n") ;

    /*   Sortieren der Nachbarn nach index   */
    if (!SparseCompNeighIndex()) return false;

	return true;
}


INT 
NoSorting()
{
	if (!SparsePrepare(INDEX))
		return false;

	return true;
}


static INT 
PrintPData(p)
	PT             *p;
{
	sprintf(globBuf, "%3d. Point has index %3d (mark: %1d)\n", count, p->index, p->mark);
	ZIBStdOut(globBuf);
	count++;

	return true;
}

static void 
ServicePrint()
{
	count = 0;
	ApplyP(PrintPData, all);

	return;
}

  
