/*
C
C  _______________________________________________________________
C
C*   Licence
C    =======
C
C    You may use or modify this code for your own non commercial
C    purposes for an unlimited time. 
C    In any case you should not deliver this code without a special 
C    permission of ZIB.
C    In case you intend to use the code commercially, we oblige you
C    to sign an according licence agreement with ZIB.
C
C
C  _______________________________________________________________
C
*/


#include "math.h"

#include "kask.h"
#include "kaskcmd.h"
#include "kasktri.h"
#include "kaskass.h"
#include "kasksol.h"


INT StdF();
INT StdObstacle();
INT StdCauchyF();
INT StdDirichF(); 

 


static REAL	*peak=nil;
static char	*peakNames[] = {"peakX","peakY","peakZ","peakSize"};

static PROBLEM initProblem =
  {
    "std",
	(PROC)StdF, 
    nil,
	(PROC)StdCauchyF,
	(PROC)StdDirichF,
	nil,
	(PROC)NumAss,
	ZERO,
	true,
	nil
  };


PROBLEM *actProblem = & initProblem;



	/*		Example: The Model Problem		*/

static INT StdF(x, y, z, class,fVals,selOfVals)
	REAL x, y, z;
	INT class, selOfVals;
	REAL fVals[];
  {
	fVals[0]	=  1.0;		/*	pX	*/
	fVals[1]	=  1.0;		/*	pY	*/
	fVals[2]	=  1.0;		/*	pZ	*/
	fVals[3]	=  0.0;		/*	q	*/
	fVals[4]	= -1.0;		/*	g	*/
    return true;
  }
 
static INT StdObstacle(x, y, z, class,obstVals,selOfVals)
	REAL x, y, z;
	INT class, selOfVals;
	REAL obstVals[];
  {
	obstVals[0]	=  -1.0;		/*	lower obstacle	*/
	obstVals[1]	=   2.0;		/*	upper obstacle	*/
    return true;
  }

static INT StdCauchyF(x, y, z, class,cauVals, selOfVals)
	REAL x, y, z;
	INT class, selOfVals;
	REAL cauVals[];
  {
	cauVals[0]	= 0.0;		/*	Sigma	*/
	cauVals[1]	= 0.0;		/*	Xi	*/
    return true;
  }
 

static INT StdDirichF(x, y, z, class,dirVals)
	REAL x, y, z;
	INT class;
	REAL dirVals[];
  {
	dirVals[0] = 0.0;
    return true ;
  }


	/*		Example: The Polynom Problem		*/

static INT PolyF(x, y, z, class,fVals,selOfVals)
	REAL x, y, z;
	INT class, selOfVals;
	REAL fVals[];
  {
    REAL fx, fy, fz, fxy, fxz, fyz, fxyz;

    fx   = x*(x-1);
    fy   = y*(y-1);
    fz   = z*(z-1);
    fxy  = fx*fy;
    fxz  = fx*fz;
    fyz  = fy*fz;
    fxyz = TWO *(fyz + fxz + fxy);


	fVals[0]	=  1.0;     /*	pX	*/
	fVals[1]	=  1.0;     /*	pY	*/
	fVals[2]	=  1.0;     /*	pZ	*/
	fVals[3]	=  0.0;     /*	q	*/
	fVals[4]	= -fxyz;    /*	g	*/
    return true;
  }
 
static INT PolyDirichF(x, y, z, class,dirVals)
	REAL x, y, z;
	INT class;
	REAL dirVals[];
  {
    REAL fx, fy, fz;

    fx   = x*(x-1);
    fy   = y*(y-1);
    fz   = z*(z-1);

    dirVals[0] = fx*fy*fz;                /*   U(x,y,z)    */

    return true ;
  }

static INT PolySol(x, y, z,class, solVals,selOfVals)
	REAL x, y, z, solVals[];
	INT class, selOfVals;
  {
    REAL fx, fy, fz;

    fx   = x*(x-1);
    fy   = y*(y-1);
    fz   = z*(z-1);

    solVals[0] = fx*fy*fz;                /*   U(x,y,z)    */
    solVals[1] = (2.0*x-ONE)*fy*fz;       /*   Ux(x,y,z)   */
    solVals[2] = (2.0*y-ONE)*fx*fz;       /*   Uy(x,y,z)   */
    solVals[3] = (2.0*z-ONE)*fx*fy;       /*   Uz(x,y,z)   */
    return true;    
  }


	/*		Example:  Polynom of degree 3		*/

static INT Deg3F(x, y, z, class,fVals,selOfVals)
	REAL x, y, z;
	INT class, selOfVals;
	REAL fVals[];
  {
    REAL fxyz;

    fxyz = 6.00000000000 *(x + y + z);


    fVals[0]	=  1.0;     /*	pX	*/
    fVals[1]	=  1.0;     /*	pY	*/
    fVals[2]	=  1.0;     /*	pZ	*/
    fVals[3]	=  0.0;     /*	q	*/
    fVals[4]	= -fxyz;    /*	g	*/

    return true;
  }
 
static INT Deg3DirichF(x, y, z, class,dirVals)
	REAL x, y, z;
	INT class;
	REAL dirVals[];
  {
    REAL fx, fy, fz;

    fx   = x*x*x;
    fy   = y*y*y;
    fz   = z*z*z;

    dirVals[0] = fx+fy+fz;                /*   U(x,y,z)    */

    return true ;
  }

static INT Deg3Sol(x, y, z,class, solVals,selOfVals)
	REAL x, y, z, solVals[];
	INT class, selOfVals;
  {
    REAL fx, fy, fz;

    fx   = x*x*x;
    fy   = y*y*y;
    fz   = z*z*z;

    solVals[0] = fx+fy+fz;          /*   U(x,y,z)    */
    solVals[1] = THREE*x*x;         /*   Ux(x,y,z)   */
    solVals[2] = THREE*y*y;         /*   Uy(x,y,z)   */
    solVals[3] = THREE*z*z;         /*   Uz(x,y,z)   */
    return true;    
  }

	/*		Example:  Polynom of degree 12		*/

static INT Deg12F(x, y, z, class,fVals,selOfVals)
	REAL x, y, z;
	INT class, selOfVals;
	REAL fVals[];
  {
    REAL fx, fy, fz, fxyz;



    fVals[0]	=  1.0;     /*	pX	*/
    fVals[1]	=  1.0;     /*	pY	*/
    fVals[2]	=  1.0;     /*	pZ	*/
    fVals[3]	=  0.0;     /*	q	*/

    fx = x*x*x*x;
    fy = y*y*y*y;
    fz = z*z*z*z;

    fxyz = 12.000000000 *(x*x*fy*fz + y*y*fx*fz + z*z*fx*fy);

    fVals[4]	= -fxyz;    /*	g	*/

    return true;
  }
 
static INT Deg12DirichF(x, y, z, class,dirVals)
	REAL x, y, z;
	INT class;
	REAL dirVals[];
  {
    REAL fx, fy, fz;

    fx = x*x*x*x;
    fy = y*y*y*y;
    fz = z*z*z*z;

    dirVals[0] = fx*fy*fz;                /*   U(x,y,z)    */

    return true ;
  }

static INT Deg12Sol(x, y, z,class, solVals,selOfVals)
	REAL x, y, z, solVals[];
	INT class, selOfVals;
  {
    REAL fx, fy, fz;

    fx = x*x*x*x;
    fy = y*y*y*y;
    fz = z*z*z*z;

    solVals[0] = fx*fy*fz;              /*   U(x,y,z)    */
    solVals[1] = 3.0*x*x*fy*fz;         /*   Ux(x,y,z)   */
    solVals[2] = 3.0*y*y*fx*fz;         /*   Uy(x,y,z)   */
    solVals[3] = 3.0*z*z*fx*fy;         /*   Uz(x,y,z)   */
    return true;    
  }


	/*		Example:  Linear Problem		*/

static INT LinF(x, y, z, class,fVals,selOfVals)
	REAL x, y, z;
	INT class, selOfVals;
	REAL fVals[];
  {
	fVals[0]	=  1.0;     /*	pX	*/
	fVals[1]	=  1.0;     /*	pY	*/
	fVals[2]	=  1.0;     /*	pZ	*/
	fVals[3]	=  0.0;     /*	q	*/
	fVals[4]	=  0.0;    /*	g	*/
    return true;
  }
 
static INT LinDirichF(x, y, z, class,dirVals)
	REAL x, y, z;
	INT class;
	REAL dirVals[];
  {
    dirVals[0] = x + y + z;                 

    return true ;
  }


static INT LinCauchyF(x,y,z,class,cauVals,selOfVals)
   REAL x, y, z;
   INT class, selOfVals;
   REAL cauVals[];
{
   switch (class)
     {
      case 1: 
       cauVals[0] = 1.0;		  /*   Sigma   */
       cauVals[1] = -ONE + x + y + z;     /*   Xi      */
       break;
      case 2: 
       cauVals[0] = 0.0;		/*   Sigma   */
       cauVals[1] = ONE;		/*   Xi   */
       break;
      case 3: 
       cauVals[0] = 0.0;		/*	Sigma	*/
       cauVals[1] = -ONE;		/*	Xi	*/
       break;
      case 4: 
       cauVals[0] = 0.0;		/*	Sigma	*/
       cauVals[1] = ONE;		/*	Xi	*/
       break;
      case 5: 
       cauVals[0] = 0.0;		/*	Sigma	*/
       cauVals[1] = -ONE;		/*	Xi	*/
       break;
      case 6: 
       cauVals[0] = 0.0;		/*	Sigma	*/
       cauVals[1] = ONE;		/*	Xi	*/
       break;

     }
   return true;   
  }

static INT LinSol(x, y, z,class, solVals,selOfVals)
	REAL x, y, z, solVals[];
	INT class, selOfVals;
  {
    solVals[0] = x + y + z;                
    solVals[1] = ONE;                  /*   Ux(x,y,z)   */
    solVals[2] = ONE;                  /*   Uy(x,y,z)   */
    solVals[3] = ONE;                  /*   Uz(x,y,z)   */
    return true;    
  }


	/*		Example:  Quadratic Problem		*/

static INT QuadF(x, y, z, class,fVals,selOfVals)
	REAL x, y, z;
	INT class, selOfVals;
	REAL fVals[];
  {
	fVals[0]	=  1.0;     /*	pX	*/
	fVals[1]	=  1.0;     /*	pY	*/
	fVals[2]	=  1.0;     /*	pZ	*/
	fVals[3]	=  0.0;     /*	q	*/
	fVals[4]	=  -6.0;    /*	g	*/
    return true;
  }
 
static INT QuadDirichF(x, y, z, class,dirVals)
	REAL x, y, z;
	INT class;
	REAL dirVals[];
  {
    dirVals[0] = x*x+y*y+z*z;                 

    return true ;
  }

static INT QuadCauchyF(x,y,z,class,cauVals,selOfVals)
   REAL x, y, z;
   INT class, selOfVals;
   REAL cauVals[];
{
   switch (class)
     {
      case 1: 
       cauVals[0] = ONE;                /*   Sigma   */
       cauVals[1] = -TWO*x + x*x+y*y+z*z;             /*   Xi      */
       break;
      case 2: 
       cauVals[0] = ONE;		/*   Sigma   */
       cauVals[1] = -TWO*z + x*x+y*y+z*z;		/*   Xi      */
       break;
      case 3: 
       cauVals[0] = ONE;		/*   Sigma   */
       cauVals[1] = -TWO*y + x*x+y*y+z*z;		/*   Xi      */
       break;
      case 4: 
       cauVals[0] = ONE;		/*   Sigma   */
       cauVals[1] = TWO*x + x*x+y*y+z*z;		/*   Xi      */
       break;
      case 5: 
       cauVals[0] = ONE;                /*   Sigma   */
       cauVals[1] = TWO*z + x*x+y*y+z*z;              /*   Xi      */
       break;
      case 6: 
       cauVals[0] = ONE;                /*   Sigma   */
       cauVals[1] = TWO*y + x*x+y*y+z*z;              /*   Xi	     */
       break;

     }
   return true;   
  }

static INT QuadSol(x, y, z,class, solVals,selOfVals)
	REAL x, y, z, solVals[];
	INT class, selOfVals;
  {
    solVals[0] = x*x+y*y+z*z;                
    solVals[1] = TWO*x;                  /*   Ux(x,y,z)   */
    solVals[2] = TWO*y;                  /*   Uy(x,y,z)   */
    solVals[3] = TWO*z;                  /*   Uz(x,y,z)   */
    return true;    
  }



/*		functions for test peak 2d			*/

static INT Peak2dF(x, y, z, class,fVals,selOfVals)
	REAL x, y, z;
	INT class, selOfVals;
	REAL fVals[];
  {
	REAL a_para, b_para, peak;
   	REAL xm1,xm1x,ym1,ym1y,xma,ymb;
  	REAL sum,divisor;

	fVals[0] =  1.0;		/*	pX	*/
	fVals[1] =  1.0;		/*	pY	*/
	fVals[2] =  1.0;		/*	pZ	*/
	fVals[3] =  0.0;		/*	q	*/

    a_para = 0.5;
    b_para = 0.117;
    peak = 100.0;
    xm1  = x-1.0;
    xm1x = x*xm1;
    ym1  = y-1.0;
    ym1y = y*ym1;
    xma  = x-a_para;
    ymb  = y-b_para;

   	divisor = exp(peak*(xma*xma + ymb*ymb));
  	sum = 2.0*(xm1x+ym1y);
   	sum = sum - 4.0*peak*(xm1x*ym1y + xm1*xma*ym1y + x*xma*ym1y);
   	sum = sum + 4.0*peak*peak*(xm1x*xma*xma*ym1y + xm1x*ym1y*ymb*ymb);
   	sum = sum - 4.0*peak*(xm1x*ym1*ymb + xm1*x*y*ymb);
	fVals[4] = -sum/divisor;	/*	g	*/

    return true;
  }
 
  


static INT Peak2dDirichF(x, y, z,class,dirVals)
	REAL x, y, z;
	INT class;
	REAL dirVals[];
  {
	REAL a_para, b_para, peak;
	REAL xma, ymb, efac;

	a_para = 0.5;
	b_para = 0.117;
    peak = 100.0;
	xma  = x-a_para;
	ymb  = y-b_para;

    efac = exp(-peak*(xma*xma + ymb*ymb));
	dirVals[0] =  x*(x-1.0)*y*(y-1.0)*efac;  
    return true;
  }

static INT Peak2dSol(x, y, z, class, solVals,selOfVals)
   REAL x, y, z, solVals[];
   INT class, selOfVals;
{
   REAL a_para, b_para, peak;
   REAL xma, ymb, efac;

   a_para = 0.5;
   b_para = 0.117;
   peak = 100.0;
   xma  = x-a_para;
   ymb  = y-b_para;

   efac = exp(-peak*(xma*xma + ymb*ymb));
   solVals[0] =  x*(x-1.0)*y*(y-1.0)*efac;  
   solVals[1] = ( (x-1.0)*(y-1.0)*y + x*y*(y-1.0) -
                  2.0*peak*(x-1.0)*x*xma*y*(y-1.0) ) * efac;  
   solVals[2] = ( (x-1.0)*(y-1.0)*x + x*y*(x-1.0) -
                  2.0*peak*(x-1.0)*x*ymb*y*(y-1.0) ) * efac; 
   solVals[3] =  ZERO; 

   return true;  
}



/*		functions for test peak 3d			*/

static INT Peak3dF(x, y, z, class,fVals,selOfVals)
   REAL x, y, z;
   INT class, selOfVals;
   REAL fVals[];
{
   REAL a, b, c, p;
   REAL xm1,ym1,zm1,xm1x,ym1y,zm1z,xma,ymb,zmc;
   REAL sum,divisor;

   fVals[0] =  1.0;                     /*      pX      */
   fVals[1] =  1.0;		/*      pY      */
   fVals[2] =  1.0;		/*      pZ      */
   fVals[3] =  0.0;		/*      q       */

   if ((selOfVals == P_RHS) || (selOfVals == P_ALL))
     {
       a = peak[0];
       b = peak[1];
       c = peak[2];
       p = peak[3];
       xma  = x-a;
       ymb  = y-b;
       zmc  = z-c;

       xm1  = x-1.0;
       ym1  = y-1.0;
       zm1  = z-1.0;
       xm1x = x*xm1;
       ym1y = y*ym1;
       zm1z = z*zm1;

       divisor = exp(p*(xma*xma + ymb*ymb + zmc*zmc));

       sum =       2.0*(xm1x*ym1y+xm1x*zm1z+ym1y*zm1z);
       sum = sum - 6.0*p*(xm1x*ym1y*zm1z);
       sum = sum - 4.0*p*(xm1*xma*ym1y*zm1z + x*xma*ym1y*zm1z);
       sum = sum + 4.0*p*p*(xm1x*xma*xma*ym1y*zm1z);
       sum = sum - 4.0*p*(xm1x*ym1*ymb*zm1z + xm1x*y*ymb*zm1z);
       sum = sum + 4.0*p*p*(xm1x*ym1y*ymb*ymb*zm1z);
       sum = sum - 4.0*p*(xm1x*ym1y*zm1*zmc + xm1x*ym1y*z*zmc);
       sum = sum + 4.0*p*p*(xm1x*ym1y*zm1z*zmc*zmc);

       fVals[4] = -100.0*sum/divisor;      /*   g   */
     }
   return true;
}
 
  
static INT Peak3dDirichF(x, y, z,class,dirVals)
   REAL x, y, z;
   INT class;
   REAL dirVals[];
{
   REAL a, b, c, p;
   REAL xma, ymb, zmc, efac;

   a = peak[0];
   b = peak[1];
   c = peak[2];
   p = peak[3];
   xma  = x-a;
   ymb  = y-b;
   zmc  = z-c;

   efac = exp(-p*(xma*xma + ymb*ymb + zmc*zmc));
   dirVals[0] =  x*(x-1.0)*y*(y-1.0)*z*(z-1.0)*efac;  

/*
   dirVals[0] =  0.0; 
*/ 
   return true;
}

static INT Peak3dSol(x, y, z, class, solVals,selOfVals)
   REAL x, y, z, solVals[];
   INT class, selOfVals;
{
   REAL a, b, c, p;
   REAL xma, ymb, zmc, efac;

   a = peak[0];
   b = peak[1];
   c = peak[2];
   p = peak[3];
   xma  = x-a;
   ymb  = y-b;
   zmc  = z-c;

   efac = exp(-p*(xma*xma + ymb*ymb + zmc*zmc));
   solVals[0] =  100.0*x*(x-1.0)*y*(y-1.0)*z*(z-1.0)*efac;  
/*
   solVals[1] = ( (x-1.0)*(y-1.0)*y + x*y*(y-1.0) -
                  2.0*p*(x-1.0)*x*xma*y*(y-1.0) ) * efac;  
   solVals[2] = ( (x-1.0)*(y-1.0)*x + x*y*(x-1.0) -
                  2.0*p*(x-1.0)*x*ymb*y*(y-1.0) ) * efac; 
   solVals[3] =  ZERO; 
*/

   return true;  
}




/*	Example: Discontinuous Dirichlet Boundary Conditions	*/

static INT RandUnF(x, y, z, class,fVals,selOfVals)
	REAL x, y, z;
	INT class, selOfVals;
	REAL fVals[];
  {
    fVals[0]	=  1.0;		/*  pX	*/
    fVals[1]	=  1.0;		/*  pY	*/
    fVals[2]	=  1.0;		/*  pZ	*/
    fVals[3]	=  0.0;		/*  q	*/
    fVals[4]	=  0.0;		/*  g	*/
    return true;
  }
 
static INT RandUnDirichF(x, y, z, class,dirVals)
    REAL x, y, z;
    INT class;
    REAL dirVals[];
  {
   REAL eps =1.e-6;

    dirVals[0] = 0.0;
    if ((x<HALF-eps)&&(y<HALF-eps)&&(z<HALF-eps))
        dirVals[0] = ONE;
    else
    if ((fabs(x-HALF)<eps)&&(fabs(y)<eps)&&(z<=HALF)) dirVals[0] = HALF;
    else
    if ((fabs(x)<eps)&&(fabs(y-HALF)<eps)&&(z<=HALF)) dirVals[0] = HALF;
    else
    if ((x<=HALF)&&(fabs(y)<eps)&&(fabs(z-HALF)<eps)) dirVals[0] = HALF;
    else
    if ((fabs(x)<eps)&&(y<=HALF)&&(fabs(z-HALF)<eps)) dirVals[0] = HALF;
    else
    if ((fabs(x-HALF)<eps)&&(y<=HALF)&&(fabs(z)<eps)) dirVals[0] = HALF;
    else
    if ((x<=HALF)&&(fabs(y-HALF)<eps)&&(fabs(z)<eps)) dirVals[0] = HALF;

    return true ;
  }

/*	Example: Discontinuous Coefficients	*/

static INT RegpF(x,y,z,class,fVals,selOfVals)
   REAL x, y, z;
   INT class, selOfVals;
   REAL fVals[];
{
					/*	pX	*/
       fVals[0] = 1.0e6;
       if (class == 1) fVals[0] = ONE;

       fVals[1]	=  fVals[0];	/*  pY	*/
       fVals[2]	=  fVals[0];	/*  pZ	*/
       fVals[3]	=  0.0;	        /*  q	*/
       fVals[4]	= -1.0;	        /*  g	*/

   return true;
}
 


/*		Example: mixed boundary values			*/

static INT KreisF(x,y,z,class,fVals,selOfVals)
	REAL x, y, z;
	INT class, selOfVals;
	REAL fVals[];
  {
	fVals[0] =  1.0;		/*	pX	*/
	fVals[1] =  1.0;		/*	pY	*/
	fVals[2] =  1.0;		/*	pZ	*/
	fVals[3] =  0.0;		/*	q	*/
	fVals[4] = -4.0;		/*	g	*/

    return true;
  }
 
  
static INT KreisCauchyF(x,y,z,class,cauVals,selOfVals)
	REAL x, y, z;
	INT class, selOfVals;
	REAL cauVals[];
  {
	REAL r;
	r = sqrt(x*x + y*y);
	cauVals[0] = 0.0;		/*	Sigma	*/
	cauVals[1] = 2.0*r;		/*	Xi	*/
     return true;   
  }


static INT KreisDirichF(x, y, z,class,dirVals)
	REAL x, y, z;
	INT class;
	REAL dirVals[];
  {
	dirVals[0] = 1.0;
    return true;
  }

static INT KreisSol(x, y, z,class, solVals,selOfVals)
	REAL x, y, z, solVals[];
	INT class, selOfVals;
  {
    solVals[0] = x*x+y*y;    
    solVals[1] = 2.0*x;    
    solVals[2] = 2.0*y;    
    solVals[3] = 0.0;    
    return true;    
  }

/*         singularities        */
static int Bank1F(x, y,z, class,fVals,selOfVals)
   REAL x, y, z;
   INT class, selOfVals;
   REAL fVals[];
{
   fVals[0]	=  1.0;	      /*   pX   */
   fVals[1]	=  1.0;	      /*   pY   */
   fVals[2]	=  1.0;	      /*   pZ   */
   fVals[3]	=  0.0;	      /*   q   */
   fVals[4]	=  0.0;	      /*   g   */
   return true;
}

static REAL arccos(x)
   REAL x;
{
   REAL y=(fabs(x)<SMALL)?REALPI2:atan(sqrt(ONE-x*x)/x);
   if (x<=-SMALL) y = REALPI+y;
   return y;
}

 
static int Bank1DirichF(x, y, z, class,dirVals)
   REAL x, y, z;
   INT class;
   REAL dirVals[];
{  
   REAL r = sqrt(x*x+y*y), phi;
   phi = (r>0)?arccos(x/r):ZERO;
   if (y<ZERO) phi = TWO*REALPI-phi;
   r =pow(r,2.0/3.0);
   dirVals[0]= r*sin(2.0/3.0*phi);

   return true ;
}

static int Bank1Sol(x, y, z, class, solVals,selOfVals)
   REAL x, y, z, solVals[];
   int class, selOfVals;
{
   REAL r = sqrt(x*x+y*y), phi;
   REAL dr, dphi, drx, dry, dphix, dphiy, rp1, rp2, r2;

   phi = (r>0)?arccos(x/r):ZERO;
   if (y<ZERO) phi = TWO*REALPI-phi;
   rp1 =pow(r,2.0/3.0);
   rp2 =pow(r,1.0/3.0);
   solVals[0]= rp1*sin(2.0/3.0*phi);

   r2   = r*r;
   dr   = 2.0/(3.0*rp2)*sin(2.0/3.0*phi);
   dphi = 2.0/3.0*rp1*cos(2.0/3.0*phi);
   drx  = x/r;
   dry	 = y/r;
   dphix = -y/r2;
   dphiy = x/r2;
   solVals[1]= dr*drx + dphi*dphix;
   solVals[2]= dr*dry + dphi*dphiy;

   return true;    
}


/*         Hyperthermia        */

static INT HyperF0(x,y,z,class,fVals,selOfVals)
	REAL x, y, z;
	INT class, selOfVals;
	REAL fVals[];
  {
    float absE2;
    float xf, yf,zf;
    REAL sigma, k, rho, w, epsT, cB, rhoB;

    xf = x; 
    yf = y;
    zf = z;

    noOfFCalls++;


    switch (class)
	  {
		case BONE:				
		  k     = 4.36e-3;
		  rho   = 1.30;
		  w     = ZERO;
		  epsT  = 5.61;
		  sigma = 1.40e-3;
		  break;

		case BRAIN:	
		  k     = 6.30e-3;
		  rho   = ONE;
		  w     = 9.00e-3;
		  epsT  = 34.4;	
		  sigma = 7.3e-3;	
		  break;

		case TUMOR:
		  k     = 6.30e-3;
		  rho   = ONE;
		  w     = 9.00e-3;
		  epsT  = 34.4;	
                  sigma = 7.30e-3;
		  break;
	  }

        cB   = 3.5 ;
        rhoB = ONE;

        fVals[0] =  k;                   /*   pX   */
        fVals[1] =  k;                   /*   pY   */
        fVals[2] =  k;                   /*   pZ   */
        fVals[3] =  cB*rhoB*rho*w;       /*   q    */


        if ((selOfVals == P_RHS) || (selOfVals == P_ALL))
          {
/*
            field3_(&xf,&yf,&zf,&absE2);   
	    fVals[4] = sigma*absE2;        
*/
	    fVals[4] = sigma;        

          }

    return true;
  }
 


static INT HyperF1(x,y,z,class,fVals,selOfVals)
	REAL x, y, z;
	INT class, selOfVals;
	REAL fVals[];
  {
    float absE2;
    float xf, yf,zf;
    REAL sigma, k, rho, w, epsT, cB, rhoB;

    xf = x; 
    yf = y;
    zf = z;

    noOfFCalls++;


    switch (class)
	  {
		case BONE:				
		  k     = 4.36e-3;
		  rho   = 1.30;
		  w     = ZERO;
		  epsT  = 5.61;
		  sigma = 1.40e-3;
		  break;

		case BRAIN:	
		  k     = 6.30e-3;
		  rho   = ONE;
		  w     = 9.00e-3;
		  epsT  = 34.4;	
		  sigma = 7.3e-3;	
		  break;

		case TUMOR:
		  k     = 6.30e-3;
		  rho   = ONE;
		  w     = 4.50e-3;
		  epsT  = 34.4;	
                  sigma = 7.30e-3;
		  break;
	  }

        cB   = 3.5 ;
        rhoB = ONE;

        fVals[0] =  k;                   /*   pX   */
        fVals[1] =  k;                   /*   pY   */
        fVals[2] =  k;                   /*   pZ   */
        fVals[3] =  cB*rhoB*rho*w;       /*   q    */


        if ((selOfVals == P_RHS) || (selOfVals == P_ALL))
          {

/*
            field3_(&xf,&yf,&zf,&absE2);   
	    fVals[4] = sigma*absE2;        
*/
	    fVals[4] = sigma;        

          }

    return true;
  }

  
static INT HyperDirichF(x, y, z,class,dirVals)
	REAL x, y, z;
	INT class;
	REAL dirVals[];
  {
	dirVals[0] = ZERO;
    return true;
  }

static INT HyperCauchyF(x,y,z,class,cauVals,selOfVals)
   REAL x, y, z;
   INT class, selOfVals;
   REAL cauVals[];
{
   REAL tAmb, h, k;

   switch (class)
      {
        case 0:				
            h    = 7.44e-4;
            tAmb = -10.8;
            break;

        case 1:	
            h    = 2.67e-4;
            tAmb = -6.0;
            break;

        case 2:
            h    = 6.50e-3;
            tAmb = ZERO;
            break;

        case 3:
            h    = 1.6e-3;
            tAmb = -6.0;
            break;

        case 4:
            h    = 1.6e-3;
            tAmb = -1.0;
            break;

	  }
    
   cauVals[0] = h;             /*   Sigma   */
   cauVals[1] = h*tAmb;        /*   Xi      */
   return true;   
}

/********************************************************/
/*	Problem-Command: Select a special example	*/
/********************************************************/

INT StdProblem(cmd)
  COMMAND *cmd;
  {
	INT index = 0;
	char *tp;

	if (ParsCheck(cmd,0,1)) return false;

	if ((cmd->noOfPars)==1)
	  {
	    tp = (cmd->pars)[1];
	    index = CheckName(&tp, cmd->names, nameClass);
	  }
	if (index==-1)
	  {
	    sprintf(globBuf,"Parameter %s of %s not allowed\n", (cmd->pars)[1], (cmd->pars)[0]);
		ZIBStdOut(globBuf);
	    return false;
	  }

/*		 Presets for Solve		*/
/*	
    if (!CheckPreSets()) return false;    
	actSolve->Direct = nil;
	actSolve->Estimate = nil;
	actSolve->Refine = nil;
	actSolve->Iterate = nil;
*/	
/*	select a special problem		*/

	actProblem->Obstacle    = nil;
	actProblem->Sol         = nil;
	actProblem->NumAss      = NumAss;
	actProblem->class       = ZERO;
	actProblem->symP        = SYMMETRIC;

	switch (index)
	  {
		case 0:				/*	Std		*/
		  actProblem->F	       = (PROC)StdF;
		  actProblem->CauchyF  = (PROC)StdCauchyF;
		  actProblem->DirichF  = (PROC)StdDirichF;	
		  break;

		case 1:				/*	linear		*/
	  	  actProblem->F	       = (PROC)LinF;
		  actProblem->CauchyF  = (PROC)LinCauchyF;
		  actProblem->DirichF  = (PROC)LinDirichF;	
		  actProblem->Sol      = (PROC)LinSol;	
		  break;

		case 2:	                       /*   quadratic   */
	  	  actProblem->F	       = (PROC)QuadF;
		  actProblem->CauchyF  = (PROC)QuadCauchyF;
		  actProblem->DirichF  = (PROC)QuadDirichF;	
		  actProblem->Sol      = (PROC)QuadSol;	
		  break;

		case 3:                            /*   Polynom    */
	  	  actProblem->F	       = (PROC)PolyF;
		  actProblem->CauchyF  = (PROC)StdCauchyF;
		  actProblem->DirichF  = (PROC)PolyDirichF;	
		  actProblem->Sol      = (PROC)PolySol;	
		  break;


		case 4:				/*   peak3D	  */
                  if (peak == nil)
                    {
                     peak = (REAL*)
                            NewParamList("tompeak",4,
                                         sizeof(REAL),1,peakNames,nil,nil);
                     peak[0] = 0.5;
                     peak[1] = 0.5;
                     peak[2] = 0.5;
                     peak[3] = 150.0;
		    }

	  	  actProblem->F	       = (PROC)Peak3dF;
		  actProblem->CauchyF  = (PROC)StdCauchyF;
		  actProblem->DirichF  = (PROC)Peak3dDirichF;	
		  actProblem->Sol      = (PROC)Peak3dSol;	
		  break;

		case 5:				/*   bounddisc   */
	  	  actProblem->F	       = (PROC)RandUnF;
		  actProblem->CauchyF  = (PROC)StdCauchyF;	
		  actProblem->DirichF  = (PROC)RandUnDirichF;
		  break;

		case 6:	                          /*   region   */
	  	  actProblem->F	       = (PROC)RegpF;
		  actProblem->CauchyF  = (PROC)StdCauchyF;
		  actProblem->DirichF  = (PROC)StdDirichF;
		  break;
             }
/*
	SetBound(R_SOL);
	InitNumAss(N_STD);
*/
	if ((cmd->noOfPars)==1) actProblem->name = (cmd->names)[index];
	return true;
  }

INT InfPrb(cmd)
  COMMAND *cmd;
  {
	sprintf(globBuf,"Problem: current problem is %s \n",actProblem->name);
	ZIBStdOut(globBuf);
	return true;
  }
