/*
C
C  _______________________________________________________________
C
C*   Licence
C    =======
C
C    You may use or modify this code for your own non commercial
C    purposes for an unlimited time. 
C    In any case you should not deliver this code without a special 
C    permission of ZIB.
C    In case you intend to use the code commercially, we oblige you
C    to sign an according licence agreement with ZIB.
C
C
C  _______________________________________________________________
C
*/


#include <ctype.h>

#include "kask.h"
#include "kasktri.h"
#include "kaskcmd.h"
#include "kaskgraph.h"

static INT countP, countE, countT;
static INT noOfGreen1Tds, noOfGreen2Tds, noOfGreen3Tds, noOfWhiteTds;
static INT noOfBoundPts, noOfBoundEds, noOfBoundTrs;

static INT CountTdTypes(t)
   TR3 *t;
{
   if( (t->type) == T_GREEN_1) noOfGreen1Tds++;
   if( ((t->type) == T_GREEN_2A) || ((t->type) == T_GREEN_2B) ) noOfGreen2Tds++;
   if( (t->type) == T_GREEN_3) noOfGreen3Tds++;
   if( (t->type) == T_WHITE  ) noOfWhiteTds++;

   return true;
}

static INT CountPtsTypes(p)
   PT *p;
{
   noOfBoundPts++;
   return true;
}

static INT CountEdsTypes(ed)
   EDG *ed;
{
   noOfBoundEds++;
   return true;
}

static INT CountTrsTypes(t)
   TR2 *t;
{
   if( (t->boundP) != INTERIOR) noOfBoundTrs++;
   return true;
}

INT TriInf(cmd)
  COMMAND *cmd;
{
   INT i = 0, no =0;
   TRIANGULATION *tri = firstTriang;

   if (ParsCheck(cmd,0,0)) return false;
   if (actTriang==nil)
     {
       sprintf(globBuf, "Tri: no triangulation (%s)\n",(cmd->pars)[0]);
       ZIBStdOut(globBuf);
       return false;
     }

   sprintf(globBuf, "Tri: current triangulation '%s' from '%s'\n",
           actTriang->name, actTriang->fileName);
   ZIBStdOut(globBuf);
   sprintf(globBuf, "       %-20.20s: % 4d\n", "noOfPoints",
           actTriang->noOfPoints);
   ZIBStdOut(globBuf);
   sprintf(globBuf, "       %-20.20s: % 4d\n", "noOfEdges",
           actTriang->noOfEdges);
   ZIBStdOut(globBuf);
   sprintf(globBuf, "       %-20.20s: % 4d\n", "noOfTriangles",
           actTriang->noOfTriangles);
   ZIBStdOut(globBuf);
   sprintf(globBuf, "       %-20.20s: % 4d\n", "noOfTetrahedra",
           actTriang->noOfTetrahedra);
   ZIBStdOut(globBuf);
   no =   actTriang->noOfPoints - actTriang->noOfEdges 
        + actTriang->noOfTriangles - actTriang->noOfTetrahedra;
   sprintf(globBuf, "       %-20.20s: % 4d\n", "Euler-characteristic", no);
   ZIBStdOut(globBuf);

   noOfGreen1Tds = 0;
   noOfGreen2Tds = 0;
   noOfGreen3Tds = 0;
   noOfWhiteTds  = 0;
   ApplyT3(CountTdTypes,all);
   sprintf(globBuf, "       %-20.20s: % 4d\n", "noOfWhiteTetrahedra",
           noOfWhiteTds);
   ZIBStdOut(globBuf);
   sprintf(globBuf, "       %-20.20s: % 4d\n", "noOfGreen1Tetrahedra",
           noOfGreen1Tds);
   ZIBStdOut(globBuf);
   sprintf(globBuf, "       %-20.20s: % 4d\n", "noOfGreen2Tetrahedra",
           noOfGreen2Tds);
   ZIBStdOut(globBuf);
   sprintf(globBuf, "       %-20.20s: % 4d\n", "noOfGreen3Tetrahedra",
           noOfGreen3Tds);
   ZIBStdOut(globBuf);

   noOfBoundPts = 0;
   ApplyP(CountPtsTypes,boundHier);
   sprintf(globBuf, "       %-20.20s: % 4d\n", "noOfBoundaryPoints",
           noOfBoundPts);
   ZIBStdOut(globBuf);

   noOfBoundEds = 0;
   ApplyE(CountEdsTypes,boundNodal);
   sprintf(globBuf, "       %-20.20s: % 4d\n", "noOfBoundaryEdges",
           noOfBoundEds);
   ZIBStdOut(globBuf);

   noOfBoundTrs = 0;
   ApplyT(CountTrsTypes,all);
   sprintf(globBuf, "       %-20.20s: % 4d\n", "noOfBoundaryTrs",
           noOfBoundTrs);
   ZIBStdOut(globBuf);


   sprintf(globBuf, "       %-20.20s: % 4d\n", "noOfInitPoints",
           actTriang->noOfInitPoints);
   ZIBStdOut(globBuf);
   sprintf(globBuf, "       %-20.20s: % 4d\n", "noOfInitEdges",
           actTriang->noOfInitEdges);
   ZIBStdOut(globBuf);
   sprintf(globBuf, "       %-20.20s: % 4d\n", "noOfInitTriangles",
           actTriang->noOfInitTriangles);
   ZIBStdOut(globBuf);
   sprintf(globBuf, "       %-20.20s: % 4d\n", "noOfInitTetrahedra",
           actTriang->noOfInitTetrahedra);
   ZIBStdOut(globBuf);
   sprintf(globBuf, "       %-20.20s: % 4d\n", 
           "refLevel", actTriang->refLevel);
   ZIBStdOut(globBuf);
   sprintf(globBuf, "       %-20.20s: % 4d\n", 
           "maxDepth", actTriang->maxDepth);
   ZIBStdOut(globBuf);

   if ((firstTriang==actTriang)&&((actTriang->next)==nil))
     return true;
   ZIBStdOut("Tri: other triangulation\n");
   while (tri!=nil)
        {
         if (i==0) ZIBStdOut("       ");
         else { if ((i%3)==0) { ZIBNL(); ZIBStdOut("       "); } }
         if ( tri != actTriang )
           {
             sprintf(globBuf, "	   %-20.20s\n", tri->name);
             ZIBStdOut(globBuf);
           }
         tri = tri->next;
         i++;
        }
   ZIBNL();

   return true;
}

static INT compString(s1, s2)
  char *s1, *s2;
  {
    char c1 = *s1, c2 = *s2;

	while (c1==c2)
	{
	  if ((c1==NUL)&&(c2==NUL)) return true;
	  s1++;     s2++;
	  c1 = *s1; c2 = *s2;
	  if (isupper(c1)) c1 = tolower(c1);
	  if (isupper(c2)) c2 = tolower(c2);
	}
	return false;
  }

INT TriSel(cmd)
  COMMAND *cmd;
  {
	TRIANGULATION *tri = firstTriang;

	if (ParsCheck(cmd,1,1)) return false;
	if (tri==nil)
	  {
	    sprintf(globBuf, "Tri: no triangulation (%s)\n",(cmd->pars)[0]);
		ZIBStdOut(globBuf);
		return false;
	  }

	while (tri!=nil)
	{
	  if (compString(tri->name,(cmd->pars)[1]))
	    {
		  SelTri(tri);
		  sprintf(globBuf, "Tri: current triangulation set to '%s'\n", 
                          actTriang->name);
		  ZIBStdOut(globBuf);

/*
          if( actGraph != nil )
                 ReCompScale();
*/
		  return true;
		}
	  tri = tri->next;
	}

	sprintf(globBuf, "Tri: can`t find triangulation '%s'\n", (cmd->pars)[1]);
	ZIBStdOut(globBuf);
	return false;
  }

INT TriDel(cmd)
   COMMAND *cmd;
{
   TRIANGULATION *tri = firstTriang;
   char *name;

   if (ParsCheck(cmd,0,1)) return false;
   if (tri==nil)
     {
       sprintf(globBuf, "Tri: no triangulation (%s)\n",(cmd->pars)[0]);
       ZIBStdOut(globBuf);
       return false;
     }

   if ((cmd->noOfPars)==0) tri = actTriang;
     else
       while (tri!=nil)
         {
          if (compString(tri->name,(cmd->pars)[1])) break;
          tri = tri->next;
         }
   if (tri==nil)
     {
      sprintf(globBuf, "Tri: can`t find triangulation '%s'\n", (cmd->pars)[1]);
      ZIBStdOut(globBuf);
      return false;
     }
   name = tri->name;
   CloseTri(&tri);
   sprintf(globBuf, "Tri: triangulation '%s' deleted\n", name);
   ZIBStdOut(globBuf);
   if (actTriang!=nil)
     {
       sprintf(globBuf, "Tri: current triangulation set to '%s'\n",
               actTriang->name);
       ZIBStdOut(globBuf);
     }
   return true;
}

INT CmdResetTriang(cmd)
   COMMAND *cmd;
{
   char *name;

   if (ParsCheck(cmd,0,0)) return false;

   if(!ResetTriang()) return false;;
   sprintf(globBuf, "Tri: actual triangulation reset\n");
   ZIBStdOut(globBuf);

   return true;
}

void PrintTr(s, t)
  char *s;
  TR2 *t;
  {
    INT k;
    EDG **eds = &(t->e1);
    if (t==nil)
	  { sprintf(globBuf,"%s: nil\n", s); ZIBStdOut(globBuf); return; }
    sprintf(globBuf,"%s: (%d,%d,%d) (", s, (t->p1)->index,
			(t->p2)->index, (t->p3)->index);
    ZIBStdOut(globBuf);
	for (k = 0; k<3; k++)
      {
	    sprintf(globBuf,"(%d)-(%d)", (eds[k]->p1)->index,
				(eds[k]->p2)->index);
		ZIBStdOut(globBuf);
	  }
    ZIBStdOut(")");
    if ((t->father)==nil) ZIBStdOut(" init");

    switch(t->type)
        {
          case T_RED:      ZIBStdOut(" red"); break;
          case T_GREEN_1:  ZIBStdOut(" green"); break;
          case T_WHITE:    ZIBStdOut(" white"); break;
        }
    ZIBStdOut("\n");
    return;
  }

INT PrTr(t)
  TR2 *t;
  {
    PrintTr("", t);
    return true;
  }
 
void PrintEdg(s, ed)
  char *s;
  EDG *ed;
  {
    if (ed==nil)
	  { sprintf(globBuf,"%s: nil\n", s); ZIBStdOut(globBuf); return; }
    sprintf(globBuf,"%s: (%d-%d)", s, (ed->p1)->index,
			(ed->p2)->index);
    ZIBStdOut(globBuf);
	switch(ed->type)
        {
          case T_RED:     ZIBStdOut(" red"); break;
          case T_GREEN_1: ZIBStdOut(" green"); break;
          case T_WHITE:   ZIBStdOut(" white"); break;
          case T_NOTHING: ZIBStdOut(" nothing"); break;
        }
    ZIBStdOut("\n");
    return;
  }

INT PrEdg(ed)
  EDG *ed;
  {
    PrintEdg("", ed);;
    return true;
  }

void PrintPt(s, p)
  char *s;
  PT *p;
  {
    sprintf(globBuf,"%s",s);
	ZIBStdOut(globBuf);
    PrPt(p);
    return;
  }

INT PrPt(p)
  PT *p;
  {
    if (p==nil) ZIBStdOut(": nil\n");
    else
	  {
	    sprintf(globBuf, ": %d, (%e,%e), sol=%e\n",
				p->index, p->x, p->y, RA(p,R_SOL));
		ZIBStdOut(globBuf);
	  }
    return true;
  }

static INT CntP(p)
  PT *p;
  {
	countP++;
	if ((p->vec)==nil)
	  {
	    sprintf(globBuf, "Check: missing associated array at point %d\n",
		        p->index);
		ZIBStdOut(globBuf);
	  }
	return true;
  }

static INT CntE(ed)
  EDG *ed;
  {
    INT k = 0;
	return true;
  } 

static REAL Jacobi(p1,p2,p3)
  PT *p1, *p2, *p3;
  {
    return ((p2->x)-(p1->x))*((p3->y)-(p1->y))-
           ((p3->x)-(p1->x))*((p2->y)-(p1->y));
  }

static INT CheckOrien(t)
  TR2 *t;
  {
    REAL J = Jacobi(t->p1,t->p2,t->p3);
    EDG **eds = &(t->e1);
    PT **ps = &(t->p1), *pe1, *pe2;
    INT i, k, z1, z2;
    INT notOrien = false, notThreeP = false, notOpposit = false;
 
    if ((t->type)==T_GREEN_1) return true;
    if (J<=ZERO) notOrien = true;
    for (k=0; k<3; k++)
      {
        pe1 = eds[k]->p1; pe2 = eds[k]->p2;
        z1 = 0; z2 = 0;
        for (i=0; i<3; i++)
          {
            if (pe1==ps[i]) z1++;
            if (pe2==ps[i]) z2++;
          }
        if ((z1!=1)||(z2!=1)) notThreeP = true;
        if ((pe1==ps[k])||(pe2==ps[k])) notOpposit = true;
      }
    if (notOrien)
      {
        PrintTr("Not oriented",t);
      }
    if (notThreeP)
      {
        PrintTr("Edges contain more than 3 Points",t);
      }
    if (notOpposit)
      {
        PrintTr("Points not opposit to edge",t);
      }
 
    return true;
  }

static INT CntT(t)
  TR2 *t;
  {
    CheckOrien(t);
	countT++;
	return true;
  }

void CountLevel(no)
  INT no;
  {
    countP =0; countE = 0; countT = 0;
	if (no<(actTriang->refLevel))
	  {
		ApplyPLevel(CntP, no);
		ApplyELevel(CntE, no);
		ApplyTLevel(CntT, no);
	  }
	else
	  {
		ApplyP(CntP, all);
		ApplyE(CntE, nodal);
		ApplyT(CntT, all);
	  }
	sprintf(globBuf, "Tri: level %2d - %4d points, %4d edges, %4d triangles counted\n",
			no, countP, countE, countT);
	ZIBStdOut(globBuf);
	return;
  }

INT TriChk(cmd)
  COMMAND *cmd;
  {
    INT k;

	if (ParsCheck(cmd,0,0)) return false;
	if (actTriang==nil)
	  { ZIBStdOut("Tri: no triangulation (TriChk)\n"); return false; }

	sprintf(globBuf, "Tri: counting on %d level(s)\n",
			(actTriang->refLevel)+1);
	ZIBStdOut(globBuf);
	for (k = 0; k<=(actTriang->refLevel); k++)
	  CountLevel(k);
	return true;
  }


INT CheckTri(cmd)
  COMMAND *cmd;
  {
    INT k;

	if (ParsCheck(cmd,0,0)) return false;
	if (actTriang==nil)
	  { ZIBStdOut("Tri: no triangulation (CheckTri)\n"); return false; }

	sprintf(globBuf, "Check the triangulation data\n");
	ZIBStdOut(globBuf);

     if (ApplyT3(CheckTd,all)==false)
       {
        ZIBStdOut("CheckTri: error in triangulation\n");
        return false;
       }
     else
       {
        ZIBStdOut("CheckTri: ok \n");
       }

	return true;
  }
	
